package mt.edu.um.nlgapp.docplan;

import java.util.ArrayList;
import java.util.List;

import mt.edu.um.nlgapp.microplan.LexicalTemplate;
import uk.ac.abdn.carados.OntologyInstanceHandler;

/**
 * This is the class for a document plan. A document plan has the form of a
 * tree, where each node (which is itself a <code>DocPlan</code>) has a
 * relationship to its parent (<code>null</code> by default), and an ontology
 * instance, which represents the message at that particular node. It also has a
 * status, a value from {@link mt.edu.um.nlgapp.docplan.DocStatus}, which
 * indicates whether the current node is a document, a section, a paragraph or a
 * sentence. DocPlan nodes also have a
 * {@link mt.edu.um.nlgapp.microplan.LexicalTemplate}, which should be set via calls to
 * {@link mt.edu.um.nlgapp.microplan.Microplanner#microplanDoc(DocPlan)};
 * 
 * @author bertugatt
 * 
 */
public class DocPlan {

	private DocStatus status;
	private OntologyInstanceHandler message;
	private String relationToParent;
	private DocPlan parent;
	private List<DocPlan> children;
	private LexicalTemplate template;

	/**
	 * Default (empty) constructor
	 */
	public DocPlan() {
		children = new ArrayList<DocPlan>();
	}

	/**
	 * @return the status
	 */
	public DocStatus getStatus() {
		return status;
	}

	/**
	 * Set the status of this document node, indicating what its
	 * 
	 * @param status
	 *            the status to set
	 */
	public void setStatus(DocStatus status) {
		this.status = status;
	}

	/**
	 * @return the message
	 */
	public OntologyInstanceHandler getMessage() {
		return message;
	}

	/**
	 * @param message
	 *            the message to set
	 */
	public void setMessage(OntologyInstanceHandler message) {
		this.message = message;
	}

	/**
	 * @return the relationToParent
	 */
	public String getRelationToParent() {
		return relationToParent;
	}

	/**
	 * @param relationToParent
	 *            the relationToParent to set
	 */
	public void setRelationToParent(String relationToParent) {
		this.relationToParent = relationToParent;
	}

	/**
	 * @return the parent node in the tree. This method should return
	 *         <code>null</code> if, and only if, the current node is the root
	 *         of the document.
	 */
	public DocPlan getParent() {
		return parent;
	}

	/**
	 * @return the children
	 */
	public List<DocPlan> getChildren() {
		return children;
	}

	/**
	 * @param children
	 *            the children to set
	 */
	public void setChildren(List<DocPlan> children) {

		for (DocPlan child : children) {
			addChild(child);
		}
	}

	/**
	 * Add a child to this document plan. Children are held in the order in
	 * which they are added.
	 * 
	 * @param child
	 *            The child
	 */
	public void addChild(DocPlan child) {
		child.parent = this; // set the current node as parent;
		this.children.add(child);
	}

	/**
	 * Add a child to this document plan, also specifying what the relation
	 * should be between the child and the current node (e.g. "cause").
	 * 
	 * @param child
	 *            The child
	 */
	public void addChild(DocPlan child, String parentRelation) {
		addChild(child);
		child.setRelationToParent(parentRelation);
	}

	/**
	 * Set the lexical template corresponding to the message in this node;
	 * 
	 * @param template the template
	 */
	public void setLexicalTemplate(LexicalTemplate template) {
		this.template = template;
	}

	/**
	 * @return the lexical template corresponding to this node;
	 */
	public LexicalTemplate getLexicalTemplate() {
		return this.template;
	}
}
