/*
 * File: DOMUtil
 *
 * CHANGE HISTORY
 * 0.2	16/01/09	Base line file
 * 0.3	30/01/09	Added links to DOM specification and tutorials.	
 * 0.4	09/02/09	Externalised strings.
 * 1.1	11/08/09	Added memory clean up.
 * 
 * NB - Version numbers refer to the package version not the file version!
 */

/*
 * Copyright (c) 2007-2007, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.aries;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URI;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * <p>This class contains a collection of static methods relating to an application's
 * interaction with the DOM.</p>
 * <h4>Release Notes</h4>
 * <p>Version 0.4 - externalised strings.</p>
 * <p>Version 0.3 - added links to specification and tutorials on XPath.</p>
 * @see <a href="http://www.w3.org/DOM/">W3C DOM specification</a>
 * @see <a href="http://www.w3schools.com/dom/default.asp">W3Schools DOM Tutorial</a>
 * @author D. Westwater, University of Aberdeen
 * @since Aries 0.2
 */
public abstract class DOMUtil
{
	/**
	 * Saves a DOM document to a given file location.
	 * @param doc the document to be saved.
	 * @param outputFile the file to be saved to.
	 * @throws DOMUtilException if any problems are encountered while saving 
	 * the document.
	 */
	public final static void saveDocument(Document doc, String outputFile)
	throws DOMUtilException
	{
		// create output stream
		FileOutputStream stream = null;
		TransformerFactory tFactory = null;
		Transformer transformer = null;
		DOMSource source = null;
		
		try {
			stream = new FileOutputStream(new File(outputFile));

			// create result obj
			StreamResult result = new StreamResult(stream);

			// transform doc
			tFactory = TransformerFactory.newInstance();
			transformer = tFactory.newTransformer();
			source = new DOMSource(doc);
			transformer.transform(source, result);
		} catch (FileNotFoundException e) {
			throw new DOMUtilException(AriesMessages.getString("ARIES.NO_FILE") +  //$NON-NLS-1$
					outputFile);
		} catch (TransformerConfigurationException e) {
			throw new DOMUtilException(AriesMessages.getString("ARIES.NO_TRANSFORM") +  //$NON-NLS-1$
					e.getMessage());
		} catch (TransformerException e) {
			throw new DOMUtilException(AriesMessages.getString("ARIES.NO_TRANSFORMATION") +  //$NON-NLS-1$
					e.getMessage());
		} finally {
			if (stream != null) {
				try {
					stream.close();
				} catch (IOException e) {
					// Do nothing
				}
				stream = null;
				tFactory = null;
				transformer = null;
				source = null;
			}
		}
	}

	/**
	 * Creates a DOM document from the given resource.
	 * @param uri the location of the resource.
	 * @return the created document.
	 * @throws DOMUtilException if any problems are encountered while loading 
	 * the document.
	 */
	public final static Document loadDocument(URI uri) throws DOMUtilException
	{
		Document doc = null;
		DocumentBuilder builder = null;
		
		try {
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			if (factory != null)
			{
				builder = factory.newDocumentBuilder();
				if (builder != null)
				{
					doc = builder.parse(uri.toString());
				}
				else
				{
					throw new DOMUtilException(AriesMessages.getString("ARIES.NO_DOC_BUILDER")); //$NON-NLS-1$
				}
			}
			else
			{
				throw new 
				DOMUtilException(AriesMessages.getString("ARIES.NO_DOC_BUILDER_FACTORY")); //$NON-NLS-1$
			}
		} catch (ParserConfigurationException e) {
			throw new DOMUtilException(AriesMessages.getString("ARIES.NO_PARSE_CONFIG") +  //$NON-NLS-1$
					e.getMessage());
		} catch (SAXException e) {
			throw new DOMUtilException(AriesMessages.getString("ARIES.NO_PARSE_RESOURCE") + e.getMessage()); //$NON-NLS-1$
		} catch (IOException e) {
			throw new DOMUtilException(AriesMessages.getString("ARIES.NO_READING_RESOURCE") + e.getMessage()); //$NON-NLS-1$
		} finally {
			builder = null;
		}
		return doc;
	}
}
