/*
 * File: ParameterMap
 *
 * CHANGE HISTORY
 * 0.6	18/03/09	Base line file
 * 0.7	27/03/09	Added getBooleanParameter(String), getParameterNames(), 
 * 					numberOfParameters(), decodeParameters(Document).
 * 1.1	11/08/09	Added memory clean up.
 */

/*
 * Copyright (c) 2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.aries;

import java.util.Date;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * <p>This is a convenience wrapper around a Map of key, object values. It is particularly
 * useful for generating parameters from strings within an XML configuration file. It 
 * can also be used as a general purpose data-storage for passing information between
 * different applications, modules or classes.</p>
 * 
 * <h4>Change History</h4>
 * <p>
 * Version 0.7
 * <ul><li>Added new methods: <code>getBooleanParameter(String)</code>, 
 * <code>getParameterNames()</code>, <code>numberOfParameters()</code>, 
 * <code>decodeParameters(Document)</code>
 * </li>
 * </ul>
 * </p>
 * 
 * @author D Westwater, University of Aberdeen.
 * @since 0.6
 */
public class ParameterMap {

	/** The map of parameters linking a key (string) to a value (object). */
	protected Map<String, Object> _parameters = new TreeMap<String, Object>();

	/** The node path for parameters in the DOM. */
	final static private String PARAMETER_NODE_PATH = "/configuration/parameter"; //$NON-NLS-1$
	
	/** The tag for name nodes. */
	final private static String NAME_NODE = "name"; //$NON-NLS-1$
	
	/** The tag for parameter value nodes. */
	final private static String VALUE_NODE = "value"; //$NON-NLS-1$

	/**
	 * Adds a number of parameters to this scenario's current list. Parameters
	 * are stored as (name:String, value:Object) pairs.
	 * @param parameters the map of parameters to be added.
	 */
	final public void setParameters(Map<String, Object> parameters) {
		this._parameters.putAll(parameters);
	}

	/**
	 * Adds a single parameter for this scenario.
	 * @param name the name of the parameter.
	 * @param value the parameter's value.
	 */
	final public void setParameter(String name, Object value) {
		this._parameters.put(name, value);
	}

	/**
	 * Removes the named parameter from the list
	 * @param name the name of the parameter to be removed.
	 */
	final public void removeParameter(String name) {
		this._parameters.remove(name);
	}

	/**
	 * Clears all parameters from this scenario.
	 */
	final public void clearParameters() {
		this._parameters.clear();
	}

	/**
	 * Retrieves the object value for the named parameter.
	 * @param name the name of the parameter.
	 * @return the value of the parameter, this may be <code>null</code>
	 * if the parameter does not exist.
	 */
	final public Object getParameter(String name)
	{
		return name != null ? this._parameters.get(name) : null;
	}

	/**
	 * Retrieves the string value for the named parameter. If the value cannot
	 * be cast as a String then the <code>toString</code> method is called. 
	 * @param name the name of the parameter.
	 * @return the string representation of the value, this may be <code>null</code>
	 * if the parameter does not exist.
	 */
	final public String getStringParameter(String name) {
		String string = null;
		if (name != null) {
			Object value = this._parameters.get(name);
			if (value != null) {
				if (value instanceof String) {
					string = (String)value;
				}
				else {
					string = value.toString();
				}
			}
			value = null;
		}
		return string;
	}

	/**
	 * Retrieves the float value for the named parameter. If the value cannot
	 * be cast as a float then <code>Float.NaN</code> is returned. 
	 * @param name the name of the parameter.
	 * @return the float value, this may be <code>Float.NaN</code>
	 * if the parameter does not exist or is not a float.
	 */
	final public float getFloatParameter(String name) {
		float number = Float.NaN;
		if (name != null) {
			Object value = this._parameters.get(name);
			if (value != null) {
				if (value instanceof Number) {
					number = ((Number)value).floatValue();
				}
				else if (value instanceof String) {
					try {
						number = Float.parseFloat((String)value);
					}
					catch (NumberFormatException nfe) {
						number = Float.NaN;
					}
				}
			}
			value = null;
		}
		return number;
	}

	/**
	 * Retrieves the double value for the named parameter. If the value cannot
	 * be cast as a double then <code>Double.NaN</code> is returned. 
	 * @param name the name of the parameter.
	 * @return the double value, this may be <code>Double.NaN</code>
	 * if the parameter does not exist or is not a double.
	 */
	final public double getDoubleParameter(String name) {
		double number = Double.NaN;
		if (name != null) {
			Object value = this._parameters.get(name);
			if (value != null) {
				if (value instanceof Number) {
					number = ((Number)value).doubleValue();
				}
				else if (value instanceof String) {
					try {
						number = Double.parseDouble((String)value);
					}
					catch (NumberFormatException nfe) {
						number = Double.NaN;
					}
				}
			}
			value = null;
		}
		return number;
	}

	/**
	 * Retrieves the integer value for the named parameter. If the value cannot
	 * be cast as an integer then <code>0</code> is returned. 
	 * @param name the name of the parameter.
	 * @return the integer value, which will be 0 if the casting fails or the 
	 * parameter does not exist.
	 */
	final public int getIntParameter(String name) {
		int number = 0;
		if (name != null) {
			Object value = this._parameters.get(name);
			if (value != null) {
				if (value instanceof Number) {
					number = ((Number)value).intValue();
				}
				else if (value instanceof String) {
					try {
						number = Integer.parseInt((String)value);
					}
					catch (NumberFormatException nfe) {
						number = 0;
					}
				}
			}
			value = null;
		}
		return number;
	}

	/**
	 * Retrieves the long value for the named parameter. If the value cannot
	 * be cast as a long then <code>0</code> is returned. 
	 * @param name the name of the parameter.
	 * @return the long value, which may be 0 if the casting fails or the parameter
	 * does not exist.
	 */
	final public long getLongParameter(String name) {
		long number = 0L;
		if (name != null) {
			Object value = this._parameters.get(name);
			if (value != null) {
				if (value instanceof Number) {
					number = ((Number)value).longValue();
				}
				else if (value instanceof String) {
					try {
						number = Long.parseLong((String)value);
					}
					catch (NumberFormatException nfe) {
						number = 0L;
					}
				}
			}
			value = null;
		}
		return number;
	}
	
	/**
	 * Retrieves the data value for the named parameter. If the value cannot be 
	 * case as a Date then <code>null</code> is returned.
	 * @param name the name of the parameter.
	 * @return the Date value, which may be null if the casting fails or the 
	 * parameter does not exist.
	 */
	final public Date getDateParameter(String name) {
		Object value = this._parameters.get(name);
		
		return (value instanceof Date) ? (Date)value : null;
	}
	
	
	/**
	 * Retrieves the boolean value for the named parameter. If the value cannot
	 * be cast as a boolean then <code>false</code> is returned. 
	 * @param name the name of the parameter.
	 * @return the boolean value, which will be false if the casting fails or the 
	 * parameter does not exist.
	 * @since 0.7
	 */
	final public boolean getBooleanParameter(String name) {
		boolean bool = false;
		if (name != null) {
			Object value = this._parameters.get(name);
			if (value != null) {
				if (value instanceof Boolean) {
					bool = ((Boolean)value).booleanValue();
				}
				else if (value instanceof String) {
					bool = Boolean.parseBoolean((String)value);
				}
			}
			value = null;
		}
		return bool;
	}
	
	/**
	 * Returns the names of all of the parameters 
	 * 
	 * @return a Set of parameter names as Strings
	 */
	final public Set<String> getParameterNames(){
		return this._parameters.keySet();
	}

	 /**
	 * Returns the number of parameters
	 * @return the number of parameters
	 */
	final public int numberOfParameters(){
		return this._parameters.size();
	}
	/**
	 * Decodes the parameters from the DOM and adds them to the scenario.
	 * @param doc the DOM document being used.
	 * @throws XPathExpressionException if there is a problem with the XPath 
	 * expression
	 * @throws XPathUtilException if there is a problem evaluating the XPath.
	 */
	public final void decodeParameters(Document doc) 
	throws XPathExpressionException, XPathUtilException {
		
		XPath xpath = XPathFactory.newInstance().newXPath();
		NodeList nodelist = (NodeList)xpath.evaluate(
				PARAMETER_NODE_PATH, doc, XPathConstants.NODESET);
		
		if (nodelist != null) {
			int length = nodelist.getLength();
			Node currentNode = null;
			String name = null;
			String value = null;
			
			for (int i = 0; i < length; i++)
			{
				currentNode = nodelist.item(i);
				if (currentNode != null)
				{
					name = XPathUtil.extractValue(currentNode, NAME_NODE);
					value = XPathUtil.extractValue(currentNode, VALUE_NODE);
					
					if (name != null && value != null)
					{
						setParameter(name,value);
					}
				}
			}
			currentNode = null;
			name = null;
			value = null;
		}
		xpath = null;
		nodelist = null;
	}
	
	public Map<String, Object> getMap() {
		return this._parameters;
	}
	
	public void setParameters(ParameterMap parameterMap) {
		this._parameters = parameterMap.getMap();
	}
	
	public void addParameters(ParameterMap parameterMap) {
		this._parameters.putAll(parameterMap.getMap());
	}
}
