/*
 * File: XPathUtil
 *
 * CHANGE HISTORY
 * 0.1	12/01/09	Base line file
 * 0.3	20/01/09	Fixed problem with evaluating paths from nodes.
 * 		30/01/09	Added links to DOM specification and tutorials.
 * 0.4	09/02/09	Externalised strings
 * 0.6	18/03/09	The values retrieved by extractValue and getAttributeValue are now 
 * 					trimmed of leading and trailing white spaces.
 * 1.1	11/08/09	Added memory clean up.
 * 
 * NB - Version numbers refer to the package version not the file version!
 */

/*
 * Copyright (c) 2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.aries;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

/**
 * <p>
 * This class contains a collection of static methods relating to an application's
 * interaction with the XPath facility used by DOM.
 * </p>
 * <h4>Release Notes</h4>
 * <p>Version 0.6<br>
 * <ul>
 * <li>The values retrieved by extractValue and getAttributeValue are now trimmed of
 * leading and trailing white spaces.</li>
 * </ul>
 * </p>
 * <p>Version 0.5 - added a new method for extracting attribute values from a node.
 * </p>
 * <p>Version 0.4 - externalised strings.</p>
 * <p>Version 0.3 - added links to specification and tutorials on XPath. 
 * Fixed problem with the extraction of values from child
 * nodes.</p>
 * @see <a href="http://www.w3.org/Style/XSL/">W3C XPath specification</a>
 * @see <a href="http://www.w3schools.com/XPath/default.asp">W3Schools XPath Tutorial</a>
 * @author D. Westwater, University of Aberdeen
 */
public abstract class XPathUtil {
	/**
	 * <p>Extracts the value of the node path contained within a top level node.</p>
	 * <h4>Example:</h4>
	 * <p>
	 * Given the following XML extract:</br>
	 * <blockquote>
	 * <code>
	 * &lt;pets&gt;<br>
	 * &nbsp;&nbsp;&nbsp;&lt;dog&gt;fido&lt;/dog&gt;<br>
	 * &nbsp;&nbsp;&nbsp;&lt;cat&gt;buttons&lt;/cat&gt;<br>
	 * &lt;/pets&gt;<br>
	 * </code>
	 * </blockquote>
	 * The call: <code>extractValue(animalNode, "dog");<code> would return the 
	 * value <em>fido</em>.
	 * </p>
	 * @param node the top level node to search down
	 * @param targetName the target path of the node to look for
	 * @return the value of this node. <code>null</code> is returned if the node
	 * contains new value, or the specified node cannot be found.
	 * @throws XPathUtilException if there is a problem reading the XPath
	 */
	static public final String extractValue(Node node, String targetName)
	throws XPathUtilException
	{
		XPath xPath = XPathFactory.newInstance().newXPath();
		String nodeValue = null;
		
		if (xPath != null)
		{
			Node childNode = null;
			try
			{
				childNode = (Node)xPath.evaluate(targetName, node, 
						XPathConstants.NODE);
			} catch (XPathExpressionException e)
			{
				throw new XPathUtilException(AriesMessages.getString("ARIES.NODE_PATH_ERROR") +  //$NON-NLS-1$
						targetName);
			}
			if (childNode != null)
			{
				Node childChildNode = childNode.getFirstChild(); 
				if (childChildNode != null)
				{
					nodeValue = childChildNode.getNodeValue();
					childChildNode = null;
				}
				childNode = null;
			}
		}
		xPath = null;
		return nodeValue != null ? nodeValue.trim() : null;
	}
	
	/**
	 * Retrieves the value of a required attribute from the given node.
	 * @param attributeName the name of the attribute being sought.
	 * @param parentNode the node the attribute belongs to.
	 * @return a string representation of the attribute value.
	 * @since 0.5
	 */
	static public final String getAttributeValue(String attributeName, Node parentNode) {
		String value = null;
		NamedNodeMap map = parentNode.getAttributes();
		Node attribute = map.getNamedItem(attributeName);
		map = null;
		
		if (attribute != null) {
			value = attribute.getNodeValue();
			attribute = null;
		}

		return value != null ? value.trim() : null;
	}
}
