/*
 * File: OntologyClassHandler
 *
 * CHANGE HISTORY
 * 0.1	15/01/09	Base line file
 * 0.4	02/02/09	The getInstances and getSortedInstances methods can now take an 
 * 					OntologyInstanceFilter to determine which instances should be 
 * 					included in the collection.
 * 0.11	23/06/09	Added new sorting method for allowing the specification of direction to 
 * 					sort in.
 */

/*
 * Copyright (c) 2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados;

import java.util.Collection;

import uk.ac.abdn.carados.util.OntologySortTree;

/**
 * <p>This interface is used to define how an application will interact with a
 * ontology classes in a generic ontology. In most cases, the actual
 * implementation of an ontology is irrelevant. This handler maps the common
 * features that all ontology classes should have.</p>
 * 
 * <h4>Release Notes</h4>
 * <p>
 * Version 0.4 - Added the OntologyInstanceFilter onto <code>getInstances</code> 
 * and <code>getSortedInstance</code>. This allows only those instances that meet 
 * the requirements of the filter to be included.
 * </p>
 * @author D. Westwater, University of Aberdeen
 */
public interface OntologyClassHandler extends OntologyEntityHandler {
	
	/**
	 * Adds a new super class to this class.
	 * 
	 * @param superclass
	 *            the new super class.
	 */
	public void addDirectSuperClass(OntologyClassHandler superclass);

	/**
	 * Adds a new subclass to this class.
	 * 
	 * @param subclass
	 *            the new subclass
	 */
	public void addSubClass(OntologyClassHandler subclass);
	
	/**
	 * Removes the given class from this classes list of super classes.
	 * 
	 * @param superclass the super class to be removed.
	 */
	public void removeDirectSuperClass(OntologyClassHandler superclass);

	/**
	 * Removes the given class from this classes list of sub classes.
	 * 
	 * @param subclass the subclass to be removed.
	 */
	public void removeSubClass(OntologyClassHandler subclass);

	/**
	 * <p>
	 * Retrieves all the instances of this class. The parameter determines if
	 * the call is transitive or not. A transitive call, where the parameter is
	 * <code>true</code>, returns all the instances of subclasses as well as
	 * this class.
	 * </p>
	 * 
	 * <p>
	 * The order the instances appear in the collection is completely arbitrary.
	 * See {@link #getSortedInstances(String,boolean) getSortedInstances} for
	 * retrieving the instances in a sorted order.
	 * </p>
	 * 
	 * @param transitive
	 *            set to <code>true</code> if instances of subclasses have to be
	 *            returned as well, otherwise this is <code>false</false>.
	 * @return a collection of <code>OntologyInstanceHandler</code>s
	 *         representing all the instances in the ontology.
	 * @see #getSortedInstances(String, boolean)
	 */
	public Collection<OntologyInstanceHandler> getInstances(boolean transitive);

	/**
	 * <p>
	 * Retrieves all the instances of this class that meet the requirements 
	 * defined by the filter. The second parameter determines if the call is 
	 * transitive or not. A transitive call, where the parameter is
	 * <code>true</code>, returns all the instances of subclasses as well as
	 * this class.
	 * </p>
	 * 
	 * <p>
	 * The order the instances appear in the collection is completely arbitrary.
	 * See {@link #getSortedInstances(String,boolean) getSortedInstances} for
	 * retrieving the instances in a sorted order.
	 * </p>
	 * 
	 * @param filter the filter for determining whether an instance should appear
	 * in the list or not.
	 * @param transitive
	 *            set to <code>true</code> if instances of subclasses have to be
	 *            returned as well, otherwise this is <code>false</false>.
	 * @return a collection of <code>OntologyInstanceHandler</code>s
	 *         representing all the instances in the ontology.
	 * @see #getSortedInstances(String, boolean)
	 * @since 0.4
	 */
	public Collection<OntologyInstanceHandler> getInstances(OntologyInstanceFilter filter, 
			boolean transitive);
	
	/**
	 * <p>This method returns all the instances of this ontology class as a
	 * sorted collection based on the given property name. The second parameter 
	 * determines if the call is transitive or not. A transitive call, where 
	 * the parameter is <code>true</code>, returns all the instances of 
	 * subclasses as well as this class. Sorting is done on the given property 
	 * name and is the natural order, e.g. a,b,c and 1,2,3. Sorting is 
	 * <em>only</em> done on direct instances.</p>
	 * 
	 * @param propertyName
	 *            the name of the property to sort on
	 * @param transitive
	 *            set to <code>true</code> if instances of subclasses have to be
	 *            returned as well, otherwise this is <code>false</false>.
	 * @return a collection of instances
	 */
	public Collection<OntologyInstanceHandler> getSortedInstances(
			String propertyName, boolean transitive);
	
	/**
	 * <p>This method returns all the instances of this ontology class, which 
	 * meet the requirements of the provided filter, as a sorted collection 
	 * based on the given property name. The third parameter determines if the 
	 * call is transitive or not. A transitive call, where the parameter is 
	 * <code>true</code>, returns all the instances of subclasses as well as 
	 * this class. Sorting is done on the given property name and is the 
	 * natural order, e.g. a,b,c and 1,2,3. Sorting is <em>only</em> done on 
	 * direct instances.</p>
	 * 
	 * @param filter the filter for determining whether an instance should appear
	 * in the list or not.
	 * @param propertyName
	 *            the name of the property to sort on
	 * @param transitive
	 *            set to <code>true</code> if instances of subclasses have to be
	 *            returned as well, otherwise this is <code>false</false>.
	 * @return a collection of instances
	 * @since 0.4
	 */
	public Collection<OntologyInstanceHandler> getSortedInstances(
			OntologyInstanceFilter filter, String propertyName, boolean transitive);

	/**
	 * Gets all the subclasses of this class. The boolean parameter specifies
	 * whether the subclass relation is transitive.
	 * 
	 * @param transitive
	 *            if <code>true</code>, returns the transitive closure of the
	 *            subclass relation starting from the current class.
	 * @return the subclasses, if any, an empty collection otherwise.
	 */
	public Collection<OntologyClassHandler> getSubClasses(boolean transitive);

	/**
	 * Gets all the superclasses of this class. The boolean parameter specifies
	 * whether the superclass relation is transitive.
	 * 
	 * @param transitive
	 *            if <code>true</code>, returns the transitive closure of the
	 *            superclass relation starting from the current class.
	 * @return the superclasses, if any, an empty collection otherwise.
	 */
	public Collection<OntologyClassHandler> getSuperClasses(boolean transitive);

	/**
	 * Check whether this class has a subclass with the given name.
	 * 
	 * @param className
	 *            the name of the subclass to check
	 * @return <code>true</code> if the this class has the subclass
	 */
	public boolean hasSubClass(OntologyClassHandler className,
			boolean transitive);

	/**
	 * Check whether this class has a superclass with the given name.
	 * 
	 * @param className
	 *            the name of the superclass
	 * @return <code>true</code> if this class has the superclass
	 */
	public boolean hasSuperClass(OntologyClassHandler className,
			boolean transitive);

	/**
	 * Sets the default value of the given property to the given value. What
	 * constitutes a default value is entirely dependent on the underlying
	 * ontology being used. If the value is <code>null</code> then the default
	 * value should be deleted.
	 * 
	 * @param propertyName
	 *            the name of the property to be set
	 * @param value
	 *            the new value the property has
	 */
	public void setDefaultValue(String propertyName, Object value);

	/**
	 * Set the superclass of this class. This method replaces all previously
	 * defined superclasses.
	 * 
	 * @param superclass
	 *            the new superclass
	 */
	public void setDirectSuperClass(OntologyClassHandler superclass);

	/**
	 * Creates an instance of this class in the underlying ontology. The instance
	 * is given a default name.
	 * 
	 * @return the reference to the instance.
	 */
	public OntologyInstanceHandler createInstance();

	/**
	 * Creates an instance of this class in the underlying ontology. The instance
	 * is given the name supplied.
	 * @param name the name for this instance.
	 * @return the reference to the instance.
	 */
	public OntologyInstanceHandler createInstance(String name);
	
	/**
	 * This method performs the sorting on the instances of a class.
	 * @param filter the filter to be used, may be <code>null</code> for all instances.
	 * @param propertyName the property to be sorted on.
	 * @param transitive if <code>true</code> then all instances of sub-classes are also returned.
	 * If this is <code>false</code> then all the instance of this class are used.
	 * @param order a flag to determine the ordering. Use <code>OntologySortTree.ASCENDING_ORDER</code>
	 * or <code>OntologySortTree.DESCENDING_ORDER</code>.
	 * @return the sorted collection as an <code>ArrayList</code>.
	 * @see OntologySortTree
	 * @since 0.11
	 */
	Collection<OntologyInstanceHandler> getSortedInstances(
			OntologyInstanceFilter filter, String propertyName,
			boolean transitive, boolean order);
}
