/*
 * File: OntologyEntityHandler
 *
 * CHANGE HISTORY
 * 0.1	16/01/09	Base line file
 * 0.3	29/01/09	Added isA(String) method for efficiency with Jess.
 */

/*
 * Copyright (c) 2007-2007, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados;

import java.util.Collection;
import java.util.TreeMap;

/**
 * <p>This interface is used to define how an application will interact with an 
 * ontology entity in a generic ontology. In most cases, the actual implementation
 * of an ontology and its entities is irrelevant. This handler maps the common 
 * features that all ontology entities should have. An ontology entity is the common
 * super interface of both ontology classes and ontology instances.</p>
 * <h4>Release Notes</h4>
 * <p>
 * Version 0.3 - Added <em>isA(String)</em> method to the OntologyEntityHandler. 
 * This is then applied to ClassHandlers and InstanceHandlers.
 * </p>
 * @author D. Westwater & A. Gatt, University of Aberdeen
 */
public interface OntologyEntityHandler
{
	/**
	 * Deletes this element from the ontology.
	 */
	public abstract void delete();

	/**
	 * Retrieves the default value for the given property. The 
	 * definition of a default value is completely dependent on 
	 * the underlying ontology being used.
	 * @param propertyName the property whose value we want
	 * @param transitive <code>true</code> if the super classes have to be searched,
	 * <code>false</code> otherwise.
	 * @return the value of the default property
	 */
	public abstract Object getDefaultValue(String propertyName, boolean transitive);

	/**
	 * Retrieves all the default values for this entity. The 
	 * definition of a default value is completely dependent on 
	 * the underlying ontology being used.
	 * @param transitive <code>true</code> if the super classes have to be searched,
	 * <code>false</code> otherwise.
	 * @return a map of the <property name, value> pairs.
	 */
	public abstract TreeMap<String, Object> getAllDefaultValues(boolean transitive);

	/**
	 * Retrieves the handler for the underlying ontology for this entity.
	 * @return a reference to the <code>OntologyHandler</code> that represents
	 * the ontology.
	 */
	public OntologyHandler getOntology();

	/**
	 * Retrieves the name of this element. The actual construction of the name
	 * will be dependent on the ontology being used.
	 * @return a string representation of the element's name.
	 */
	public String getName();

	/**
	 * Retrieves all the property names for this element.
	 * @return a collection of strings representing the property names
	 */
	public Collection<String> getPropertyNames();

	/**
	 * Determines if the entity is of the same type as the given class.
	 * @param ontologyClass the class to be checked against
	 * @return <code>true</code> if the element is of this type, 
	 * <code>false</code> otherwise
	 */
	public boolean isA(OntologyClassHandler ontologyClass);

	/**
	 * Determines if the entity is of the same type as the given class.
	 * @param className the name of the class to be checked against
	 * @return <code>true</code> if the element is of this type, 
	 * <code>false</code> otherwise
	 * @since 0.3
	 */
	public boolean isA(String className);
	
	/**
	 * Checks the property list for this entity and determines if it 
	 * has the property associated with it.
	 * @param propertyName the name of the property.
	 * @return <code>true</code> if the entity has the property, <code>false</code>
	 * otherwise.
	 */
	public boolean hasProperty(String propertyName);
}
