/*
 * File: OntologyHandler
 *
 * CHANGE HISTORY
 * 0.1	14/01/09	Base line file
 * 0.8	05/03/09	Added new method for accepting a filter when grabbing all the 
 * 					instances in the ontology.
 */

/*
 * Copyright (c) 2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados;

import java.net.URI;
import java.util.Collection;
import java.util.Date;

/**
 * <p>
 * The <code>OntologyHandler</code> is used to define how an application will 
 * interact with a generic ontology. In most cases, the actual implementation
 * of an ontology is irrelevant. This handler maps the common features that 
 * all ontologies should have.
 * </p>
 * <h4>Release Notes</h4>
 * <p>Version 0.8 - Added new method: 
 * <code>getAllInstances(OntologyInstanceHandler)</code>
 * </p>
 * @author D. Westwater & A. Gatt, University of Aberdeen
 */
public interface OntologyHandler
{
	/**
	 * Creates a class with the given name. The class is stored in the ontology.
	 * Note that class names are unique; hence, if a class with the given name
	 * exists, this method should replace it.
	 * @param className the name of the class to be created
	 * @return the <code>OntologyClassHandler</code> for the ontology class, 
	 * or <code>null</code> if the class could not be created.
	 */
	public OntologyClassHandler createClass(String className);

	/**
	 * Gets the class with the given name.
	 * 
	 * @param className the class name
	 * @return the <code>OntologyClassHandler</code> for the class, if defined, 
	 * <code>null</code> otherwise
	 */
	public OntologyClassHandler getOntologyClass(String className);

	/**
	 * Gets all the classes defined in the ontology.
	 * 
	 * @return a Collection of <code>OntologyClassHandler<code>s for all the classes in the
	 * ontology. If there are no classes, an empty collection is returned.
	 */
	public Collection<OntologyClassHandler> getAllOntologyClasses();

	/**
	 * Check whether a class with the given name exists in this ontology
	 * @param className the name of the class
	 * @return <code>true</code> if the class has been defined in the ontology
	 */
	public boolean containsClass(String className);
	
	/**
	 * Checks whether the referred class is contained within this ontology.
	 * @param classHandler the <code>OntologyClassHandler</code> reference to an ontology class.
	 * @return <code>true</code> if the class has been defined in the ontology
	 */
	public boolean containsClass(OntologyClassHandler classHandler);

	/**
	 * Retrieves all the instances in the ontology.
	 * 
	 * @return a collection of <code>OntologyInstanceHandler</code>s for all the instances
	 * in the ontology. If there are no instances, an empty collection is returned.
	 */
	public Collection<OntologyInstanceHandler> getAllInstances();

	/**
	 * Retrieves all the instances in the ontology.
	 * 
	 * @param filter the filter to be used for determining if the instances should be
	 * included in the returned collection.
	 * @return a collection of <code>OntologyInstanceHandler</code>s for all the instances
	 * in the ontology. If there are no instances, an empty collection is returned.
	 * @since 0.8
	 */
	public Collection<OntologyInstanceHandler> getAllInstances(OntologyInstanceFilter filter);

	/**
	 * Checks whether this ontology contains the instance referred to by the given
	 * <code>OntologyInstanceHandler</code>.
	 * @param instanceHandler the <code>OntolgyInstanceHandler</code> referring to the 
	 * instance.
	 * @return <code>true</code> if this ontology contains such an instance.
	 */
	public boolean containsInstance(OntologyInstanceHandler instanceHandler);

	/**
	 * Retrieves all the names of properties defined in the ontology.
	 * 
	 * @return a <code>Collection</code> of <code>String</code>s representing
	 * the property names.
	 */
	public Collection<String> getPropertyNames();

	/**
	 * Checks whether the ontology contains the given property.
	 * @param propertyName the property name
	 * @return <code>true</code> if the property is defined in this ontology.
	 */
	public boolean containsProperty(String propertyName);

	/**
	 * Retrieves the root class of the hierarchy. The root class is defined as the
	 * top-most class in the taxonomy, such that all classes in the ontology are
	 * subclasses of this class.
	 * 
	 * @return a <code>OntologyClassHandler</code> for the root class. If there is 
	 * no root class, this should return <code>null</code>
	 */
	public OntologyClassHandler getRootClass();

	/**
	 * Loads the ontology from a specified URI location.
	 * @param uri the location of the file to be loaded.
	 * @throws OntologyException if any exception is encountered during the 
	 * process of loading the file.
	 */
	public void loadOntology(URI uri) throws OntologyException;

	/**
	 * Outputs the ontology to a file
	 * 
	 * @param fileName the output file
	 * @throws OntologyException if any exception is encountered during the process of writing
	 * the file.
	 */
	public void saveOntology(String fileName) throws OntologyException;
	
	/**
	 * Translates a Date object into an object that is usable by the ontology. Not all 
	 * ontologies will handle dates in the same manner. 
	 *
	 * @param date the date object to be translated
	 * @return the translated object, which is intended to be sent directly to the ontology.
	 */
	public Object translateDate(Date date);
	
	/**
	 * Translates a timestamp (long) into an object that is usable by the ontology. 
	 * Not all ontologies will handle dates in the same manner. 
	 * @param timestamp the long timestamp representing a date that is to be translated.
	 * @return the translated object, which is intended to be sent directly to the ontology.
	 */
	public Object translateDate(long timestamp);
	
	/**
	 * Converts the ontology object representing a date into a Java <code>Date</code> object.
	 * @param object the object to be converted.
	 * @return the converted <code>Date</code> object
	 */
	public Date translateToDate(Object object);
	
	/**
	 * Converts the ontology object representing a date into a timestamp (long).
	 * @param object the object to be converted.
	 * @return the converted timestamp (long).
	 */
	public long translateToTimestamp(Object object);

	/**
	 * Attaches an <code>OntologyListener</code> to the handler. This allows
	 * diagnostics to be run on the ontology.
	 * @param listener a <code>OntologyListener</code> to be attached.
	 */
	public void attachListener(OntologyListener listener);
	
	/**
	 * Removes the given <code>OntologyListener</code> from this handler's list of listeners.
	 * @param listener the <code>OntologyListener</code> to be removed.
	 */
	public void removeListener(OntologyListener listener);
	
	/**
	 * Removes all the <code>OntologyListener</code>s from this handler.
	 */
	public void removeAllListeners();
}
