/*
 * File: OntologyInstanceHandler
 *
 * CHANGE HISTORY
 * 0.1	16/01/09	Base line file
 * 0.4	03/02/09	Added new methods for doing the casting of property values 
 * 					locally.
 * 0.5	12/02/09	Added getOwnValueAsFloat() method.
 * 0.9	28/05/09	Added getOwnValueAsDouble() method.
 * 2.3	19/02/10	Added new method.
 */

/*
 * Copyright (c) 2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados;

import java.util.Collection;
import java.util.Date;

/**
 * <p>
 * This interface is used to define how an application will interact with a
 * ontology instances (or individuals) in a generic ontology. In most cases, the
 * actual implementation of an ontology is irrelevant. This handler maps the
 * common features that all ontology individuals should have.
 * </p>
 * <h4>Release Notes</h4>
 * <p>
 * Version 2.2 - Added getOwnValueAsCollection(String, Class&lt;T&gt;) method.
 * </p>
 * <p>
 * Version 0.9 - Added getOwnValueAsDouble()
 * </p>
 * <p>
 * Version 0.5 - Added getOwnValueAsFloat()
 * </p>
 * <p>
 * Version 0.4 - Added additional methods for local casting of property values.
 * </p>
 * 
 * @author D. Westwater & A. Gatt, University of Aberdeen
 */
public interface OntologyInstanceHandler extends OntologyEntityHandler {
	/**
	 * Retrieves the ontology class to which this instance belongs.
	 * 
	 * @return the handler to the ontology class
	 */
	public OntologyClassHandler getOntologyClass();

	/**
	 * Retrieves the value of a property for this instance. If the instance has
	 * no value set for the property, the implementation <b>may</b> return the
	 * default value for this instance's class.
	 * 
	 * @param propertyName
	 *            the property name
	 * @return the value of the property
	 */
	public Object getOwnValue(String propertyName);

	/**
	 * Retrieves the string value of a property for this instance. If the
	 * instance has no value set for the property, the implementation <b>may</b>
	 * return the default value for this instance's class. If no default can be
	 * found then <code>null</code> is returned. If the value of the property is
	 * not a string then the string representation is returned using the
	 * <code>toString</code> method.
	 * 
	 * @param propertyName
	 *            the property name
	 * @return the string value of the property
	 * @since 0.4
	 */
	public String getOwnValueAsString(String propertyName);

	/**
	 * Retrieves the value of a property for this instance as a date. If the
	 * instance has no value set for the property, the implementation <b>may</b>
	 * return the default value for this instance's class. If no value is found
	 * then <code>null</code> is returned.
	 * 
	 * @param propertyName
	 *            the property name
	 * @return the date value of the property
	 * @since 0.4
	 */
	public Date getOwnValueAsDate(String propertyName);

	/**
	 * Retrieves the value of a property for this instance as an Integer. If the
	 * instance has no value set for the property, the implementation <b>may</b>
	 * return the default value for this instance's class. If no value for the
	 * property is found then <code>null</code> is returned.
	 * 
	 * @param propertyName
	 *            the property name
	 * @return the value of the property
	 * @since 0.4
	 */
	public Integer getOwnValueAsInteger(String propertyName);

	/**
	 * Retrieves the value of a property for this instance as a Float. If the
	 * instance has no value set for the property, the implementation <b>may</b>
	 * return the default value for this instance's class. If no value for the
	 * property is found then <code>NULL</code> is returned.
	 * 
	 * @param propertyName
	 *            the property name
	 * @return the value of the property
	 * @since 0.5
	 */
	public Float getOwnValueAsFloat(String propertyName);

	/**
	 * Retrieves the value of a property for this instance as a Double. If the
	 * instance has no value set for the property, the implementation <b>may</b>
	 * return the default value for this instance's class. If no value for the
	 * property is found then <code>NULL</code> is returned.
	 * 
	 * @param propertyName
	 *            the property name
	 * @return the value of the property
	 * @since 0.9
	 */
	public Double getOwnValueAsDouble(String propertyName);

	/**
	 * Retrieves the value of a property for this instance. If the instance has
	 * no value set for the property, the implementation <b>may</b> return the
	 * default value for this instance's class. If there are no values for this
	 * property then the empty array is returned.
	 * 
	 * @param propertyName
	 *            the property name
	 * @return the value of the property
	 */
	public Object[] getOwnValueAsArray(String propertyName);

	/**
	 * Retrieves the values of a property for this instance as a Collection of
	 * objects. If the instance has no value set for the property, the
	 * implementation <b>may</b> return the default value for this instance's
	 * class. If no values are found then the empty collection is returned.
	 * 
	 * @param propertyName
	 *            the property name
	 * @return a collection of all the values for this property.
	 * @since 0.4
	 */
	public Collection<Object> getOwnValueAsCollection(String propertyName);

	/**
	 * Retrieves the values of a property for this instance as a Collection of
	 * the specified type. If the instance has no value set for the property,
	 * the implementation <b>may</b> return the default value for this
	 * instance's class. If no values are found then the empty collection is
	 * returned.
	 * 
	 * @param propertyName
	 *            the property name
	 * @return a collection of all the values for this property.
	 * @since 2.2
	 */
	public <T> Collection<T> getOwnValueAsCollection(String propertyName,
			Class<T> type);

	/**
	 * Sets the value for a property. This value may be different from the
	 * default value specified in the instance's parent class. Setting the
	 * property value will overwrite all previous values for this property.
	 * 
	 * @param propertyName
	 *            the property name to be set
	 * @param value
	 *            the value to be set
	 */
	public void setOwnValue(String propertyName, Object value);

	/**
	 * Adds the given value to the <em>"list"</em> of existing values. This is
	 * contrast to setOwnValue which overwrites existing values.
	 * 
	 * @param propertyName
	 *            the property name to be added to.
	 * @param value
	 *            the value to be added.
	 */
	public void addOwnValue(String propertyName, Object value);
}
