/*
 * File: OntologyListener
 *
 * CHANGE HISTORY
 * 0.1	15/01/09	Base line file
 * 0.8	27/02/09	Added new message for warning about no property existing.
 * 0.10	18/06/09	Added a new method for reporting on type conflicts.
 */

/*
 * Copyright (c) 2007-2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados;

/**
 * <p>
 * The <code>OntologyListener</code> is designed primarily to give diagnostic 
 * feedback concerning the use of the various OntologyHandlers. Listeners are 
 * informed of the following events:
 * <blockquote>
 * <ul>
 * <li>a new class is created in the ontology.</li>
 * <li>a new instance is created in the ontology.</li>
 * <li>the ontology is saved.</li>
 * <li>a class is given a new super class.</li>
 * <li>a class is given a new sub class.</li>
 * <li>a default value is changed.</li>
 * <li>a property value is changed.</li>
 * </ul>
 * </blockquote>
 * </p>
 * <h4>Release Notes</h4>
 * <p>Version 0.10 - added new method: typeConflict(String).</p>
 * <p>
 * Verison 0.8 - added a method for reporting missing properties.
 * </p>
 * @author Dave Westwater, University of Aberdeen
 */
public interface OntologyListener
{
	/**
	 * This method is called when a new class is created.
	 * @param ontologyClass the reference to the newly created class.
	 */
	public void classCreated(OntologyClassHandler ontologyClass);

	/**
	 * This method reports that a class has been deleted from the ontology.
	 * @param className the reference to the deleted class.
	 */
	public void classDeleted(String className);
	 
	/**
	 * Informs the listener that the ontology has been saved.
	 * @param fileName the location where the file has been saved.
	 * @param ontologyName the name of the ontology.
	 */
	public void saved(String fileName, String ontologyName);

	/**
	 * This method informs the listener that a super class has been added
	 * to the base class. The <code>overwritten</code> flag is used to 
	 * show if this super class addition replaces existing super classes.
	 * @param baseClass the reference to the base class.
	 * @param superClass the reference to the new super class.
	 * @param overwritten <code>true</code> if the new super class has 
	 * replaced all other super classes, <code>false</code> otherwise.
	 */
	public void superClassAdded(OntologyClassHandler baseClass, 
			OntologyClassHandler superClass, boolean overwritten);

	/**
	 * This method is called when a sub classes has been added to a class.
	 * @param baseClass the reference to the base class.
	 * @param subClass the reference to the new subclass.
	 */
	public void subClassAdded(OntologyClassHandler baseClass, 
			OntologyClassHandler subClass);
	
	/**
	 * Informs the listener that a default value on a property has been changed.
	 * @param ontologyClass the reference to the altered class.
	 * @param propertyName the name of the property that has been changed.
	 * @param propertyValue the new value of the property.
	 */
	public void defaultValueChanged(OntologyClassHandler ontologyClass, 
			String propertyName, Object propertyValue);

	/**
	 * This method informs the listener that a new instance has been created in
	 * the ontology.
	 * @param instance the reference to the new instance.
	 */
	public void instanceCreated(OntologyInstanceHandler instance);

	/**
	 * This method reports that an instance has been deleted from the ontology.
	 * @param instanceName the reference to the deleted instance.
	 */
	public void instanceDeleted(String instanceName);

	/**
	 * This method is called when a property value of an instance is changed.
	 * In the case where a new value is added to the <em>list</em> of values 
	 * (in other words, when <code>overwritten</code> is <code>false</code>)
	 * the <code>propertyValue</code> refers only to the newly added value and 
	 * not the complete <em>list</em> of values.
	 * @param instance the reference to the instance that has been altered.
	 * @param propertyName the name of the property that has been changed.
	 * @param propertyValue the new value given to the property.
	 * @param overwritten <code>true</code> if the new value has replaced 
	 * all other values, <code>false</code> otherwise.
	 */
	public void propertyValueChanged(OntologyInstanceHandler instance, 
			String propertyName, Object propertyValue, boolean overwritten);

	/**
	 * This method informs the listener that a super class has been removed
	 * to the base class. 
	 * @param baseClass the reference to the base class.
	 * @param superClass the reference to the new super class.
	 */
	public void superClassRemoved(OntologyClassHandler baseClass, 
			OntologyClassHandler superClass);

	/**
	 * This method is called when a sub classes has been removed from a class.
	 * @param baseClass the reference to the base class.
	 * @param subClass the reference to the new subclass.
	 */
	public void subClassRemoved(OntologyClassHandler baseClass, 
			OntologyClassHandler subClass);

	/**
	 * This method is called when the property being set doesn't already exist
	 * in the ontology.
	 * @param property the name of the property.
	 * @since 0.8
	 */
	public void missingProperty(String property);

	/**
	 * This method reports that a value for a property is in conflict with the defined 
	 * type for the property. The implementation may operate a strict typing policy that
	 * only allows values of the specified type. Other implementations, however, may 
	 * try a conversion between types first before throwing this error.
	 * @param message the resulting error message.
	 * @since 0.10
	 */
	public void typeConflict(String property, Object value);
}
