/*
 * File: Owl2CaradosClass
 *
 * CHANGE HISTORY
 * 0.1	15/01/09	Base line file
 * 0.2	20/01/09	Added check for null class in the constructor
 * 		21/01/09	Fixed bug with createInstance(String) entering an infinite loop. 
 * 					Corrected class description in JavaDoc.
 * 					The instance counter is removed from this file.
 * 0.3	29/01/09	Added isA(String) method for efficiency with Jess.
 * 0.4	03/02/09	Added filtering option to getInstances and getSortedInstances.
 * 		05/02/09	Added toString, hashCode and equals methods.
 * 0.5	10/02/09	Externalised strings.
 * 0.10	12/06/09	Fixed major problem with sorted instances. Did use TreeMap which 
 * 					mean is two values are the same only one is recorded.
 * 0.11	23/06/09	Rewrote the getSortedInstances routine again
 */

/*
 * Copyright (c) 2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados.owl2;

import java.util.Collection;
import java.util.HashSet;
import java.util.TreeMap;

import org.semanticweb.owl.model.OWLClass;

import uk.ac.abdn.carados.CaradosMessages;
import uk.ac.abdn.carados.OntologyClassHandler;
import uk.ac.abdn.carados.OntologyHandler;
import uk.ac.abdn.carados.OntologyInstanceFilter;
import uk.ac.abdn.carados.OntologyInstanceHandler;
import uk.ac.abdn.carados.util.OntologySortTree;

/**
 * <p>An example of the OntologyClassHandler, representing a class in the
 * Owl2CaradosOntology. It essentially wraps around an OWLClass, thus 
 * requiring the OWL2.0 library.
 * </p>
 * <h4>Release Notes</h4>
 * <p>Version 0.5 - Externalised strings.</p>
 * <p>
 * Version 0.4 - Added a filtering option to the getInstances and getSortedInstances
 * methods. Only the instances meeting the conditions of the filter are added to the 
 * retrieved collection. Added the <code>toString</code>, <code>hashCode</code> and 
 * <code>equals</code> methods.
 * </p>
 * <p>
 * Version 0.3 - Added <em>isA(String)</em> method to the OntologyEntityHandler. 
 * This is then applied to ClassHandlers and InstanceHandlers.
 * </p>
 * <p>
 * Version 0.2 - there is now a check for <code>null</code> class in the constructor.
 * In such a case, no class is stored for the instance. Also fixed a bug with the 
 * <code>createInstance(String)</code> method. Corrected class description in the 
 * JavaDoc. Moved the instance counter from the class to the ontology handler.
 * </p>
 * @author A. Gatt & D. Westwater, University of Aberdeen 
 */
public class Owl2CaradosClass implements OntologyClassHandler, Comparable<Object> {
	
	/** The name of this class. */
	private String _name;
	
	/** The reference of the ontology that created this class. */
	private Owl2CaradosOntology _ontology;
	
	/** The underlying OWLClass that this handler refers to. */
	private OWLClass _owlClass;
	
	/**
	 * Constructs a new ontology class handler.
	 * @param owlClass the underlying owlClass.
	 * @param ontology the ontology this class belongs to.
	 */
	Owl2CaradosClass(OWLClass owlClass, Owl2CaradosOntology ontology)
	{
		if (owlClass != null) {
			this._name = owlClass.getURI().getFragment();
		}
		this._owlClass = owlClass;
		this._ontology = ontology;
	}
	
	@Override
	final public void addDirectSuperClass(OntologyClassHandler superClass) {
		if (superClass instanceof Owl2CaradosClass && this._ontology != null)
		{
			this._ontology.addSuperClass(this, (Owl2CaradosClass)superClass, false);
		}
	}
	
	/**
	 * Retrieves the underlying OWLClass for this handler.
	 * @return the underlying OWLClass.
	 */
	final OWLClass getOntologyClass()
	{
		return this._owlClass;
	}

	@Override
	final public void addSubClass(OntologyClassHandler subClass) {
		if (subClass instanceof Owl2CaradosClass && this._ontology != null)
		{
			this._ontology.addSubClass((Owl2CaradosClass)subClass, this);
		}
	}

	@Override
	final public Collection<OntologyInstanceHandler> getInstances
	(boolean transitive)
	{
		return getInstances(null, transitive);
	}

	@Override
	final public Collection<OntologyInstanceHandler> getInstances
	(OntologyInstanceFilter filter, boolean transitive)
	{
		HashSet<OntologyInstanceHandler> allInstances = 
			new HashSet<OntologyInstanceHandler>();
		
		if (this._ontology != null && this._owlClass != null) {
			this._ontology.getInstances(this._owlClass, filter, transitive, allInstances);
		}
		return allInstances;
	}

	@Override
	final public Collection<OntologyInstanceHandler> getSortedInstances(
			String propertyName, boolean transitive) {
		
		return getSortedInstances(null, propertyName, transitive);
	}

	@Override
	final public Collection<OntologyInstanceHandler> getSortedInstances(
			OntologyInstanceFilter filter, String propertyName, boolean transitive) {
		return getSortedInstances(filter, propertyName, transitive, OntologySortTree.ASCENDING_ORDER);
	}

	@Override
	final public Collection<OntologyInstanceHandler> getSortedInstances(
			OntologyInstanceFilter filter, String propertyName, boolean transitive, 
			boolean order) {
		
		Collection <OntologyInstanceHandler> allInstances = 
			getInstances(filter, transitive);
		
		if (allInstances != null && allInstances.size() > 0) {
			OntologySortTree tree = new OntologySortTree(propertyName);
			allInstances = tree.sort(allInstances, order);
		}
		return allInstances;
	}

	@Override
	final public Collection<OntologyClassHandler> 
	getSubClasses(boolean transitive)
	{
		Collection<OntologyClassHandler> subClasses = 
			new HashSet<OntologyClassHandler>();
		
		if (this._ontology != null && this._owlClass != null) {
			this._ontology.getSubClasses(this._owlClass, transitive, subClasses);
		}
		return subClasses;
	}

	@Override
	final public Collection<OntologyClassHandler> 
	getSuperClasses(boolean transitive) {
		Collection<OntologyClassHandler> superClasses = 
			new HashSet<OntologyClassHandler>();
		
		if (this._ontology != null && this._owlClass != null) {
			this._ontology.getSuperClasses(this._owlClass, transitive, superClasses);
		}
		return superClasses;
	}

	@Override
	final public boolean hasSubClass(OntologyClassHandler className,
			boolean transitive) {
		
		return className != null ? className.hasSuperClass(this, transitive)
				: false;
	}

	@Override
	final public boolean hasSuperClass(OntologyClassHandler className,
			boolean transitive) {
		
		boolean hasSuperClass = false;
		
		if (this._ontology != null && className != null && 
				className instanceof Owl2CaradosClass && this._owlClass != null) {
			
			hasSuperClass = this._ontology.hasSuperClass(this._owlClass, 
						((Owl2CaradosClass)className).getOntologyClass(), 
						transitive);
		}
		return hasSuperClass;
	}

	@Override
	final public void setDefaultValue(String propertyName, Object value) {
		if (propertyName != null && this._owlClass != null && this._ontology != null)
		{
			this._ontology.setDefaultValue(this, propertyName, value);
		}
	}

	@Override
	final public void setDirectSuperClass(OntologyClassHandler superClass) {
		if (superClass != null && superClass instanceof Owl2CaradosClass && 
				this._ontology != null)
		{
			this._ontology.addSuperClass(this, (Owl2CaradosClass)superClass, true);
		}
	}

	@Override
	final public void delete() {
		if (this._ontology != null)
		{
			this._ontology.delete(this);
		}
	}

	@Override
	final public Object getDefaultValue(String propertyName, boolean transitive) {
		Object value = null;
		
		if (propertyName != null && this._owlClass != null && this._ontology != null)
		{
			value = this._ontology.getDefaultValue(this._owlClass, propertyName, 
					transitive);
		}
		return value;
	}

	@Override
	final public String getName() {
		return this._name;
	}

	@Override
	final public OntologyHandler getOntology() {
		return this._ontology;
	}

	@Override
	final public Collection<String> getPropertyNames() {
		return this._ontology != null && this._owlClass != null ? 
				this._ontology.getPropertyNames(this._owlClass) : 
					new HashSet<String>();
	}

	@Override
	final public boolean isA(OntologyClassHandler ontologyClass) {
		
		return (ontologyClass == null) ? false : 
			this.compareTo(ontologyClass) == 0 || 
			this.hasSuperClass(ontologyClass, true);
	}

	@Override
	final public boolean isA(String className) {
		OntologyClassHandler ontologyClass = null;
		if (className != null && this._name != null) {
			ontologyClass = 
				this._ontology.getOntologyClass(className);
		}
		
		return (ontologyClass == null) ? false : 
			this.compareTo(ontologyClass) == 0 || 
			this.hasSuperClass(ontologyClass, true);
	}

	@Override
	final public OntologyInstanceHandler createInstance() {
		return createInstance(null);
	}

	@Override
	final public OntologyInstanceHandler createInstance(String name) {
		OntologyInstanceHandler instance = this._ontology.createInstance(this, name);
		return instance;
	}

	@Override
	final public TreeMap<String, Object> getAllDefaultValues(boolean transitive) {
		return this._ontology != null && this._owlClass != null ? 
				this._ontology.getAllDefaultValues(this._owlClass, transitive) : 
					new TreeMap<String, Object>();
	}

	@Override
	final public int compareTo(Object o) {
		int comparison = 0;
		
		if (o instanceof Owl2CaradosClass)
		{
			comparison = this._name.compareTo(((Owl2CaradosClass)o).getName());
		}
		else
		{
			throw new ClassCastException(
					CaradosMessages.getString("Carados.NO_CLASS_CASTING") + this.getClass().getName() + ": " +  //$NON-NLS-1$ //$NON-NLS-2$
					o.getClass().getName());
		}
		return comparison;
	}

	@Override
	final public boolean hasProperty(String propertyName) {
		return this._ontology != null && this._owlClass != null && 
		this._ontology.hasProperty(this._owlClass, propertyName);
	}
	
	@Override
	final public int hashCode()
	{
		return this._name.hashCode();
	}

	@Override
	final public void removeDirectSuperClass(OntologyClassHandler superClass) {
		if (this._ontology != null && superClass != null && 
				superClass instanceof Owl2CaradosClass) {
			
			this._ontology.removeDirectSuperClass(this, (Owl2CaradosClass)superClass);
		}
	}

	@Override
	final public void removeSubClass(OntologyClassHandler subClass) {
		if (this._ontology != null && subClass != null && 
				subClass instanceof Owl2CaradosClass) {
			
			this._ontology.removeSubClass(this, (Owl2CaradosClass)subClass);
		}
	}
	
	@Override
	final public String toString() {
		return this._name;
	}

	@Override
	public boolean equals(Object obj) {
		return obj instanceof Owl2CaradosClass ? 
				this._name.equals(((Owl2CaradosClass)obj).getName()) : false;
	}
}
