/*
 * File: Owl2CaradosClass
 *
 * CHANGE HISTORY
 * 0.1	16/01/09	Base line file
 * 0.2	21/01/09	Corrected class description in JavaDoc.
 * 0.3	29/01/09	Added isA(String) method for efficiency with Jess.
 * 0.4	02/02/09	Changed getOwnValue so that it returns the default value if the instances
 * 					own value is not set.
 * 					Added new methods for doing the casting of property values locally.
 * 		05/02/09	Added toString, hashCode and equals methods.
 * 0.5	12/02/09	Fixed compiler warnings.
 * 					Added getOwnValueAsFloat().
 * 0.9	28/05/09	Added getOwnValueAsDouble().
 * 0.12	01/07/09	Fixed bug in getOwnValueAsDate() now that getOwnValue() returns a Date.
 * 2.2	09/02/10	Added a new getOwnValueAsCollection method.
 */

/*
 * Copyright (c) 2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados.owl2;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.TreeMap;

import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLIndividual;

import uk.ac.abdn.carados.OntologyClassHandler;
import uk.ac.abdn.carados.OntologyHandler;
import uk.ac.abdn.carados.OntologyInstanceHandler;

/**
 * <p>An example of the OntologyInstanceHandler, representing an instance in the
 * Owl2CaradosOntology. It essentially wraps around an OWLClass, thus 
 * requiring the OWL2.0 library.
 * </p>
 * <h4>Release Notes</h4>
 * <p>
 * Version 2.2 - Added getOwnValueAsCollection(String, Class&lt;T&gt;)
 * </p>
 * <p>
 * Version 0.9 - Added getOwnValueAsDouble()
 * </p>
 * <p>
 * Version 0.5 - added getOwnValueAsFloat().
 * </p>
 * <p>
 * Version 0.4 - getOwnValue now returns the default value if the instance does not have its
 * own value set. Added casting methods for retrieving a specific type of 
 * property value. Added the <code>toString</code>, <code>hashCode</code> and 
 * <code>equals</code> methods.
 * </p>
 * <p>
 * Version 0.3 - Added <em>isA(String)</em> method to the OntologyEntityHandler. 
 * This is then applied to ClassHandlers and InstanceHandlers.
 * </p>
 * <p>
 * Version 0.2 - Corrected class description in the JavaDoc. 
 * </p>
 * @author A. Gatt & D. Westwater, University of Aberdeen 
 */
public class Owl2CaradosInstance implements OntologyInstanceHandler {
	
	/** The name of this instance. */
	private String _name;
	
	/** The reference of the ontology that created this instance. */
	private Owl2CaradosOntology _ontology;
	
	/** The reference to the carados ontology class this instance belongs to. */
	private Owl2CaradosClass _parentCaradosClass;
	
	/** The underlying OWLIndividual. */
	private OWLIndividual _owlIndividual;

	/** Convenience field for Jess. */
	private String className;
	
	/** Convenience method for Jess. */
	public String getClassName() {
		return this.className;
	}
	
	/**
	 * Constructs a new ontology instance handler.
	 * @param owlIndividual the instance.
	 * @param parentClass the class this instance belongs to.
	 * @param ontology the ontology this instance belongs to.
	 */
	Owl2CaradosInstance(OWLIndividual owlIndividual, 
			OWLClass parentClass, Owl2CaradosOntology ontology)
	{
		this._name = owlIndividual.getURI().getFragment();
		this._ontology = ontology;
		this._owlIndividual = owlIndividual;
		this._parentCaradosClass = new Owl2CaradosClass(parentClass, ontology);
		this.className = this._parentCaradosClass.getName();
	}

	@Override
	final public void addOwnValue(String propertyName, Object value) {
		if (this._ontology != null && value != null)
		{
			this._ontology.setInstanceProperty(this, propertyName, value, false);
		}
	}

	@Override
	final public OntologyClassHandler getOntologyClass() {
		return this._parentCaradosClass;
	}

	@Override
	final public Object getOwnValue(String propertyName) {
		Object value = null;
		
		if (this._ontology != null && this._owlIndividual != null && propertyName != null) {
			value = this._ontology.getPropertyValue(this._owlIndividual, propertyName);
		}
		return value == null ? this.getDefaultValue(propertyName, true) : value;
	}
	
	@Override
	final public void setOwnValue(String propertyName, Object value) {
		if (this._ontology != null && propertyName != null)
		{
			this._ontology.setInstanceProperty(this, propertyName, value, true);
		}
	}

	@Override
	final public void delete() {
		if (this._ontology != null)
		{
			this._ontology.delete(this);
		}
	}

	@Override
	final public TreeMap<String, Object> getAllDefaultValues(boolean transitive) {
		return this._parentCaradosClass != null ? 
				this._parentCaradosClass.getAllDefaultValues(transitive) : 
					new TreeMap<String, Object>();
	}

	@Override
	final public Object getDefaultValue(String propertyName, boolean transitive) {
		return this._parentCaradosClass != null ? 
				this._parentCaradosClass.getDefaultValue(propertyName, transitive) : null;
	}

	@Override
	final public String getName() {
		return this._name;
	}

	@Override
	final public OntologyHandler getOntology() {
		return this._ontology;
	}

	@Override
	final public Collection<String> getPropertyNames() {
		return this._parentCaradosClass != null ? 
				this._parentCaradosClass.getPropertyNames() :  
					new HashSet<String>();
	}

	@Override
	final public boolean isA(OntologyClassHandler ontologyClass) {
		return this._parentCaradosClass != null ? 
			this._parentCaradosClass.isA(ontologyClass) : false;
	}

	@Override
	final public boolean isA(String name) {
		return name == null || 
		this._parentCaradosClass == null ? false : 
				this._parentCaradosClass.isA(name);
	}

	@Override
	final public boolean hasProperty(String propertyName) {
		return this._parentCaradosClass != null ? 
				this._parentCaradosClass.hasProperty(propertyName) : false;
	}

	/**
	 * Retrieves the underlying OWLIndividual that this handler wraps.
	 * @return the OWLIndividual.
	 */
	final OWLIndividual getOntologyIndividual() {
		return this._owlIndividual;
	}

	/**
	 * Sets the parent reference for this instance.
	 * @param owl2CaradosClass the parent class reference.
	 */
	final void setParentReference(Owl2CaradosClass owl2CaradosClass) {
		this._parentCaradosClass = owl2CaradosClass;
	}

	@Override
	public Object[] getOwnValueAsArray(String propertyName) {
		Object singleValue = this.getOwnValue(propertyName);
		Object[] arrayValues = null;
		
		if (singleValue != null)
		{
			if (singleValue instanceof Collection)
			{
				arrayValues = ((Collection<?>)singleValue).toArray();
			}
			else if (singleValue.getClass().isArray())
			{
				arrayValues = (Object[])singleValue;
			}
			else
			{
				arrayValues = new Object[]{singleValue};
			}
		}
		return arrayValues;
	}

	@SuppressWarnings("unchecked")
	@Override
	public Collection<Object> getOwnValueAsCollection(String propertyName) {
		Object singleValue = this.getOwnValue(propertyName);
		Collection<Object> setValues = null;
		
		if (singleValue != null)
		{
			if (singleValue instanceof Collection)
			{
				setValues = (Collection<Object>)singleValue;
			}
			else if (singleValue.getClass().isArray())
			{
				setValues = Arrays.asList((Object[])singleValue);
			}
			else
			{
				setValues = Collections.singleton(singleValue);
			}
		}
		return setValues;
	}

	@Override
	public Date getOwnValueAsDate(String propertyName) {
		Object value = getOwnValue(propertyName);
		Date dateValue = null;
		
		if (value != null) {
			if (value instanceof Date) {
				dateValue = (Date) value;
			}
			else {
				try {
					dateValue = this._ontology.translateToDate(value);
				}
				catch (Exception e) {
					// Do nothing
				}
			}
		}
		return dateValue;
	}

	/**
	 * Retrieves the default value as a Double. This implementation returns
	 * Float objects intact. Number objects are cast as Doubles using 
	 * <code>Number.doubleValue()</code> and Strings are parsed for Double values.
	 * @param propertyName the name of the property whose value is desired.
	 * @param transitive <code>true</code> if the super classes have to be 
	 * checked, <code>false</code> otherwise.
	 * @return a Double representation of the property value or <code>null</code>
	 * if the property does not exist or could not be converted to a Double.
	 */
	@Override
	public Double getOwnValueAsDouble(String propertyName) {
		Object value = getOwnValue(propertyName);
		Double doubleValue = null;
		
		if (value instanceof Double) {
			doubleValue = (Double)value;
		}
		else if (value instanceof Number) {
			doubleValue = new Double(((Number)value).doubleValue());
		}
		else if (value instanceof String) {
			try {
				doubleValue = Double.valueOf((String)value);
			} catch (NumberFormatException e) {
				doubleValue = null;
			}
		}
		return doubleValue;
	}

	/**
	 * Retrieves the default value as an Integer. This implementation returns
	 * Integer objects intact. Number objects are cast as Integers using 
	 * <code>Number.intValue()</code> and Strings are parsed for integer values.
	 * @param propertyName the name of the property whose value is desired.
	 * @return an Integer representation of the property value or 
	 * <code>null</code> if the property does not exist or could not be 
	 * converted to an Integer.
	 */
	@Override
	public Integer getOwnValueAsInteger(String propertyName) {
		Object value = getOwnValue(propertyName);
		Integer intValue = null;
		
		if (value instanceof Integer) {
			intValue = (Integer)value;
		}
		else if (value instanceof Number) {
			intValue = new Integer(((Number)value).intValue());
		}
		else if (value instanceof String) {
			try {
				intValue = Integer.valueOf((String)value);
			} catch (NumberFormatException e) {
				intValue = null;
			}
		}
		return intValue;
	}

	/**
	 * Retrieves the property value as a Float. This implementation returns
	 * Float objects intact. Number objects are cast as Floats using 
	 * <code>Number.floatValue()</code> and Strings are parsed for float values.
	 * @param propertyName the name of the property whose value is desired.
	 * @return a Float representation of the property value or <code>null</code>
	 * if the property does not exist or could not be converted to a Float.
	 */
	@Override
	public Float getOwnValueAsFloat(String propertyName) {
		Object value = getOwnValue(propertyName);
		Float floatValue = null;
		
		if (value instanceof Float) {
			floatValue = (Float)value;
		}
		else if (value instanceof Number) {
			floatValue = new Float(((Number)value).floatValue());
		}
		else if (value instanceof String) {
			try {
				floatValue = Float.valueOf((String)value);
			} catch (NumberFormatException e) {
				floatValue = null;
			}
		}
		return floatValue;
	}

	@Override
	public String getOwnValueAsString(String propertyName) {
		Object value = getOwnValue(propertyName);
		String stringValue = null;
		
		if (value != null) {
			if (value instanceof String) {
				stringValue = (String)value;
			}
			else {
				stringValue = value.toString();
			}
		}
		return stringValue;
	}
	
	@Override
	final public String toString() {
		return this._name;
	}
	
	@Override
	final public int hashCode()
	{
		return this._name.hashCode();
	}

	@Override
	public boolean equals(Object obj) {
		return obj instanceof Owl2CaradosInstance ? 
				this._name.equals(((Owl2CaradosInstance)obj).getName()) : false;
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T> Collection<T> getOwnValueAsCollection(String propertyName,
			Class<T> type) {

		Object singleValue = this.getOwnValue(propertyName);
		Collection<T> setValues = null;

		if (singleValue != null) {
			if (singleValue instanceof Collection<?>) {
				setValues = (Collection<T>) singleValue;
			} else if (singleValue.getClass().isArray()) {
				setValues = Arrays.asList((T[]) singleValue);
			} else {
				setValues = (Collection<T>) Collections
						.singleton((T) singleValue);
			}
		}
		return setValues == null ? new HashSet<T>() : setValues;
	}
}
