/*
 * File: Owl2CaradosOntology
 *
 * CHANGE HISTORY
 * 0.1		16/01/09	Base line file
 * 0.2		20/01/09	Fixed bug with casting in getAllDefaultValues(...).
 * 			21/01/09	Fixed bug with setting an instance's own value.
 * 						Corrected class descripton in JavaDoc.
 * 						The instance counter is moved to this file.
 * 			22/01/09	Incorrect default instance name being given, now in lower case.
 * 			29/01/09	Fixed bug with the hasSuperClass recursive call that took the base 
 * 					class again for searching and not the super class.
 * 0.4		03/02/09	Added a filtering option to the getInstances methods.
 * 			04/02/09	Fixed a bug with the recursive call on the getDefaultValues.
 * 0.5		09/02/09	Externalised all strings.
 * 			12/02/09	Fixed compiler warnings.
 * 0.8		27/02/09	Added a listener method for reporting missing properties.
 * 			02/03/09	Fixed bug with the missing properties reporting method.
 * 			05/03/09	Added new method for accepting a filter when grabbing all the 
 * 						instances in the ontology.
 * 0.10		19/06/09	Added basic type checking to data and object properties.
 * 0.12		30/06/09	Fixed bug with multiple object hasValues not working.
 * 			01/07/09	Fixed bug with the handling of dateTime objects in OWL.
 * 0.13		09/07/09	Removed debug code.
 * 						Fixed bug with class restriction on properties not being correctly
 * 						handled when no ranges exist.
 * 1.2		06/08/09	Fixed problem with getParameterNames() not checking for union of 
 *	 					classes in the domains.
 * 2.2		26/08/09	Added new methods for retrieving the model, factory and manager.
 * 2.2.1	15/09/09	Clean up warnings
 */

/*
 * Copyright (c) 2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados.owl2;

import java.net.URI;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import org.semanticweb.owl.apibinding.OWLManager;
import org.semanticweb.owl.model.AddAxiom;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLClassAssertionAxiom;
import org.semanticweb.owl.model.OWLConstant;
import org.semanticweb.owl.model.OWLDataFactory;
import org.semanticweb.owl.model.OWLDataOneOf;
import org.semanticweb.owl.model.OWLDataProperty;
import org.semanticweb.owl.model.OWLDataPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLDataPropertyExpression;
import org.semanticweb.owl.model.OWLDataRange;
import org.semanticweb.owl.model.OWLDataType;
import org.semanticweb.owl.model.OWLDataValueRestriction;
import org.semanticweb.owl.model.OWLDescription;
import org.semanticweb.owl.model.OWLIndividual;
import org.semanticweb.owl.model.OWLObject;
import org.semanticweb.owl.model.OWLObjectAllRestriction;
import org.semanticweb.owl.model.OWLObjectIntersectionOf;
import org.semanticweb.owl.model.OWLObjectProperty;
import org.semanticweb.owl.model.OWLObjectPropertyAssertionAxiom;
import org.semanticweb.owl.model.OWLObjectPropertyExpression;
import org.semanticweb.owl.model.OWLObjectUnionOf;
import org.semanticweb.owl.model.OWLObjectValueRestriction;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyChangeException;
import org.semanticweb.owl.model.OWLOntologyCreationException;
import org.semanticweb.owl.model.OWLOntologyManager;
import org.semanticweb.owl.model.OWLOntologyStorageException;
import org.semanticweb.owl.model.OWLProperty;
import org.semanticweb.owl.model.OWLRestriction;
import org.semanticweb.owl.model.OWLSubClassAxiom;
import org.semanticweb.owl.model.OWLTypedConstant;
import org.semanticweb.owl.model.RemoveAxiom;
import org.semanticweb.owl.model.UnknownOWLOntologyException;
import org.semanticweb.owl.util.OWLEntityRemover;

import uk.ac.abdn.carados.CaradosMessages;
import uk.ac.abdn.carados.OntologyClassHandler;
import uk.ac.abdn.carados.OntologyException;
import uk.ac.abdn.carados.OntologyHandler;
import uk.ac.abdn.carados.OntologyInstanceFilter;
import uk.ac.abdn.carados.OntologyInstanceHandler;
import uk.ac.abdn.carados.OntologyListener;

/**
 * <p>
 * This is an implementation of the Carados Ontology Package that wraps around an OWL2
 * based ontology. To run this version of Carados will require the OWL2.0 API. 
 * </p>
 * 
 * <h4>Release Notes</h4>
 * <p>Version 2.2 - Added new methods for retrieving the underlying OWLOntology, 
 * OWLOntologyManager and the OWLDataFactory.
 * </p>
 * <p>
 * Version 0.8 - Ontology will now report to listener any missing properties that are 
 * being set. Add a new method: <code>getAllInstances(OntologyInstanceFilter)</code>
 * </p>
 * <p>
 * Version 0.4 - Added a filtering option to the get instances. Only the instances
 * meeting the conditions of the filter are added to the retrieved collection.
 * </p>
 * <p>
 * Version 0.3 - Fixed a bug with the hasSuperClass call that entered an infinite loop.
 * </p>
 * <p>
 * Version 0.2 - Fixed two bugs: a casting bug in getDefaultObjectValue(...); and a 
 * problem with setting a value on an instance's property. Corrected class description
 * in JavaDoc. Moved the instance counter from the class to the ontology handler. Fixed 
 * bug with incorrect default name being created for new instances.
 * </p>
 * @author A. Gatt & D. Westwater, University of Aberdeen
 */
public class Owl2CaradosOntology implements OntologyHandler {
	/** The list of listeners for this handler. */
	private ArrayList<OntologyListener> _listeners = new ArrayList<OntologyListener>();

	/** The name of the ontology. */
	private String _name = ""; //$NON-NLS-1$

	/** The owl ontology manager, used for loading and saving */
	private OWLOntologyManager _manager;

	/** 
	 * The ontology data factory, used for accessing axioms, classes, 
	 * individuals, properties and so on.
	 */
	private OWLDataFactory _dataFactory;

	/** The owl ontology model */
	private OWLOntology _owlModel;

	/** The full URI of the ontology */
	private String _ontologyURI;

	/** The instance count for appending to the end of default names. */
	private long _instanceCount;

	private OWLDataType _owlInt = null;
	private OWLDataType _owlString = null;
	private OWLDataType _owlBoolean = null;
	private OWLDataType _owlFloat = null;
	private OWLDataType _owlDateTime = null;
	private static final SimpleDateFormat DATE_TIME_FORMAT = 
		new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss"); //$NON-NLS-1$

	/**
	 * Basic constructor for the Owl2CaradosOntology.
	 */
	public Owl2CaradosOntology() {
		this._manager = OWLManager.createOWLOntologyManager();
		this._dataFactory = this._manager.getOWLDataFactory();
		this._owlInt = 
			this._dataFactory.getOWLDataType(URI.create("http://www.w3.org/2001/XMLSchema#int")); //$NON-NLS-1$
		this._owlString = 
			this._dataFactory.getOWLDataType(URI.create("http://www.w3.org/2001/XMLSchema#string")); //$NON-NLS-1$
		this._owlFloat = 
			this._dataFactory.getOWLDataType(URI.create("http://www.w3.org/2001/XMLSchema#float")); //$NON-NLS-1$
		this._owlBoolean = 
			this._dataFactory.getOWLDataType(URI.create("http://www.w3.org/2001/XMLSchema#boolean")); //$NON-NLS-1$
		this._owlDateTime = 
			this._dataFactory.getOWLDataType(URI.create("http://www.w3.org/2001/XMLSchema#dateTime")); //$NON-NLS-1$
	}

	/**
	 * Constructs a URI for the given element name.
	 * @param name the element name.
	 * @return a URI representing this element for the given ontology.
	 */
	final private URI constructURI (String name)
	{
		StringBuilder uri = new StringBuilder();

		if (!name.startsWith(this._ontologyURI))
		{
			uri.append(this._ontologyURI);
		}
		uri.append(name);
		return URI.create(uri.toString());
	}

	@Override
	final public void loadOntology(URI uri) throws OntologyException {
		try
		{
			this._owlModel = this._manager.loadOntologyFromPhysicalURI(uri);
		}
		catch (Exception e)
		{
			try {
				this._owlModel = this._manager.createOntology(uri);
			} catch (OWLOntologyCreationException e1) {
				throw new OntologyException(CaradosMessages.getString("Carados.NO_ONTOLOGY_CREATED") +  //$NON-NLS-1$
						uri.toString());
			}
		}
		this._name = this._owlModel.getURI().toString();
		this._ontologyURI = this._name + '#';
	}

	@Override
	final public void attachListener(OntologyListener listener) {
		if (this._listeners != null && listener != null) {
			this._listeners.add(listener);
		}
	}

	@Override
	final public boolean containsClass(String className) {
		return className == null ? false :
			this._owlModel.containsClassReference(constructURI(className));
	}

	@Override
	final public boolean containsClass(OntologyClassHandler classHandler)
	{
		return classHandler == null ? false : 
			containsClass(classHandler.getName());
	}

	@Override
	final public boolean containsInstance(OntologyInstanceHandler instanceHandler)
	{
		return instanceHandler == null ? false : 
			this._owlModel.containsIndividualReference(constructURI(
					instanceHandler.getName()));
	}

	@Override
	final public boolean containsProperty(String propertyName) {
		URI uri = null;

		if (propertyName != null)
		{
			uri = constructURI(propertyName);
		}
		return uri == null ? false : 
			this._owlModel.containsDataPropertyReference(uri) ||
			this._owlModel.containsObjectPropertyReference(uri);
	}

	@Override
	final public OntologyClassHandler createClass(String className) {
		Owl2CaradosClass newClass = null;
		URI classURI = constructURI(className);

		if (className != null)
		{
			OWLClass owlClass = 
				this._dataFactory.getOWLClass(classURI);

			if (!this._owlModel.containsClassReference(classURI))
			{
				OWLAxiom axiom = this._dataFactory.getOWLDeclarationAxiom(owlClass);
				AddAxiom addAxiom = new AddAxiom(this._owlModel, axiom);
				try {
					this._manager.applyChange(addAxiom);
				} catch (OWLOntologyChangeException e) {
					owlClass = null;
				}
			}
			if (owlClass != null) {
				newClass = new Owl2CaradosClass(owlClass, this);
				if (this._listeners != null)
				{
					for (OntologyListener eachListener : this._listeners) {
						eachListener.classCreated(newClass);
					}
				}
			}
		}
		return newClass;
	}

	@Override
	final public Collection<OntologyInstanceHandler> getAllInstances() {
		return getAllInstances(null);
	}

	@Override
	final public Collection<OntologyInstanceHandler> 
	getAllInstances(OntologyInstanceFilter filter) {

		HashSet<OntologyInstanceHandler> allInstances = 
			new HashSet<OntologyInstanceHandler>();

		OntologyInstanceHandler instance = null;

		for (OWLIndividual owlIndividual : 
			this._owlModel.getReferencedIndividuals())
		{
			instance = new Owl2CaradosInstance(owlIndividual, 
					getIndividualClass(owlIndividual), this);

			if (filter == null || filter.canInclude(instance)) {
				allInstances.add(instance);
			}
		}
		return allInstances;
	}

	/**
	 * Retrieves the associated class of the given individual.
	 * @param individual the OWLIndividual.
	 * @return the appropriate OWLClass.
	 */
	private final OWLClass getIndividualClass(OWLIndividual individual)
	{
		OWLClass owlClass = null;
		Iterator<OWLDescription> eachDescription = 
			individual.getTypes(this._owlModel).iterator();

		boolean found = false;

		if (eachDescription != null)
		{
			OWLDescription desc = null;
			while (!found && eachDescription.hasNext())
			{
				desc = eachDescription.next();

				if (desc instanceof OWLClass)
				{
					found = true;
					owlClass = desc.asOWLClass();
				}
			}
		}
		return owlClass;
	}

	@Override
	final public Collection<OntologyClassHandler> getAllOntologyClasses() {
		HashSet<OntologyClassHandler> allClasses = 
			new HashSet<OntologyClassHandler>();

		for (OWLClass owlClass : this._owlModel.getReferencedClasses())
		{
			allClasses.add(new Owl2CaradosClass(owlClass, this));
		}
		return allClasses;
	}

	@Override
	final public OntologyClassHandler getOntologyClass(String className) {
		OntologyClassHandler ontologyClass = null;

		if (className != null)
		{
			if (this.containsClass(className))
			{
				ontologyClass = new Owl2CaradosClass(
						this._dataFactory.getOWLClass(
								constructURI(className)), this);
			}
		}
		return ontologyClass;
	}

	@Override
	final public Collection<String> getPropertyNames() {
		Collection<String> names = this.getDataPropertyNames();

		if (names == null)
		{
			names = this.getObjectPropertyNames();
		}
		else
		{
			names.addAll(this.getObjectPropertyNames());
		}
		return names;
	}

	/**
	 * Retrieves all the names of the data properties in the ontology
	 * @return a collection of strings representing the property names
	 */
	final private Collection<String> getDataPropertyNames()
	{
		Collection<String> allProperties = new HashSet<String>();

		for (OWLDataProperty owlProperty : 
			this._owlModel.getReferencedDataProperties())
		{
			allProperties.add(owlProperty.toString());
		}
		return allProperties;
	}

	/**
	 * Retrieves all the names of the object properties in the ontology
	 * @return a collection of strings representing the property names
	 */
	final private Collection<String> getObjectPropertyNames()
	{
		Collection<String> allProperties = new HashSet<String>();

		for (OWLObjectProperty owlProperty : 
			this._owlModel.getReferencedObjectProperties())
		{
			allProperties.add(owlProperty.toString());
		}
		return allProperties;
	}

	@Override
	final public OntologyClassHandler getRootClass() {
		return new Owl2CaradosClass(this._dataFactory.getOWLThing(), this);
	}

	@Override
	final public void removeAllListeners() {
		this._listeners.clear();
	}

	@Override
	final public void removeListener(OntologyListener listener) {
		this._listeners.remove(listener);
	}

	@Override
	final public void saveOntology(String fileName) throws OntologyException {
		try {
			this._manager.saveOntology(this._owlModel, URI.create(fileName));
		} catch (UnknownOWLOntologyException e) {
			throw new OntologyException(CaradosMessages.getString("Carados.NO_ONTOLOGY_SAVING") +   //$NON-NLS-1$
					e.getMessage());

		} catch (OWLOntologyStorageException e) {
			throw new OntologyException(CaradosMessages.getString("Carados.NO_ONTOLOGY_SAVING") +   //$NON-NLS-1$
					e.getMessage());
		}
	}

	@Override
	public Object translateDate(Date date) {
		return DATE_TIME_FORMAT.format(date);
	}

	@Override
	public Object translateDate(long timestamp) {
		return DATE_TIME_FORMAT.format(new Date(timestamp));
	}

	@Override
	public Date translateToDate(Object kbObject) {
		Date date = null;
		if (kbObject instanceof Date) {
			date = (Date)kbObject;
		}
		else if (kbObject instanceof String) {
			try {
				date = DATE_TIME_FORMAT.parse((String)kbObject);
			} catch (ParseException e) {
				try {
					date = new Date((Long.parseLong((String)kbObject)));
				}
				catch (NumberFormatException nfe) {
					// Do nothing
				}
			}
		}
		return date;
	}

	@Override
	public long translateToTimestamp(Object kbObject) {
		Date date = translateToDate(kbObject);
		return date == null ? 0 : date.getTime();
	}

	@Override
	final public String toString()
	{
		return this._name;
	}

	/**
	 * Adds a hierarchical relationship between the two classes.
	 * @param subCaradosClass the sub (or child) class.
	 * @param superCaradosClass the super (or parent) class.
	 */
	void addSubClass(Owl2CaradosClass subCaradosClass, Owl2CaradosClass superCaradosClass)
	{
		OWLClass subClass = subCaradosClass.getOntologyClass();
		OWLClass superClass = superCaradosClass.getOntologyClass();
		if (subClass != null && superClass != null)
		{
			OWLAxiom axiom = 
				this._dataFactory.getOWLSubClassAxiom(subClass, superClass);

			AddAxiom addAxiom = new AddAxiom(this._owlModel, axiom);
			try {
				this._manager.applyChange(addAxiom);

				if (this._listeners != null)
				{
					for (OntologyListener eachListener : this._listeners) {
						eachListener.subClassAdded(superCaradosClass, subCaradosClass);
					}
				}
			} catch (OWLOntologyChangeException e) {
				// Do nothing
			}
		}
	}

	/**
	 * Adds a hierarchical relationship between the two classes.
	 * @param subCaradosClass the sub (or child) class.
	 * @param superCaradosClass the super (or parent) class.
	 * @param overwrite <code>true</code> if the new super class is to be the only
	 * super class of the subclass, <code>false</code> otherwise.
	 */
	void addSuperClass(Owl2CaradosClass subCaradosClass, 
			Owl2CaradosClass superCaradosClass, boolean overwrite) {

		OWLClass subClass = subCaradosClass.getOntologyClass();
		OWLClass superClass = superCaradosClass.getOntologyClass();

		if (subClass != null && superClass != null && this._owlModel != null)
		{
			if (overwrite)
			{
				List<RemoveAxiom> removeAll = new ArrayList<RemoveAxiom>();

				Collection<OWLDescription> allSuperClasses = 
					subClass.getSuperClasses(this._owlModel);

				for (OWLDescription eachDescription : allSuperClasses)
				{
					if (eachDescription instanceof OWLClass)
					{
						removeAll.add(new RemoveAxiom(this._owlModel, 
								this._dataFactory.getOWLSubClassAxiom(subClass, eachDescription)));
					}
				}
				try {
					this._manager.applyChanges(removeAll);
				} catch (OWLOntologyChangeException e) {
					//Do nothing
				}
			}
			AddAxiom addAxiom = new AddAxiom(this._owlModel, 
					this._dataFactory.getOWLSubClassAxiom(subClass, superClass));
			try {
				this._manager.applyChange(addAxiom);

				if (this._listeners != null)
				{
					for (OntologyListener eachListener : this._listeners) {
						eachListener.superClassAdded(subCaradosClass, 
								superCaradosClass, false);
					}
				}
			} catch (OWLOntologyChangeException e) {
				// Do nothing
			}
		}
	}

	/**
	 * Retrieves all the instances for a given class.
	 * @param owlClass the class we want the instances for.
	 * @param filter the interface for filtering the instances retrieved. Only
	 * those instances matching the conditions of the filter are added. The 
	 * filter can be <code>null</code>, which has the effect of returning all 
	 * the instances.
	 * @param transitive <code>true</code> if the instances of the sub-classes
	 * have also to be retrieved, <code>false</code> otherwise.
	 * @param allInstances the collection to which the instances are added. 
	 */
	final void getInstances(OWLClass owlClass,
			OntologyInstanceFilter filter, boolean transitive, 
			Collection<OntologyInstanceHandler> allInstances) {

		if (allInstances != null) {
			OntologyInstanceHandler instance = null;
			for (OWLIndividual owlIndividual : 
				owlClass.getIndividuals(this._owlModel)) {

				instance = new Owl2CaradosInstance(owlIndividual, owlClass, this);
				if (filter == null || filter.canInclude(instance)) {
					allInstances.add(instance);
				}
			}
			if (transitive)
			{
				for (OWLDescription owlSubClass : 
					owlClass.getSubClasses(this._owlModel))
				{
					if (owlSubClass instanceof OWLClass)
					{
						this.getInstances(owlSubClass.asOWLClass(), filter, true, 
								allInstances);
					}
				}
			}
		}
	}

	/**
	 * Retrieves all the subclasses for the given class. If the call is 
	 * transitive, then it is recursively called to collect all known 
	 * direct and indirect sub classes.
	 * @param baseClass the base class to work from
	 * @param transitive determines if the full hierarchy should be examined
	 * @return a collection of BTOWLClasses
	 */
	final void getSubClasses(OWLClass baseClass, 
			boolean transitive, Collection<OntologyClassHandler> allClasses) {
		
		for (OWLDescription owlSubClass : baseClass.getSubClasses(this._owlModel))
		{
			if (owlSubClass instanceof OWLClass)
			{
				allClasses.add(new Owl2CaradosClass((OWLClass)owlSubClass, this));
				if (transitive)
				{
					getSubClasses((OWLClass)owlSubClass, transitive, allClasses);
				}
			}
		}
	}

	/**
	 * Retrieves all the super classes for the given class. If the call is 
	 * transitive, then it is recursively called to collect all known 
	 * direct and indirect super classes.
	 * @param baseClass the base class to work from
	 * @param transitive determines if the full hierarchy should be examined
	 * @return a collection of BTOWLClasses
	 */
	final void getSuperClasses(OWLClass baseClass, 
			boolean transitive, Collection<OntologyClassHandler> allClasses) {

		for (OWLDescription owlSuperClass : 
			baseClass.getSuperClasses(this._owlModel))
		{
			if (owlSuperClass instanceof OWLClass)
			{
				allClasses.add(new Owl2CaradosClass((OWLClass)owlSuperClass, this));
				if (transitive)
				{
					getSuperClasses((OWLClass)owlSuperClass, transitive, allClasses);
				}
			}
		}
	}

	/**
	 * Checks to see if there is a hiearchical link between the two classes.
	 * @param descendant the descendant class.
	 * @param ancestor the ancestor class.
	 * @param transitive <code>true</code> if the entire hierarchy is to be 
	 * searched, otherwise <code>false</code> if only the direct super classes
	 * have to be checked.
	 * @return <code>true</code> if the ancestor is a super class of 
	 * the descendant, <code>false</code> otherwise.
	 */
	final boolean hasSuperClass(OWLClass descendant, OWLClass ancestor, 
			boolean transitive)
	{
		boolean found = false;

		OWLSubClassAxiom subClassAxiom = 
			this._dataFactory.getOWLSubClassAxiom(descendant, ancestor);

		if (subClassAxiom != null)
		{
			found = this._owlModel.containsAxiom(subClassAxiom);

			if (!found && transitive)
			{
				Iterator<OWLDescription> eachSuperClass = 
					descendant.getSuperClasses(this._owlModel).iterator();

				OWLDescription currentDescription = null;

				while (!found && eachSuperClass.hasNext())
				{
					currentDescription = eachSuperClass.next();

					if (currentDescription instanceof OWLClass)
					{
						found = this.hasSuperClass(currentDescription.asOWLClass(), 
								ancestor, transitive);
					}
				}

			}
		}
		return found;
	}

	/**
	 * Sets a default value for a property on the given class. If the value is <code>null</code>, 
	 * this has the effect of removing the default value restriction.
	 * @param owlCaradosClass the class the property belongs to.
	 * @param propertyName the name of the property.
	 * @param value the new value of the property.
	 */
	final void setDefaultValue(Owl2CaradosClass owlCaradosClass, String propertyName,
			Object value) {

		boolean success = false;
		URI propertyURI = constructURI(propertyName);
		OWLClass owlClass = owlCaradosClass.getOntologyClass();

		if (this._owlModel.containsDataPropertyReference(propertyURI))
		{
			OWLDataProperty dataProperty = 
				this._dataFactory.getOWLDataProperty(propertyURI);

			if (dataProperty != null)
			{
				removeDefaultDataValue(owlClass, dataProperty);
				if (value != null)
				{
					success = setDefaultDataValue(owlClass, dataProperty, value);
				}
			}
		}
		else if (this._owlModel.containsObjectPropertyReference(propertyURI))
		{
			OWLObjectProperty objectProperty = 
				this._dataFactory.getOWLObjectProperty(propertyURI);

			if (objectProperty != null)
			{
				removeDefaultObjectValue(owlClass, objectProperty);
				if (value != null)
				{
					success = setDefaultObjectValue(owlClass, objectProperty, value);
				}
			}
		}
		if (success && this._listeners != null)
		{
			for (OntologyListener eachListener : this._listeners) {
				eachListener.defaultValueChanged(owlCaradosClass, 
						propertyName, value);
			}
		}
	}

	/**
	 * Removes the default data value on the property.
	 * @param owlClass the class the property belongs to.
	 * @param property the data property to be removed.
	 */
	private final void removeDefaultDataValue(OWLClass owlClass, OWLDataProperty property)
	{
		Collection<OWLDescription> allSubClasses = owlClass.getSubClasses(this._owlModel);
		boolean found = false;
		Iterator<OWLDescription> iterator = allSubClasses.iterator();
		OWLDescription currentDescription = null;

		while (!found && iterator.hasNext())
		{
			currentDescription = iterator.next();
			if (currentDescription instanceof OWLDataValueRestriction &&
					((OWLDataValueRestriction)currentDescription).getProperty().equals(property))
			{
				found = true;
				RemoveAxiom remove = new RemoveAxiom(this._owlModel, 
						this._dataFactory.getOWLSubClassAxiom(owlClass, 
								currentDescription));

				try {
					this._manager.applyChange(remove);
				} catch (OWLOntologyChangeException e) {
					// Do nothing
				}
			}
		}
	}

	/**
	 * Removes the default object value on the property.
	 * @param property the object property to be removed.
	 * @param owlClass the class the property belongs to.
	 */
	private final void removeDefaultObjectValue(OWLClass owlClass, OWLObjectProperty property)
	{
		Collection<OWLDescription> allSubClasses = owlClass.getSubClasses(this._owlModel);
		boolean found = false;
		Iterator<OWLDescription> iterator = allSubClasses.iterator();
		OWLDescription currentDescription = null;

		while (!found && iterator.hasNext())
		{
			currentDescription = iterator.next();
			if (currentDescription instanceof OWLObjectValueRestriction &&
					((OWLObjectValueRestriction)currentDescription).getProperty().equals(property))
			{
				found = true;
				RemoveAxiom remove = new RemoveAxiom(this._owlModel, 
						this._dataFactory.getOWLSubClassAxiom(owlClass, 
								currentDescription));

				try {
					this._manager.applyChange(remove);
				} catch (OWLOntologyChangeException e) {
					// Do nothing
				}
			}
		}
	}

	/**
	 * Sets the default value on a data property.
	 * @param owlClass the class the property belongs to.
	 * @param dataProperty the data property.
	 * @param value the new value of the property.
	 * @return <code>true</code> if the value was set correctly, <code>false</code> otherwise
	 */
	private final boolean setDefaultDataValue(OWLClass owlClass, OWLDataProperty dataProperty,
			Object value) {

		boolean success = false;

		if (dataProperty != null)
		{
			OWLDataValueRestriction dataRestrict = 
				this._dataFactory.getOWLDataValueRestriction(dataProperty, 
						parseObjectType(value));

			if (dataRestrict != null)
			{
				AddAxiom add = new AddAxiom(this._owlModel, 
						this._dataFactory.getOWLSubClassAxiom(
								owlClass, dataRestrict));

				try {
					this._manager.applyChange(add);
					success = true;
				} catch (OWLOntologyChangeException e) {
					// Do nothing
				}
			}
		}
		return success;
	}

	/**
	 * Sets the default value on a data property.
	 * @param owlClass the class the property belongs to.
	 * @param objectProperty the data property.
	 * @param value the new value of the property.
	 * @return <code>true</code> if the value was set correctly, <code>false</code> otherwise
	 */
	private final boolean setDefaultObjectValue(OWLClass owlClass, 
			OWLObjectProperty objectProperty, Object value) {

		boolean success = false;
		OWLObject target = null;

		if (value instanceof Owl2CaradosClass)
		{
			target = ((Owl2CaradosClass)value).getOntologyClass();
		}
		else if (value instanceof OWLClass)
		{
			target = ((OWLClass)value);
		}
		else if (value instanceof Owl2CaradosInstance)
		{
			target = ((Owl2CaradosInstance)value).getOntologyIndividual();
		}
		else if (value instanceof OWLIndividual)
		{
			target = ((OWLIndividual)value);
		}

		if (target != null)
		{
			OWLRestriction<?> objectRestrict = null;

			if (target instanceof OWLIndividual)
			{
				objectRestrict = 
					this._dataFactory.getOWLObjectValueRestriction(objectProperty, 
							(OWLIndividual)target);
			}
			else if (target instanceof OWLClass)
			{
				objectRestrict = this._dataFactory.getOWLObjectAllRestriction(
						objectProperty, (OWLDescription)target);
			}
			if (objectRestrict != null)
			{
				AddAxiom add = new AddAxiom(this._owlModel, 
						this._dataFactory.getOWLSubClassAxiom(
								owlClass, objectRestrict));

				try {
					this._manager.applyChange(add);
					success = true;
				} catch (OWLOntologyChangeException e) {
					// Do nothing
				}
			}
		}
		return success;
	}

	/**
	 * Parses an object to a particular OWLConstant
	 * @param value the object to be "cast"
	 * @return the OWLConstant
	 */
	final private OWLConstant parseObjectType(Object value) {
		OWLConstant constant = null;
		if (value instanceof Float)
		{
			constant = this._dataFactory.getOWLTypedConstant(
					((Float)value).floatValue());
		}
		else if (value instanceof Integer)
		{
			constant = this._dataFactory.getOWLTypedConstant(
					((Integer)value).intValue());
		}
		else if (value instanceof Long)
		{
			constant = this._dataFactory.getOWLTypedConstant(
					((Long)value).intValue());
		}
		else if (value instanceof String)
		{
			constant = this._dataFactory.getOWLTypedConstant(
					(String)value);
		}
		else if (value instanceof Double)
		{
			constant = this._dataFactory.getOWLTypedConstant(
					((Double)value).doubleValue());
		}
		else if (value instanceof Boolean)
		{
			constant = this._dataFactory.getOWLTypedConstant(
					((Boolean)value).booleanValue());
		}
		else if (value instanceof Date)
		{
			constant = this._dataFactory.getOWLTypedConstant(
					((Integer)(translateDate((Date) value))).intValue());
		}
		else
		{
			constant = this._dataFactory.getOWLUntypedConstant(value.toString());
		}
		return constant;
	}

	/**
	 * Deletes the class from the ontology.
	 * @param caradosClass the class to be deleted.
	 */
	final void delete(Owl2CaradosClass caradosClass) {

		OWLClass owlClass = caradosClass.getOntologyClass();

		OWLEntityRemover remover = 
			new OWLEntityRemover(this._manager, Collections.singleton(this._owlModel));

		remover.reset();
		remover.visit(owlClass);

		try {
			this._manager.applyChanges(remover.getChanges());
			if (this._listeners != null)
			{
				for (OntologyListener eachListener : this._listeners) {
					eachListener.classDeleted(caradosClass.getName());
				}
			}
		} catch (OWLOntologyChangeException e)
		{
			// Do nothing
		}
	}

	/**
	 * Deletes the instance from the ontology.
	 * @param caradosInstance the class to be deleted.
	 */
	final void delete(Owl2CaradosInstance caradosInstance) {

		OWLIndividual owlIndividual = caradosInstance.getOntologyIndividual();

		OWLEntityRemover remover = 
			new OWLEntityRemover(this._manager, Collections.singleton(this._owlModel));

		remover.reset();
		remover.visit(owlIndividual);

		try {
			this._manager.applyChanges(remover.getChanges());
			if (this._listeners != null)
			{
				for (OntologyListener eachListener : this._listeners) {
					eachListener.instanceDeleted(caradosInstance.getName());
				}
			}
		} catch (OWLOntologyChangeException e)
		{
			// Do nothing
		}
	}

	/**
	 * Retrieves the default value for the class.
	 * @param owlClass the class the property belongs to.
	 * @param propertyName the name of the property.
	 * @param transitive <code>true</code> if super classes have to be searched, 
	 * <code>false</code> otherwise.
	 * @return the default value as an object
	 */
	final Object getDefaultValue(OWLClass owlClass, String propertyName,
			boolean transitive) {

		Object value = null;
		URI propertyURI = constructURI(propertyName);

		if (this._owlModel.containsDataPropertyReference(propertyURI))
		{
			OWLDataProperty dataProperty = 
				this._dataFactory.getOWLDataProperty(propertyURI);

			if (dataProperty != null)
			{
				value = getDefaultDataValue(owlClass, dataProperty, transitive);
			}
		}
		else if (this._owlModel.containsObjectPropertyReference(propertyURI))
		{
			OWLObjectProperty objectProperty = 
				this._dataFactory.getOWLObjectProperty(propertyURI);

			if (objectProperty != null)
			{
				value = getDefaultObjectValue(owlClass, objectProperty, transitive);
			}
		}
		return value;
	}

	/**
	 * Retrieves the value of the 'hasValue' axiom for the specified property on the 
	 * specified class.
	 * @param owlClass the class the being searched.
	 * @param dataProperty the property being sought.
	 * @param transitive <code>true</code> if the super classes have to be searched, 
	 * <code>false</code> otherwise.
	 * @return the default value as an Object
	 */
	final Object getDefaultDataValue(OWLClass owlClass, 
			OWLDataProperty dataProperty, boolean transitive) {

		Object value = null;
		Iterator<OWLDescription> iterator = owlClass.getSuperClasses(this._owlModel).iterator();
		OWLDescription currentDescription = null;
		Collection<OWLClass> allSuperClasses = new HashSet<OWLClass>();

		while (value == null && iterator.hasNext())
		{
			// Check the current class for the property, this involves going through the 
			// "super classes" while we do this build a set of actual super classes.
			currentDescription = iterator.next();
			if (currentDescription instanceof OWLDataValueRestriction && 
					((OWLDataValueRestriction)currentDescription).getProperty()
					.equals(dataProperty))
			{
				value = parseDataConstant((((OWLDataValueRestriction)currentDescription).getValue()));
			}
			else if (currentDescription instanceof OWLClass)
			{
				allSuperClasses.add((OWLClass)currentDescription);
			}
		}
		if (value == null && transitive)
		{
			// The property is not in this class so check up the tree
			Iterator<OWLClass> classIterator = allSuperClasses.iterator();

			while (value == null && classIterator.hasNext())
			{
				value = getDefaultDataValue(classIterator.next(), dataProperty, transitive);
			}
		}
		return value;
	}

	/**
	 * Retrieves the value of the 'hasValue' or 'allValuesFrom' axiom for the specified 
	 * property on the specified class. The 'hasValue' will return instances while 
	 * the 'allValuesFrom' will return classes.
	 * @param owlClass the class the being searched.
	 * @param objectProperty the property being sought.
	 * @param transitive <code>true</code> if the super classes have to be searched, 
	 * <code>false</code> otherwise.
	 * @return the default value as an Object
	 */
	final Object getDefaultObjectValue(OWLClass owlClass, 
			OWLObjectProperty objectProperty, boolean transitive) {

		Object value = null;

		Iterator<OWLDescription> iterator = 
			owlClass.getSuperClasses(this._owlModel).iterator();

		Collection<OWLClass> allSuperClasses = new HashSet<OWLClass>();

		while (value == null && iterator.hasNext())
		{
			value = parseDefaultObjectValue(iterator.next(), allSuperClasses, objectProperty);
		}
		if (value == null && transitive)
		{
			// The property is not in this class so check up the tree
			Iterator<OWLClass> classIterator = allSuperClasses.iterator();

			while (value == null && classIterator.hasNext())
			{
				value = getDefaultObjectValue(classIterator.next(), objectProperty, transitive);
			}
		}
		return value;
	}

	/**
	 * Checks the specific types of the OWLDescriptions forming the 'super classes'. 
	 * @param currentDescription the OWLDescription currently being examined.
	 * @param allSuperClasses the <em>real</em> super classes, i.e. those that correspond to an 
	 * OWLClass.
	 * @param objectProperty the property being searched on.
	 * @return the value of the property for this description.
	 * @since 0.12
	 */
	private Object parseDefaultObjectValue(OWLDescription currentDescription, 
			Collection<OWLClass> allSuperClasses, OWLObjectProperty objectProperty) {
		Object value = null;

		if (currentDescription instanceof OWLObjectValueRestriction && 
				((OWLObjectValueRestriction)currentDescription).getProperty()
				.equals(objectProperty))
		{
			OWLIndividual owlIndividual = 
				((OWLObjectValueRestriction)currentDescription).getValue();

			if (owlIndividual != null)
			{
				value = new Owl2CaradosInstance(owlIndividual, 
						this.getIndividualClass(owlIndividual), this);
			}
		}
		else if (currentDescription instanceof OWLObjectAllRestriction &&
				((OWLObjectAllRestriction)currentDescription).getProperty()
				.equals(objectProperty))
		{
			OWLDescription owlDescription = 
				((OWLObjectAllRestriction)currentDescription).getFiller();

			if (owlDescription instanceof OWLClass)
			{
				value = new Owl2CaradosClass(owlDescription.asOWLClass(), this);
			}
			else if (owlDescription instanceof OWLObjectUnionOf)
			{
				Set<Owl2CaradosClass> set = new HashSet<Owl2CaradosClass>();
				for (OWLDescription eachDesc : 
					((OWLObjectUnionOf)owlDescription).getOperands())
				{
					if (eachDesc instanceof OWLClass)
					{
						set.add(new Owl2CaradosClass(eachDesc.asOWLClass(), this));
					}
				}
				value = set;
			}
			else if (owlDescription instanceof OWLObjectIntersectionOf)
			{
				Set<Owl2CaradosClass> set = new HashSet<Owl2CaradosClass>();
				for (OWLDescription eachDesc : 
					((OWLObjectIntersectionOf)owlDescription).getOperands())
				{
					if (eachDesc instanceof OWLClass)
					{
						set.add(new Owl2CaradosClass(eachDesc.asOWLClass(), this));
					}
				}
				value = set;
			}
		}
		else if (currentDescription instanceof OWLObjectIntersectionOf){
			Collection<Owl2CaradosInstance> instances = new HashSet<Owl2CaradosInstance>();
			Object subList = null;
			Collection<OWLDescription> descriptions = 
				((OWLObjectIntersectionOf)currentDescription).getOperands();

			for (OWLDescription eachDescription : descriptions) {
				subList = parseDefaultObjectValue(eachDescription, allSuperClasses, objectProperty);
				if (subList instanceof Owl2CaradosInstance) {
					instances.add((Owl2CaradosInstance)subList);
				}
				else if (subList instanceof Collection) {
					for (Object eachSubList : (Collection<?>)subList) {
						if (eachSubList instanceof Owl2CaradosInstance) {
							instances.add((Owl2CaradosInstance)eachSubList);
						}
					}
				}
			}
			value = instances.size() > 0 ? instances : null;
		}
		else if (currentDescription instanceof OWLClass)
		{
			allSuperClasses.add((OWLClass)currentDescription);
		}
		return value;
	}

	/**
	 * Parses an OWLConstant to a java type.
	 * @param constant the OWLConstant.
	 * @return the converted java object.
	 */
	private Object parseDataConstant(OWLConstant constant)
	{
		Object returnValue = null;

		if (constant.isTyped())
		{
			String dataType = constant.asOWLTypedConstant().getDataType().toString();
			String literal = constant.getLiteral();

			if (dataType.equalsIgnoreCase("string")) //$NON-NLS-1$
			{
				returnValue = literal;
			}
			else if (dataType.equalsIgnoreCase("integer") ||  //$NON-NLS-1$
					(dataType.equalsIgnoreCase("int"))) //$NON-NLS-1$
			{
				returnValue = new Integer(literal);
			}
			else if (dataType.equalsIgnoreCase("float")) //$NON-NLS-1$
			{
				returnValue = new Float(literal);
			}
			else if (dataType.equalsIgnoreCase("boolean")) //$NON-NLS-1$
			{
				returnValue = new Boolean(literal);
			}
			else if (dataType.equalsIgnoreCase("double")) //$NON-NLS-1$
			{
				returnValue = new Double(literal);
			}
			else if (dataType.equalsIgnoreCase("datetime")) //$NON-NLS-1$
			{
				returnValue = translateToDate(literal);
			}
		}
		return returnValue;
	}

	/**
	 * Retrieves the property names for the given class.
	 * @param owlClass the class to check
	 * @return a collection of Strings representing the property names
	 */
	final Collection<String> getPropertyNames(OWLClass owlClass) {
		Collection<String> allProperties = new HashSet<String>();
		Collection<OWLDescription> allDescriptions = null;

		for (OWLDataProperty eachDataProperty : 
			this._owlModel.getReferencedDataProperties())
		{
			allDescriptions = eachDataProperty.getDomains(this._owlModel);
			if (checkDescriptionsForClass(allDescriptions, owlClass)) {
				allProperties.add(eachDataProperty.getURI().getFragment());
			}
		}
		for (OWLObjectProperty eachObjectProperty : 
			this._owlModel.getReferencedObjectProperties())
		{
			allDescriptions = eachObjectProperty.getDomains(this._owlModel);
			if (checkDescriptionsForClass(allDescriptions, owlClass)) {
				allProperties.add(eachObjectProperty.getURI().getFragment());
			}
		}

		for (OWLDescription eachSuperClass : 
			owlClass.getSuperClasses(this._owlModel))
		{
			if (eachSuperClass instanceof OWLClass)
			{
				allProperties.addAll(getPropertyNames((OWLClass)eachSuperClass));
			}
		}
		return allProperties;
	}

	private boolean checkDescriptionsForClass(
			Collection<OWLDescription> allDescriptions, OWLClass owlClass) {

		boolean containsClass = false;

		if (allDescriptions != null) {
			if (allDescriptions.contains(owlClass)) {
				containsClass = true;
			}
			else {
				for (OWLDescription eachDescription : allDescriptions) {
					if (eachDescription instanceof OWLObjectUnionOf) {
						containsClass = checkDescriptionsForClass(
								((OWLObjectUnionOf)(eachDescription)).getOperands(), 
								owlClass);
					}
					else if (eachDescription instanceof OWLClass) {
						//ignore
					}
					else {
						int a = 0;
						a++;
					}
				}
			}
		}
		return containsClass;
	}

	/**
	 * Creates an instance of the specific class type and with the given name.
	 * @param caradosClass the class the instance is a kind of.
	 * @param instanceName the name of the instance. If the name is <code>null</code>
	 * then a default name is provided.
	 * @return the reference to the newly created instance or <code>null</code> 
	 * if there was a problem creating the instance.
	 */
	final OntologyInstanceHandler createInstance(Owl2CaradosClass caradosClass, 
			String instanceName) {

		Owl2CaradosInstance newInstance = null;
		OWLClass owlClass = caradosClass.getOntologyClass();
		String localInstanceName = instanceName;

		if (localInstanceName == null) {
			localInstanceName = 
				caradosClass.getName().toLowerCase() + '_' + this._instanceCount;
		}

		URI instanceURI = constructURI(localInstanceName);

		while (this._owlModel.containsIndividualReference(instanceURI))
		{
			instanceURI = constructURI(instanceURI.getFragment()+ "_0"); //$NON-NLS-1$
		}

		OWLIndividual individual = 
			this._dataFactory.getOWLIndividual(instanceURI);

		if (individual != null)
		{
			OWLClassAssertionAxiom axiom = 
				this._dataFactory.getOWLClassAssertionAxiom(individual, owlClass);

			try {
				this._manager.applyChange(new AddAxiom(this._owlModel, axiom));
				newInstance = new Owl2CaradosInstance(individual, owlClass, this);

				if (this._listeners != null)
				{
					for (OntologyListener eachListener : this._listeners) {
						eachListener.instanceCreated(newInstance);
					}
				}
			} catch (OWLOntologyChangeException e) {
				// Do nothing
			}
		}
		this._instanceCount++;
		return newInstance;
	}

	/**
	 * Retrieves all the default values for this class.
	 * @param owlClass the class being searched.
	 * @param transitive <code>true</code> if all the default values from super 
	 * classes have to be retrieved, <code>false</code> otherwise.
	 * @return a map of all the property name, value pairs.
	 */
	final TreeMap<String, Object> getAllDefaultValues(OWLClass owlClass, 
			boolean transitive) {

		TreeMap<String, Object> valueMap = 
			new TreeMap<String, Object>();

		Object value = null;
		OWLIndividual owlIndividual = null;
		OWLDescription owlDescription = null;
		Set<Owl2CaradosClass> set = null;
		String propertyName = null;
		Collection<OWLClass> superClasses = new HashSet<OWLClass>();

		for (OWLDescription eachSuperClass : owlClass.getSuperClasses(this._owlModel))
		{
			if (eachSuperClass instanceof OWLDataValueRestriction)
			{
				value = parseDataConstant((((OWLDataValueRestriction)eachSuperClass).getValue()));
				propertyName = ((OWLDataValueRestriction)eachSuperClass).getProperty()
				.asOWLDataProperty().getURI().getFragment();
			}
			else if (eachSuperClass instanceof OWLObjectValueRestriction)
			{
				owlIndividual = 
					((OWLObjectValueRestriction)eachSuperClass).getValue();

				if (owlIndividual == null)
				{
					value = new Owl2CaradosInstance(owlIndividual, owlClass, this);

					propertyName = ((OWLObjectValueRestriction)eachSuperClass).getProperty()
					.asOWLObjectProperty().getURI().getFragment();
				}
			}
			else if (eachSuperClass instanceof OWLObjectAllRestriction)
			{
				owlDescription = 
					((OWLObjectAllRestriction)eachSuperClass).getFiller();

				propertyName = ((OWLObjectAllRestriction)eachSuperClass).getProperty()
				.asOWLObjectProperty().getURI().getFragment();

				if (owlDescription instanceof OWLClass)
				{
					value = new Owl2CaradosClass(owlDescription.asOWLClass(), this);
				}
				else if (owlDescription instanceof OWLObjectUnionOf)
				{
					set = new HashSet<Owl2CaradosClass>();
					for (OWLDescription eachDesc : 
						((OWLObjectUnionOf)owlDescription).getOperands())
					{
						if (eachDesc instanceof OWLClass)
						{
							set.add(new Owl2CaradosClass(eachDesc.asOWLClass(), this));
						}
					}
					value = set;
				}
			}
			else if (eachSuperClass instanceof OWLClass)
			{
				value = null;
				superClasses.add(eachSuperClass.asOWLClass());
			}
			if (value != null)
			{
				valueMap.put(propertyName, value);
			}
		}
		if (transitive)
		{
			for (OWLClass eachSuperClass : superClasses)
			{
				valueMap.putAll(this.getAllDefaultValues(eachSuperClass, transitive));
			}
		}

		return valueMap;
	}

	/**
	 * Checks to see if the class has the named property.
	 * @param owlClass the class to be searched.
	 * @param propertyName the name of the property.
	 * @return <code>true</code> if the property exists on the class, <code>false</code> otherwise.
	 */
	final boolean hasProperty(OWLClass owlClass, String propertyName) {
		boolean found = false;

		URI propertyURI = constructURI(propertyName);

		OWLDataProperty dataProperty = 
			this._dataFactory.getOWLDataProperty(propertyURI);

		OWLObjectProperty objectProperty = 
			this._dataFactory.getOWLObjectProperty(propertyURI);

		if (this._owlModel.getReferencedDataProperties().contains(dataProperty))
		{
			found = recursiveHasProperty(owlClass, dataProperty);
		}
		else if (this._owlModel.getReferencedObjectProperties().contains(objectProperty))
		{
			found = recursiveHasProperty(owlClass, objectProperty);
		}

		return found;
	}

	/**
	 * This method performs a recursive search up the class hierarchy looking for a 
	 * particular property.
	 * @param owlClass the class to search.
	 * @param property the property being sought.
	 * @return <code>true</code> if the class (or a super class) contains the property,
	 * <code>false</code> otherwise.
	 */
	private final boolean recursiveHasProperty(OWLClass owlClass, OWLProperty<?,?> property) {
		boolean found = 
			property.getDomains(this._owlModel).contains(owlClass);

		if (!found)
		{
			Iterator<OWLDescription> iterator = 
				owlClass.getSuperClasses(this._owlModel).iterator();

			OWLDescription owlDescription = null;

			while (!found && iterator.hasNext())
			{
				owlDescription = iterator.next();
				if (owlDescription instanceof OWLClass)
				{
					found = recursiveHasProperty(owlDescription.asOWLClass(), property);
				}
			}
		}
		return found;
	}

	/**
	 * Removes the direct super class link between the two classes.
	 * @param subCaradosClass the sub (or child) class.
	 * @param superCaradosClass the super (or parent) class.
	 */
	final void removeDirectSuperClass(Owl2CaradosClass subCaradosClass,
			Owl2CaradosClass superCaradosClass) {

		OWLClass subClass = subCaradosClass.getOntologyClass();
		OWLClass superClass = superCaradosClass.getOntologyClass();

		RemoveAxiom remove = new RemoveAxiom(this._owlModel, 
				this._dataFactory.getOWLSubClassAxiom(subClass, superClass));
		try {
			this._manager.applyChange(remove);

			if (this._listeners != null)
			{
				for (OntologyListener eachListener : this._listeners) {
					eachListener.superClassRemoved(subCaradosClass, 
							superCaradosClass);
				}
			}
		} catch (OWLOntologyChangeException e) {
			// Do nothing
		}
	}

	/**
	 * Removes the direct sub class link between the two classes.
	 * @param superCaradosClass the super (or parent) class.
	 * @param subCaradosClass the sub (or child) class.
	 */
	final void removeSubClass(Owl2CaradosClass superCaradosClass,
			Owl2CaradosClass subCaradosClass) {

		OWLClass subClass = subCaradosClass.getOntologyClass();
		OWLClass superClass = superCaradosClass.getOntologyClass();

		RemoveAxiom remove = new RemoveAxiom(this._owlModel, 
				this._dataFactory.getOWLSubClassAxiom(subClass, superClass));
		try {
			this._manager.applyChange(remove);

			if (this._listeners != null)
			{
				for (OntologyListener eachListener : this._listeners) {
					eachListener.subClassRemoved(superCaradosClass, 
							subCaradosClass);
				}
			}
		} catch (OWLOntologyChangeException e) {
			// Do nothing
		}
	}

	/**
	 * Sets the property value for a particular instance. This may optional overwrite
	 * any previous value assigned to this property for this instance. Functional 
	 * properties will automatically overwrite the existing value.
	 * @param owl2CaradosInstance the reference to the instance.
	 * @param propertyName the name of the property.
	 * @param value the new value to be set or added, setting a the value to 
	 * <code>null</code> removes the property axiom.
	 * @param overwrite <code>true</code> if this new value is to overwrite all
	 * existing values or <code>false</code> if the value is to be added to a 
	 * list.
	 */
	final void setInstanceProperty(Owl2CaradosInstance owl2CaradosInstance,
			String propertyName, Object value, boolean overwrite) {

		URI propertyURI = constructURI(propertyName);

		OWLDataProperty owlDataProperty = 
			this._dataFactory.getOWLDataProperty(propertyURI);

		OWLObjectProperty owlObjectProperty = 
			this._dataFactory.getOWLObjectProperty(propertyURI);

		boolean success = false;
		boolean found = false;

		if (this._owlModel.getReferencedDataProperties().contains(owlDataProperty)) {
			found = true;
			success = setDataProperty(owl2CaradosInstance.getOntologyIndividual(), 
					propertyURI, value, overwrite);
		}
		else if (this._owlModel.getReferencedObjectProperties().contains(owlObjectProperty)) {
			found = true;
			success = setObjectProperty(owl2CaradosInstance.getOntologyIndividual(), 
					propertyURI, value, overwrite);
		}
		if (this._listeners != null)
		{
			if (success)
			{
				for (OntologyListener eachListener : this._listeners) {
					eachListener.propertyValueChanged(owl2CaradosInstance, propertyName, 
							value, overwrite);
				}
			}
			else if (!found) {
				for (OntologyListener eachListener : this._listeners) {
					eachListener.missingProperty(propertyName);
				}
			}
		}
	}

	/**
	 * Sets a data property for the specified individual, property and value. There is 
	 * an optional flag for determining whether existing values should be overwritten.
	 * @param owlIndividual the individual.
	 * @param propertyURI the property URI.
	 * @param value the new value.
	 * @param overwrite <code>true</code> if the value is to be overwritten, 
	 * <code>false</code> otherwise.
	 * @return <code>true</code> if the change was successfully made, <code>false<code> 
	 * otherwise.
	 */
	private final boolean setDataProperty(OWLIndividual owlIndividual, URI propertyURI, 
			Object value, boolean overwrite) {

		boolean success = false;

		OWLDataProperty owlProperty = 
			this._dataFactory.getOWLDataProperty(propertyURI);

		if (owlProperty != null)
		{
			if (overwrite || value == null || 
					owlProperty.isFunctional(this._owlModel))
			{
				// Remove the old assertions
				removeDataPropertyAssertion(owlIndividual, owlProperty);
			}

			if (value != null)
			{
				ArrayList<AddAxiom> assertAxioms = new ArrayList<AddAxiom>();
				if (value.getClass().isArray())
				{
					OWLAxiom axiom = null;
					for (Object eachValue : (Object[]) value)
					{
						if (eachValue != null)
						{
							axiom = createDataPropertyAssertion(owlIndividual, owlProperty, eachValue);
							if (axiom != null) {
								assertAxioms.add(new AddAxiom(this._owlModel, axiom));
							}
						}
					}
				}
				else if (value instanceof Collection)
				{
					OWLAxiom axiom = null;
					for (Object eachValue : (Collection<?>) value)
					{
						if (eachValue != null)
						{
							axiom = createDataPropertyAssertion(owlIndividual, owlProperty, eachValue);
							if (axiom != null) {
								assertAxioms.add(new AddAxiom(this._owlModel, axiom));
							}
						}
					}
				}
				else
				{
					OWLAxiom axiom = createDataPropertyAssertion(owlIndividual, owlProperty, value);
					if (axiom != null) {
						assertAxioms.add(new AddAxiom(this._owlModel, axiom));
					}
				}
				try {
					this._manager.applyChanges(assertAxioms);
					success = true;
				} catch (OWLOntologyChangeException e) {
					// Do nothing
				}
			}
		}
		return success;
	}

	/**
	 * Removes all data property assertions on the individual. I.e. removes
	 * existing values for a particular property.
	 * @param owlIndividual the individual whose property is to be removed.
	 * @param owlProperty the data property.
	 */
	final private void removeDataPropertyAssertion(OWLIndividual owlIndividual,
			OWLDataProperty owlProperty) {

		ArrayList<RemoveAxiom> removeList = new ArrayList<RemoveAxiom>();

		for (OWLDataPropertyAssertionAxiom assertAxiom : 
			this._owlModel.getDataPropertyAssertionAxioms(owlIndividual))
		{
			if (assertAxiom.getProperty().equals(owlProperty))
			{
				removeList.add(new RemoveAxiom(this._owlModel, assertAxiom));
			}
		}
		try {
			this._manager.applyChanges(removeList);
		} catch (OWLOntologyChangeException e) {
			// Do nothing
		}
	}

	/**
	 * Creates a new axiom linking the individual, the property and the particular value.
	 * @param individual the owl individual.
	 * @param owlProperty the property to be set.
	 * @param value the new value.
	 * @return an axiom linking these facts together.
	 */
	final private OWLAxiom createDataPropertyAssertion(OWLIndividual individual,
			OWLDataProperty owlProperty, Object value) {

		OWLDataPropertyAssertionAxiom assertAxiom = null;
		if (value instanceof Float)
		{
			assertAxiom = parseFloatValue(individual, owlProperty, (Float)value);
		}
		else if (value instanceof Integer)
		{
			assertAxiom = parseIntegerValue(individual, owlProperty, (Integer)value);
		}
		else if (value instanceof Long)
		{
			assertAxiom = parseLongValue(individual, owlProperty, (Long)value);
		}
		else if (value instanceof Double)
		{
			assertAxiom = parseDoubleValue(individual, owlProperty, (Double)value);
		}
		else if (value instanceof String)
		{
			assertAxiom = parseStringValue(individual, owlProperty, (String)value);
		}
		else if (value instanceof Boolean)
		{
			assertAxiom = parseBooleanValue(individual, owlProperty, (Boolean)value);
		}
		else if (value instanceof Date)
		{
			assertAxiom = parseDateValue(individual, owlProperty, (Date)value);
		}
		if (assertAxiom == null && this._listeners != null) {
			for (OntologyListener listener : this._listeners) {
				listener.typeConflict(owlProperty.toString(), value);
			}
		}
		return assertAxiom;
	}

	/**
	 * Creates a property assertion between the individual, the property and the string value.
	 * The string is parsed to match the underlying property range restrictions.
	 * @param individual the individual whose property is to be set.
	 * @param owlProperty the property being set.
	 * @param value the new value of the property.
	 * @return the OWLDataPropertyAssertionAxiom.
	 * @since 0.10
	 */
	private OWLDataPropertyAssertionAxiom parseStringValue(OWLIndividual individual, 
			OWLDataProperty owlProperty, String value) {

		OWLDataPropertyAssertionAxiom assertAxiom = null;
		Collection<OWLDataRange> data = owlProperty.getRanges(this._owlModel);
		boolean stored = false;

		try {
			if (data == null || data.size() == 0 || data.contains(this._owlString)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value);
				stored = true;
			}
			if (!stored && data.contains(this._owlFloat)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, Float.parseFloat(value));
				stored = true;
			}
			if (!stored && data.contains(this._owlInt)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, Integer.parseInt(value));
				stored = true;
			}
			if (!stored && data.contains(this._owlBoolean)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, Boolean.parseBoolean(value));
				stored = true;
			}
			if (!stored && data.contains(this._owlDateTime)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, DATE_TIME_FORMAT.format(DATE_TIME_FORMAT.parse(value)));
				stored = true;
			}
			if (!stored) {
				Iterator<OWLDataRange> iterator = data.iterator();
				OWLDataRange nextRange = null;
				Collection<OWLConstant> oneOfValues = null;
				OWLTypedConstant valueConstant = 
					this._dataFactory.getOWLTypedConstant(value);

				while (!stored && iterator.hasNext()) {
					nextRange = iterator.next();
					if (nextRange instanceof OWLDataOneOf) {
						oneOfValues = ((OWLDataOneOf)(nextRange)).getValues();
						if (oneOfValues.contains(valueConstant)) {
							assertAxiom = 
								this._dataFactory.getOWLDataPropertyAssertionAxiom(
										individual, owlProperty, valueConstant);
							stored = true;
						}
					}
				}
			}
		}
		catch (Exception e) {
			assertAxiom = null;
		}
		return assertAxiom;
	}

	/**
	 * Creates a property assertion between the individual, the property and the float value.
	 * The float is parsed to match the underlying property range restrictions.
	 * @param individual the individual whose property is to be set.
	 * @param owlProperty the property being set.
	 * @param value the new value of the property.
	 * @return the OWLDataPropertyAssertionAxiom.
	 * @since 0.10
	 */
	private OWLDataPropertyAssertionAxiom parseFloatValue(OWLIndividual individual, 
			OWLDataProperty owlProperty, Float value) {

		OWLDataPropertyAssertionAxiom assertAxiom = null;
		Collection<OWLDataRange> data = owlProperty.getRanges(this._owlModel);
		boolean stored = false;

		try {
			if (data == null || data.size() == 0 || data.contains(this._owlFloat)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.floatValue());
				stored = true;
			}
			if (!stored && data.contains(this._owlString)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.toString());
				stored = true;
			}
			if (!stored && data.contains(this._owlInt)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.intValue());
				stored = true;
			}
			if (!stored && data.contains(this._owlDateTime)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, 
							DATE_TIME_FORMAT.format(new Date(value.longValue())));
				stored = true;
			}
		} catch (Exception e) {
			assertAxiom = null;
		}
		return assertAxiom;
	}

	/**
	 * Creates a property assertion between the individual, the property and the double value.
	 * The double is parsed to match the underlying property range restrictions.
	 * @param individual the individual whose property is to be set.
	 * @param owlProperty the property being set.
	 * @param value the new value of the property.
	 * @return the OWLDataPropertyAssertionAxiom.
	 * @since 0.10
	 */
	private OWLDataPropertyAssertionAxiom parseDoubleValue(OWLIndividual individual, 
			OWLDataProperty owlProperty, Double value) {

		OWLDataPropertyAssertionAxiom assertAxiom = null;
		Collection<OWLDataRange> data = owlProperty.getRanges(this._owlModel);
		boolean stored = false;

		try {
			if (data == null || data.size() == 0 || data.contains(this._owlFloat)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.doubleValue());
				stored = true;
			}
			if (!stored && data.contains(this._owlString)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.toString());
				stored = true;
			}
			if (!stored && data.contains(this._owlInt)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.longValue());
				stored = true;
			}
			if (!stored && data.contains(this._owlDateTime)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, 
							DATE_TIME_FORMAT.format(new Date(value.longValue())));
				stored = true;
			}
		} catch (Exception e) {
			assertAxiom = null;
		}
		return assertAxiom;
	}

	/**
	 * Creates a property assertion between the individual, the property and the integer value.
	 * The integer is parsed to match the underlying property range restrictions.
	 * @param individual the individual whose property is to be set.
	 * @param owlProperty the property being set.
	 * @param value the new value of the property.
	 * @return the OWLDataPropertyAssertionAxiom.
	 * @since 0.10
	 */
	private OWLDataPropertyAssertionAxiom parseIntegerValue(OWLIndividual individual, 
			OWLDataProperty owlProperty, Integer value) {

		OWLDataPropertyAssertionAxiom assertAxiom = null;
		Collection<OWLDataRange> data = owlProperty.getRanges(this._owlModel);
		boolean stored = false;

		try {
			if (data == null || data.size() == 0 || data.contains(this._owlInt)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.intValue());
				stored = true;
			}
			if (!stored && data.contains(this._owlString)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.toString());
				stored = true;
			}
			if (!stored && data.contains(this._owlFloat)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.floatValue());
				stored = true;
			}
			if (!stored && data.contains(this._owlDateTime)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, 
							DATE_TIME_FORMAT.format(new Date(value.longValue())));
				stored = true;
			}
		} catch (Exception e) {
			assertAxiom = null;
		}
		return assertAxiom;
	}

	/**
	 * Creates a property assertion between the individual, the property and the long value.
	 * The long is parsed to match the underlying property range restrictions.
	 * @param individual the individual whose property is to be set.
	 * @param owlProperty the property being set.
	 * @param value the new value of the property.
	 * @return the OWLDataPropertyAssertionAxiom.
	 * @since 0.10
	 */
	private OWLDataPropertyAssertionAxiom parseLongValue(OWLIndividual individual, 
			OWLDataProperty owlProperty, Long value) {

		OWLDataPropertyAssertionAxiom assertAxiom = null;
		Collection<OWLDataRange> data = owlProperty.getRanges(this._owlModel);
		boolean stored = false;

		try {
			if (data == null || data.size() == 0 || data.contains(this._owlInt)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.longValue());
				stored = true;
			}
			if (!stored && data.contains(this._owlString)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.toString());
				stored = true;
			}
			if (!stored && data.contains(this._owlFloat)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.doubleValue());
				stored = true;
			}
			if (!stored && data.contains(this._owlDateTime)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, 
							DATE_TIME_FORMAT.format(new Date(value.longValue())));
				stored = true;
			}
		} catch (Exception e) {
			assertAxiom = null;
		}
		return assertAxiom;
	}

	/**
	 * Creates a property assertion between the individual, the property and the boolean value.
	 * The boolean is parsed to match the underlying property range restrictions.
	 * @param individual the individual whose property is to be set.
	 * @param owlProperty the property being set.
	 * @param value the new value of the property.
	 * @return the OWLDataPropertyAssertionAxiom.
	 * @since 0.10
	 */
	private OWLDataPropertyAssertionAxiom parseBooleanValue(OWLIndividual individual, 
			OWLDataProperty owlProperty, Boolean value) {

		OWLDataPropertyAssertionAxiom assertAxiom = null;
		Collection<OWLDataRange> data = owlProperty.getRanges(this._owlModel);
		boolean stored = false;

		try {
			if (data == null || data.size() == 0 || data.contains(this._owlBoolean)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.booleanValue());
				stored = true;
			}
			if (!stored && data.contains(this._owlString)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.toString());
				stored = true;
			}
		} catch (Exception e) {
			assertAxiom = null;
		}
		return assertAxiom;
	}

	/**
	 * Creates a property assertion between the individual, the property and the date value.
	 * The date is parsed to match the underlying property range restrictions.
	 * @param individual the individual whose property is to be set.
	 * @param owlProperty the property being set.
	 * @param value the new value of the property.
	 * @return the OWLDataPropertyAssertionAxiom.
	 * @since 0.10
	 */
	private OWLDataPropertyAssertionAxiom parseDateValue(OWLIndividual individual, 
			OWLDataProperty owlProperty, Date value) {

		OWLDataPropertyAssertionAxiom assertAxiom = null;
		Collection<OWLDataRange> data = owlProperty.getRanges(this._owlModel);
		boolean stored = false;

		try {
			if (data == null || data.size() == 0 || data.contains(this._owlDateTime)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, this._dataFactory.getOWLTypedConstant(
									DATE_TIME_FORMAT.format(value), this._owlDateTime));
				stored = true;
			}
			if (!stored && data.contains(this._owlString)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.toString());
				stored = true;
			}
			if (!stored && data.contains(this._owlInt)) {
				assertAxiom = 
					this._dataFactory.getOWLDataPropertyAssertionAxiom(
							individual, owlProperty, value.getTime());
				stored = true;
			}
		} catch (Exception e) {
			assertAxiom = null;
		}
		return assertAxiom;
	}

	/**
	 * Sets a data property for the specified individual, property and value. There is 
	 * an optional flag for determining whether existing values should be overwritten.
	 * @param owlIndividual the individual.
	 * @param propertyURI the property URI.
	 * @param value the new value.
	 * @param overwrite <code>true</code> if the value is to be overwritten, 
	 * <code>false</code> otherwise.
	 * @return <code>true</code> if the change was successfully made, <code>false<code> 
	 * otherwise.
	 */
	private final boolean setObjectProperty(OWLIndividual owlIndividual, URI propertyURI, 
			Object value, boolean overwrite) {

		boolean success = false;

		OWLObjectProperty owlProperty = 
			this._dataFactory.getOWLObjectProperty(propertyURI);

		if (owlProperty != null)
		{
			if (overwrite || value == null || 
					owlProperty.isFunctional(this._owlModel))
			{
				// Remove the old assertions
				removeObjectPropertyAssertion(owlIndividual, owlProperty);
			}

			if (value != null)
			{
				ArrayList<AddAxiom> assertAxioms = new ArrayList<AddAxiom>();
				if (value.getClass().isArray())
				{
					for (Object eachValue : (Object[]) value)
					{
						if (eachValue != null)
						{
							OWLAxiom axiom = createObjectPropertyAssertion(owlIndividual, owlProperty, eachValue);
							if (axiom != null) {
								assertAxioms.add(new AddAxiom(this._owlModel, axiom));
							}
						}
					}
				}
				else if (value instanceof Collection)
				{
					for (Object eachValue : (Collection<?>) value)
					{
						if (eachValue != null)
						{
							OWLAxiom axiom = createObjectPropertyAssertion(owlIndividual, owlProperty, eachValue);
							if (axiom != null) {
								assertAxioms.add(new AddAxiom(this._owlModel, axiom));
							}
						}
					}
				}
				else
				{
					OWLAxiom axiom = createObjectPropertyAssertion(owlIndividual, owlProperty, value);
					if (axiom != null) {
						assertAxioms.add(new AddAxiom(this._owlModel, axiom));
					}
				}
				try {
					this._manager.applyChanges(assertAxioms);
					success = true;
				} catch (OWLOntologyChangeException e) {
					// Do nothing
				}
			}
		}
		return success;
	}

	/**
	 * Removes all object property assertions on the individual. I.e. removes
	 * existing values for a particular property.
	 * @param owlIndividual the individual whose property is to be removed.
	 * @param owlProperty the data property.
	 */
	final private void removeObjectPropertyAssertion(OWLIndividual owlIndividual,
			OWLObjectProperty owlProperty) {

		ArrayList<RemoveAxiom> removeList = new ArrayList<RemoveAxiom>();

		for (OWLObjectPropertyAssertionAxiom assertAxiom : 
			this._owlModel.getObjectPropertyAssertionAxioms(owlIndividual))
		{
			if (assertAxiom.getProperty().equals(owlProperty))
			{
				removeList.add(new RemoveAxiom(this._owlModel, assertAxiom));
			}
		}
		try {
			this._manager.applyChanges(removeList);
		} catch (OWLOntologyChangeException e) {
			// Do nothing
		}
	}

	/**
	 * Creates a new axiom linking the individual, the property and the particular value.
	 * @param individual the owl individual.
	 * @param owlProperty the property to be set.
	 * @param value the new value.
	 * @return an axiom linking these facts together.
	 */
	final private OWLAxiom createObjectPropertyAssertion(OWLIndividual individual,
			OWLObjectProperty owlProperty, Object value) {

		OWLObjectPropertyAssertionAxiom assertAxiom = null;
		OWLIndividual target = null;
		boolean found = false;

		// Add all the new assertions
		if (value instanceof String)
		{
			target = this._dataFactory.getOWLIndividual(
					constructURI((String)value));
		}
		else if (value instanceof Owl2CaradosInstance)
		{
			target = ((Owl2CaradosInstance)value).getOntologyIndividual();
		}
		else if (value instanceof OWLIndividual)
		{
			target = (OWLIndividual)value;
		}
		if (target != null)
		{
			Collection<OWLDescription> rangeDescriptions = 
				owlProperty.getRanges(this._owlModel);

			if (rangeDescriptions != null && rangeDescriptions.size() > 0) {
				found = validateClass(rangeDescriptions, target);
			}
			else {
				found = true;
			}
			if (found) {
				assertAxiom = this._dataFactory.getOWLObjectPropertyAssertionAxiom(
						individual, owlProperty, target);
				if (owlProperty.isSymmetric(this._owlModel))
				{
					assertAxiom =this._dataFactory.getOWLObjectPropertyAssertionAxiom(
							target, owlProperty, individual);
				}
			}

		}
		if (assertAxiom == null && this._listeners != null) {
			for (OntologyListener listener : this._listeners) {
				listener.typeConflict(owlProperty.toString(), value.toString());
			}
		}
		return assertAxiom;
	}

	/**
	 * Performs a test to see if the target's type can be found in the permitted 
	 * ranges of the object property.
	 * @param rangeDescriptions the collection of range descriptions.
	 * @param target the target to be checked.
	 * @return <code>true</code> if the target matches the range descriptions, 
	 * <code>false</code> otherwise.
	 */
	private boolean validateClass(Collection<OWLDescription> rangeDescriptions,
			OWLIndividual target) {

		boolean found = false;
		Iterator<OWLDescription> eachRange = rangeDescriptions.iterator();
		Collection<OWLDescription> types = target.getTypes(this._owlModel);
		Iterator<OWLDescription> eachType = null;
		OWLDescription typeClass = null;
		OWLDescription rangeClass = null;

		while (!found && eachRange.hasNext()) {
			eachType = types.iterator();
			rangeClass = eachRange.next();
			if (rangeClass instanceof OWLObjectUnionOf) {
				Collection<OWLDescription> unionRange = ((OWLObjectUnionOf)rangeClass).getOperands();
				found = validateClass(unionRange, target);
			}
			else {
				while (!found && eachType.hasNext()) {
					typeClass = eachType.next();
					if (rangeClass instanceof OWLClass && 
							typeClass instanceof OWLClass) {

						found = rangeClass.equals(typeClass) || this.hasSuperClass(typeClass.asOWLClass(), 
								rangeClass.asOWLClass(), true); 
					}
					else {
						int a = 1;
						a += 1;

					}
				}
			}
		}
		return found;
	}

	/**
	 * Retrieves the value of an instance's property.
	 * @param individual the instance the property belongs to.
	 * @param propertyName the name of the property.
	 * @return the property value as an object.
	 */
	final Object getPropertyValue(OWLIndividual individual, String propertyName) {

		Object value = null;
		URI propertyURI = constructURI(propertyName);

		if (this._owlModel.containsDataPropertyReference(propertyURI))
		{
			OWLDataProperty dataProperty = 
				this._dataFactory.getOWLDataProperty(propertyURI);

			if (dataProperty != null)
			{
				value = getPropertyDataValue(individual, dataProperty);
			}
		}
		else if (this._owlModel.containsObjectPropertyReference(propertyURI))
		{
			OWLObjectProperty objectProperty = 
				this._dataFactory.getOWLObjectProperty(propertyURI);

			if (objectProperty != null)
			{
				value = getPropertyObjectValue(individual, objectProperty);
			}
		}
		return value;
	}

	/**
	 * Retrieves the value of the instance's data property.
	 * @param individual the instance the being searched.
	 * @param dataProperty the property being sought.
	 * @return the property value as an Object
	 */
	final Object getPropertyDataValue(OWLIndividual individual, 
			OWLDataProperty dataProperty) {

		Object value = null;
		Map<OWLDataPropertyExpression, Set<OWLConstant>> dataMap = 
			individual.getDataPropertyValues(this._owlModel);

		if (dataMap != null && dataMap.containsKey(dataProperty))
		{
			Collection<OWLConstant> allConstants = dataMap.get(dataProperty);

			if (allConstants != null)
			{
				HashSet<Object> valueList = new HashSet<Object>();
				for (OWLConstant eachConstant : allConstants)
				{
					valueList.add(this.parseDataConstant(eachConstant));
				}
				if (valueList.size() == 1)
				{
					value = valueList.toArray()[0];
				}
				else if (valueList.size() > 1)
				{
					value = valueList;
				}
			}
		}

		return value;
	}

	/**
	 * Retrieves the value of the instance's object property.
	 * @param individual the class the being searched.
	 * @param objectProperty the property being sought.
	 * @return the property value as an Object
	 */
	final Object getPropertyObjectValue(OWLIndividual individual, 
			OWLObjectProperty objectProperty) {

		Object value = null;
		Map<OWLObjectPropertyExpression, Set<OWLIndividual>> dataMap = 
			individual.getObjectPropertyValues(this._owlModel);

		if (dataMap != null && dataMap.containsKey(objectProperty))
		{
			Collection<OWLIndividual> allIndividuals = dataMap.get(objectProperty);

			if (allIndividuals != null)
			{
				HashSet<Object> valueList = new HashSet<Object>();
				for (OWLIndividual eachIndividual : allIndividuals)
				{
					valueList.add(new 
							Owl2CaradosInstance(eachIndividual, 
									getIndividualClass(eachIndividual), 
									this));
				}
				if (valueList.size() == 1)
				{
					value = valueList.toArray()[0];
				}
				else if (valueList.size() > 1)
				{
					value = valueList;
				}
			}
		}
		return value;
	}
	
	/**
	 * Retrieves the owlModel that this class wraps.
	 * @return the OWLOntology
	 * @since 2.2
	 */
	public OWLOntology getOwlModel() {
		return this._owlModel;
	}
	
	/**
	 * Retrieves the ontology manager used by this ontology.
	 * @return the OWLOntologyManager
	 * @since 2.2
	 */
	public OWLOntologyManager getOwlManager() {
		return this._manager;
	}

	/**
	 * Retrieves the data factory used by this ontology.
	 * @return the OWLDataFactory.
	 * @since 2.2
	 */
	public OWLDataFactory getOwlDataFactory() {
		return this._dataFactory;
	}

	/**
	 * Retrieves the name of this ontology.
	 * @return a string representation of the name.
	 * @since 2.2
	 */
	public String getOntologyName() {
		return this._name;
	}
}
