/*
 * File: Owl2CaradosClass
 *
 * CHANGE HISTORY
 * 0.1	16/01/09	Base line file
 * 0.3	29/01/09	Added isA(String) method for efficiency with Jess.
 * 0.4	03/02/09	Added a filtering option to the getInstances and getSortedInstances.
 * 		05/02/09	Added toString, hashCode and equals methods.
 * 0.5	09/02/09	Externalised string.
 * 		12/02/09	Fixed compiler warnings.
 * 2.2	19/02/10	Fixed bug with the increment of the instance counter, no longer 
 * 					goes up by two.
 * 2.4	05/03/10	Added getInstance(String) that retrieves the named instance of the class.
 */

/*
 * Copyright (c) 2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados.simple;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.TreeMap;

import uk.ac.abdn.carados.CaradosMessages;
import uk.ac.abdn.carados.OntologyClassHandler;
import uk.ac.abdn.carados.OntologyHandler;
import uk.ac.abdn.carados.OntologyInstanceFilter;
import uk.ac.abdn.carados.OntologyInstanceHandler;
import uk.ac.abdn.carados.util.OntologySortTree;

/**
 * <p>
 * A basic example of the OntologyClassHandler, representing a class in the
 * Owl2CaradosOntology. It is intended for applications whose knowledge
 * representation requirements are relatively simple (e.g. do not require a
 * highly expressive representation language like OWL, and do not require
 * reasoning).
 * </p>
 * <h4>Release Notes</h4>
 * <p>
 * Version 0.5 - Externalised strings.
 * </p>
 * <p>
 * Version 0.4 - Added a filtering option to the getInstances and
 * getSortedInstances methods. Only the instances meeting the conditions of the
 * filter are added to the retrieved collection. Added the <code>toString</code>, <code>hashCode</code> and <code>equals</code> methods.
 * </p>
 * <p>
 * Version 0.3 - Added <em>isA(String)</em> method to the OntologyEntityHandler.
 * This is then applied to ClassHandlers and InstanceHandlers.
 * </p>
 * <p>
 * Version 2.4 - Added a new method to retrieve a named instance of this class.
 * </p>
 * 
 * @author A. Gatt & D. Westwater, University of Aberdeen
 */
public class SimpleCaradosClass implements OntologyClassHandler,
		Comparable<Object> {

	/** The name of this class. */
	private String _name;

	/** The reference of the ontology that created this class. */
	private SimpleCaradosOntology _ontology;

	/** The map containing all the instances in this ontology. */
	private TreeMap<String, OntologyInstanceHandler> _instances = new TreeMap<String, OntologyInstanceHandler>();

	/** The default properties for this class. */
	private TreeMap<String, Object> _properties = new TreeMap<String, Object>();

	/** The subclasses of this class. */
	private HashSet<OntologyClassHandler> _subClasses = new HashSet<OntologyClassHandler>();

	/** The super classes of this class. */
	private HashSet<OntologyClassHandler> _superClasses = new HashSet<OntologyClassHandler>();

	/**
	 * The number attached to each newly created instance that has a default
	 * name.
	 */
	private long instanceCount;

	/**
	 * Constructs a new ontology class handler.
	 * 
	 * @param name
	 *            the name of this class.
	 * @param ontology
	 *            the ontology this class belongs to.
	 */
	protected SimpleCaradosClass(String name, SimpleCaradosOntology ontology) {
		this._name = name;
		this._ontology = ontology;
	}

	@Override
	final public void addDirectSuperClass(OntologyClassHandler superClass) {
		if (superClass != null) {
			this._superClasses.add(superClass);
			((SimpleCaradosClass) superClass).nonRecursiveAddSubClass(this);
			this._ontology.addedSuperClass(this, superClass, false);
		}
	}

	/**
	 * Prevents recursion from the adding of subclasses.
	 * 
	 * @param superClass
	 *            the new super class.
	 */
	final private void nonRecursiveAddSuperClass(OntologyClassHandler superClass) {
		this._superClasses.add(superClass);
	}

	@Override
	final public void addSubClass(OntologyClassHandler subClass) {
		if (subClass != null) {
			this._subClasses.add(subClass);
			((SimpleCaradosClass) subClass).nonRecursiveAddSuperClass(this);
			this._ontology.addedSubClass(this, subClass);
		}
	}

	/**
	 * Prevents recursion from the adding of super classes.
	 * 
	 * @param subClass
	 *            the new subclass.
	 */
	final private void nonRecursiveAddSubClass(OntologyClassHandler subClass) {
		this._subClasses.add(subClass);
	}

	@Override
	final public Collection<OntologyInstanceHandler> getInstances(
			boolean transitive) {
		return getInstances(null, transitive);
	}

	@Override
	final public Collection<OntologyInstanceHandler> getInstances(
			OntologyInstanceFilter filter, boolean transitive) {
		HashSet<OntologyInstanceHandler> allInstances = new HashSet<OntologyInstanceHandler>();

		if (this._instances != null) {
			for (OntologyInstanceHandler eachInstance : this._instances
					.values()) {
				if (filter == null || filter.canInclude(eachInstance)) {
					allInstances.add(eachInstance);
				}
			}

			if (transitive) {
				Collection<OntologyClassHandler> classSubClasses = this
						.getSubClasses(true);

				if (classSubClasses != null) {
					for (OntologyClassHandler eachClass : classSubClasses) {
						allInstances.addAll(eachClass.getInstances(filter,
								false));
					}
				}
			}
		}
		return allInstances;
	}

	@Override
	final public Collection<OntologyInstanceHandler> getSortedInstances(
			String propertyName, boolean transitive) {

		return getSortedInstances(null, propertyName, transitive);
	}

	@Override
	final public Collection<OntologyInstanceHandler> getSortedInstances(
			OntologyInstanceFilter filter, String propertyName,
			boolean transitive) {

		return getSortedInstances(filter, propertyName, transitive,
				OntologySortTree.ASCENDING_ORDER);
	}

	@Override
	final public Collection<OntologyClassHandler> getSubClasses(
			boolean transitive) {
		HashSet<OntologyClassHandler> allClasses = new HashSet<OntologyClassHandler>();

		if (this._subClasses != null) {
			allClasses.addAll(this._subClasses);
			if (transitive) {
				for (OntologyClassHandler eachSubClass : this._subClasses) {
					allClasses.addAll(eachSubClass.getSubClasses(true));
				}
			}
		}

		return allClasses;
	}

	@Override
	final public Collection<OntologyClassHandler> getSuperClasses(
			boolean transitive) {
		HashSet<OntologyClassHandler> allClasses = new HashSet<OntologyClassHandler>();

		if (this._superClasses != null) {
			allClasses.addAll(this._superClasses);
			if (transitive) {
				for (OntologyClassHandler eachSuperClass : this._superClasses) {
					allClasses.addAll(eachSuperClass.getSuperClasses(true));
				}
			}
		}

		return allClasses;
	}

	@Override
	final public boolean hasSubClass(OntologyClassHandler className,
			boolean transitive) {

		return className != null ? className.hasSuperClass(this, transitive)
				: false;
	}

	@Override
	final public boolean hasSuperClass(OntologyClassHandler className,
			boolean transitive) {

		boolean found = false;

		if (className != null && this._superClasses != null) {
			found = this._superClasses.contains(className);

			if (transitive && !found) {
				Iterator<OntologyClassHandler> eachClass = this
						.getSuperClasses(true).iterator();

				while (!found && eachClass.hasNext()) {
					found = eachClass.next().hasSuperClass(className, false);
				}
			}
		}
		return found;
	}

	@Override
	final public void setDefaultValue(String propertyName, Object value) {
		if (this._properties != null) {
			this._properties.put(propertyName, value);
			this._ontology.changedDefaultValue(this, propertyName, value);
		}
	}

	@Override
	final public void setDirectSuperClass(OntologyClassHandler superClass) {
		if (superClass != null) {
			for (OntologyClassHandler eachSuperClass : this._superClasses) {
				eachSuperClass.removeSubClass(this);
			}
			this._superClasses.clear();
			this._superClasses.add(superClass);
			((SimpleCaradosClass) superClass).nonRecursiveAddSubClass(this);
			this._ontology.addedSuperClass(this, superClass, true);
		}
	}

	@Override
	final public void delete() {
		this._ontology.deleteClass(this);

		if (this._instances != null) {
			for (OntologyInstanceHandler eachInstance : this._instances
					.values()) {
				eachInstance.delete();
			}
		}
		if (this._subClasses != null) {
			for (OntologyClassHandler eachClass : this._subClasses) {
				eachClass.removeDirectSuperClass(this);
			}
		}
		if (this._superClasses != null) {
			for (OntologyClassHandler eachClass : this._superClasses) {
				eachClass.removeSubClass(this);
			}
		}
		this._instances = null;
		this._properties = null;
		this._ontology = null;
		this._subClasses = null;
		this._superClasses = null;
	}

	@Override
	final public Object getDefaultValue(String propertyName, boolean transitive) {
		Object value = null;

		if (this._properties != null) {
			value = this._properties.get(propertyName);
			if (value == null && transitive && this._superClasses != null) {
				// Check the immediate super classes first
				Iterator<OntologyClassHandler> iterator = this._superClasses
						.iterator();

				while (value == null && iterator.hasNext()) {
					value = iterator.next()
							.getDefaultValue(propertyName, false);
				}

				if (value == null) {
					// Reset the iterator and try recursively up the classes
					iterator = this._superClasses.iterator();
					while (value == null && iterator.hasNext()) {
						value = iterator.next().getDefaultValue(propertyName,
								true);
					}
				}
			}
		}
		return value;
	}

	@Override
	final public String getName() {
		return this._name;
	}

	@Override
	final public OntologyHandler getOntology() {
		return this._ontology;
	}

	@Override
	final public Collection<String> getPropertyNames() {
		return this._properties != null ? this._properties.keySet()
				: new HashSet<String>();
	}

	@Override
	final public boolean isA(OntologyClassHandler ontologyClass) {
		boolean match = false;

		if (ontologyClass != null) {
			if (this._name.equals(ontologyClass.getName())
					|| this.hasSuperClass(ontologyClass, true)) {
				match = true;
			}
		}
		return match;
	}

	@Override
	final public boolean isA(String className) {
		boolean match = false;
		OntologyClassHandler ontologyClass = null;

		if (className != null) {
			ontologyClass = this._ontology.getOntologyClass(className);
		}
		if (ontologyClass != null) {
			if (this._name.equals(ontologyClass.getName())
					|| this.hasSuperClass(ontologyClass, true)) {
				match = true;
			}
		}
		return match;
	}

	@Override
	final public OntologyInstanceHandler createInstance() {
		StringBuffer name = new StringBuffer(this._name.toLowerCase());
		name.append('_');
		name.append(this.instanceCount);
		
		while (this._instances.containsKey(name.toString())) {
			name.append('_');
			name.append(this.instanceCount);
		}
		return this.createInstance(name.toString());
	}

	@Override
	final public OntologyInstanceHandler createInstance(String name) {
		OntologyInstanceHandler newInstance = this._ontology.createInstance(
				this, name);

		if (newInstance != null) {
			this._instances.put(newInstance.getName(), newInstance);
		}
		this.instanceCount++;
		return newInstance;
	}

	@Override
	final public TreeMap<String, Object> getAllDefaultValues(boolean transitive) {
		TreeMap<String, Object> allValues = new TreeMap<String, Object>();
		allValues.putAll(this._properties);

		if (transitive && this._superClasses != null) {
			for (OntologyClassHandler eachSuperClass : this._superClasses) {
				allValues.putAll(eachSuperClass.getAllDefaultValues(true));
			}
		}
		return allValues;
	}

	@Override
	final public int compareTo(Object o) {
		int comparison = 0;

		if (o instanceof SimpleCaradosClass) {
			comparison = this._name.compareTo(((SimpleCaradosClass) o)
					.getName());
		} else {
			throw new ClassCastException(CaradosMessages
					.getString("Carados.NO_CLASS_CASTING") + //$NON-NLS-1$
					this.getClass().getName() + ": " + //$NON-NLS-1$ 
					o.getClass().getName());
		}
		return comparison;
	}

	/**
	 * Checks to see if the instance is contained in this class.
	 * 
	 * @param instance
	 *            the target instance.
	 * @return <code>true</code> if the instance is part of this class,
	 *         <code>false</code> otherwise.
	 */
	final boolean containsInstance(OntologyInstanceHandler instance) {
		return this._instances != null ? this._instances
				.containsValue(instance) : false;
	}

	@Override
	final public boolean hasProperty(String propertyName) {
		return this._properties != null
				&& this._properties.containsKey(propertyName);
	}

	@Override
	final public int hashCode() {
		return this._name.hashCode();
	}

	@Override
	public void removeDirectSuperClass(OntologyClassHandler superClass) {
		if (this._superClasses != null) {
			this._superClasses.remove(superClass);
		}
	}

	@Override
	public void removeSubClass(OntologyClassHandler subClass) {
		if (this._subClasses != null) {
			this._subClasses.remove(subClass);
		}
	}

	/**
	 * Deletes the specified instance from this class' list of instances.
	 * 
	 * @param instance
	 *            the instance to be deleted.
	 */
	final void deleteInstance(SimpleCaradosInstance instance) {
		if (this._instances != null) {
			this._instances.remove(instance.getName());
		}
	}

	@Override
	final public String toString() {
		return this._name;
	}

	@Override
	public boolean equals(Object obj) {
		return obj instanceof SimpleCaradosClass ? this._name
				.equals(((SimpleCaradosClass) obj).getName()) : false;
	}

	@Override
	public Collection<OntologyInstanceHandler> getSortedInstances(
			OntologyInstanceFilter filter, String propertyName,
			boolean transitive, boolean order) {

		Collection<OntologyInstanceHandler> sorted = null;

		Collection<OntologyInstanceHandler> allInstances = getInstances(filter,
				transitive);

		if (allInstances != null && allInstances.size() > 0) {
			OntologySortTree tree = new OntologySortTree(propertyName);
			sorted = tree.sort(allInstances, order);
		}
		return sorted == null ? new ArrayList<OntologyInstanceHandler>()
				: sorted;
	}

	/**
	 * Retrieves the instance of this class that has the given name. If no such
	 * instance exists then <code>null</code> is returned.
	 * 
	 * @param instanceName
	 *            the name of the instance sought.
	 * @return the <code>OntologyInstanceHandler</code> of the instance.
	 * @since 2.4
	 */
	public OntologyInstanceHandler getInstance(String instanceName) {
		return _instances.get(instanceName);
	}
}
