/*
 * File: Owl2CaradosClass
 *
 * CHANGE HISTORY
 * 0.1	16/01/09	Base line file
 * 0.3	29/01/09	Added isA(String) method for efficiency with Jess.
 * 0.4	02/02/09	getOwnValue returns the default value if there is no instance value.
 * 					Added new methods for doing the casting of property values locally.
 * 		05/02/09	Added toString, hashCode and equals methods.
 * 0.5	12/02/09	Fixed compiler warnings.
 * 					Added getOwnValueAsFloat().
 * 0.9	28/05/09	Added getOwnValueAsDouble().
 * 2.2	19/02/10	Added new method.
 * 					Fixed bug in setOwnValue when value is null.
 */

/*
 * Copyright (c) 2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados.simple;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.TreeMap;

import uk.ac.abdn.carados.OntologyClassHandler;
import uk.ac.abdn.carados.OntologyHandler;
import uk.ac.abdn.carados.OntologyInstanceHandler;

/**
 * <p>
 * A basic example of the OntologyInstanceHandler, representing an instance in
 * the Owl2CaradosOntology. It is intended for applications whose knowledge
 * representation requirements are relatively simple (e.g. do not require a
 * highly expressive representation language like OWL, and do not require
 * reasoning).
 * </p>
 * <h4>Release Notes</h4>
 * <p>
 * Version 2.2 - Added getOwnValueAsCollection(String, Class&lt;T&gt;)
 * </p>
 * <p>
 * Version 0.9 - Added getOwnValueAsDouble()
 * </p>
 * <p>
 * Version 0.4 - getOwnValue now returns the default value if the instance does
 * not have its own value set. Added casting methods for retrieving a specific
 * type of property value. Added the <code>toString</code>,
 * <code>hashCode</code> and <code>equals</code> methods.
 * </p>
 * <p>
 * Version 0.3 - Added <em>isA(String)</em> method to the OntologyEntityHandler.
 * This is then applied to ClassHandlers and InstanceHandlers.
 * </p>
 * 
 * @author A. Gatt & D. Westwater, University of Aberdeen
 */
public class SimpleCaradosInstance implements OntologyInstanceHandler {

	/** The name of this instance. */
	private String _name;

	/** The reference of the ontology that created this instance. */
	private SimpleCaradosOntology _ontology;

	/** The reference to the ontology class this instance belongs to. */
	private SimpleCaradosClass _parentClass;

	/** The list of properties for this instance. */
	private TreeMap<String, Object> _properties = new TreeMap<String, Object>();

	/**
	 * Constructs a new ontology instance handler.
	 * 
	 * @param name
	 *            the name of this instance.
	 * @param parentClass
	 *            the class this instance belongs to.
	 * @param ontology
	 *            the ontology this instance belongs to.
	 */
	protected SimpleCaradosInstance(String name,
			SimpleCaradosClass parentClass, SimpleCaradosOntology ontology) {
		this._name = name;
		this._ontology = ontology;
		this._parentClass = parentClass;
	}

	@SuppressWarnings("unchecked")
	@Override
	final public void addOwnValue(String propertyName, Object value) {
		if (this._properties != null) {
			Object propertyValues = this._properties.get(propertyName);
			if (propertyValues != null) {
				if (propertyValues instanceof Collection) {
					((Collection<Object>) propertyValues).add(value);
				} else {
					ArrayList<Object> newValues = new ArrayList<Object>();
					newValues.add(propertyValues);
					newValues.add(value);
					this._properties.put(propertyName, newValues);
				}
			} else {
				this._properties.put(propertyName, value);
			}
			this._ontology.changedPropertyValue(this, propertyName, value,
					false);
		}
	}

	@Override
	final public OntologyClassHandler getOntologyClass() {
		return this._parentClass;
	}

	@Override
	final public Object getOwnValue(String propertyName) {
		Object value = null;
		if (propertyName != null && this._properties != null) {
			value = this._properties.get(propertyName);
		}
		return value == null ? this.getDefaultValue(propertyName, true) : value;
	}

	@Override
	final public void setOwnValue(String propertyName, Object value) {
		if (this._properties != null) {

			if (null != value)
				this._properties.put(propertyName, value);
			else
				this._properties.remove(propertyName);

			this._ontology
					.changedPropertyValue(this, propertyName, value, true);
		}
	}

	@Override
	final public void delete() {
		if (this._ontology != null) {
			this._ontology.deleteInstance(this);
		}
		if (this._parentClass != null) {
			this._parentClass.deleteInstance(this);
		}
		this._properties = null;
		this._ontology = null;
	}

	@Override
	final public TreeMap<String, Object> getAllDefaultValues(boolean transitive) {
		return this._parentClass != null ? this._parentClass
				.getAllDefaultValues(transitive)
				: new TreeMap<String, Object>();
	}

	@Override
	final public Object getDefaultValue(String propertyName, boolean transitive) {
		return this._parentClass != null ? this._parentClass.getDefaultValue(
				propertyName, transitive) : null;
	}

	@Override
	final public String getName() {
		return this._name;
	}

	@Override
	final public OntologyHandler getOntology() {
		return this._ontology;
	}

	@Override
	final public Collection<String> getPropertyNames() {
		return this._properties != null ? this._properties.keySet()
				: new HashSet<String>();
	}

	@Override
	final public boolean isA(OntologyClassHandler ontologyClass) {
		boolean found = false;

		if (this._parentClass != null) {
			if (this._parentClass.compareTo(ontologyClass) == 0) {
				found = true;
			} else {
				found = this._parentClass.isA(ontologyClass);
			}
		}
		return found;
	}

	@Override
	final public boolean isA(String className) {
		boolean found = false;

		if (this._parentClass != null && className != null) {
			found = this._parentClass.isA(className);
		}
		return found;
	}

	@Override
	final public boolean hasProperty(String propertyName) {
		return this._properties != null
				&& this._properties.containsKey(propertyName);
	}

	@Override
	public Object[] getOwnValueAsArray(String propertyName) {
		Object singleValue = this.getOwnValue(propertyName);
		Object[] arrayValues = null;

		if (singleValue != null) {
			if (singleValue instanceof Collection) {
				arrayValues = ((Collection<?>) singleValue).toArray();
			} else if (singleValue.getClass().isArray()) {
				arrayValues = (Object[]) singleValue;
			} else {
				arrayValues = new Object[] { singleValue };
			}
		}
		return arrayValues;
	}

	@SuppressWarnings("unchecked")
	@Override
	public Collection<Object> getOwnValueAsCollection(String propertyName) {
		Object singleValue = this.getOwnValue(propertyName);
		Collection<Object> setValues = null;

		if (singleValue != null) {
			if (singleValue instanceof Collection) {
				setValues = (Collection<Object>) singleValue;
			} else if (singleValue.getClass().isArray()) {
				setValues = Arrays.asList((Object[]) singleValue);
			} else {
				setValues = Collections.singleton(singleValue);
			}
		}
		return setValues == null ? new HashSet<Object>() : setValues;
	}

	@Override
	public Date getOwnValueAsDate(String propertyName) {
		Object value = getOwnValue(propertyName);
		Date dateValue = null;

		if (value != null) {
			try {
				dateValue = this._ontology.translateToDate(value);
			} catch (Exception e) {
				// Do nothing
			}
		}
		return dateValue;
	}

	@Override
	public Integer getOwnValueAsInteger(String propertyName) {
		Object value = getOwnValue(propertyName);
		return value != null && value instanceof Integer ? (Integer) value
				: null;
	}

	@Override
	public Float getOwnValueAsFloat(String propertyName) {
		Object value = getOwnValue(propertyName);
		return value != null && value instanceof Float ? (Float) value : null;
	}

	@Override
	public String getOwnValueAsString(String propertyName) {
		Object value = getOwnValue(propertyName);
		String stringValue = null;

		if (value != null) {
			if (value instanceof String) {
				stringValue = (String) value;
			} else {
				stringValue = value.toString();
			}
		}
		return stringValue;
	}

	@Override
	final public String toString() {
		return this._name;
	}

	@Override
	public boolean equals(Object obj) {
		return obj instanceof SimpleCaradosInstance ? this._name
				.equals(((SimpleCaradosInstance) obj).getName()) : false;
	}

	@Override
	final public int hashCode() {
		return this._name.hashCode();
	}

	@Override
	public Double getOwnValueAsDouble(String propertyName) {
		Object value = getOwnValue(propertyName);
		return value != null && value instanceof Double ? (Double) value : null;
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T> Collection<T> getOwnValueAsCollection(String propertyName,
			Class<T> type) {

		Object singleValue = this.getOwnValue(propertyName);
		Collection<T> setValues = null;

		if (singleValue != null) {
			if (singleValue instanceof Collection<?>) {
				setValues = (Collection<T>) singleValue;
			} else if (singleValue.getClass().isArray()) {
				setValues = Arrays.asList((T[]) singleValue);
			} else {
				setValues = (Collection<T>) Collections
						.singleton((T) singleValue);
			}
		}
		return setValues == null ? new HashSet<T>() : setValues;
	}
}
