/*
 * File: Owl2CaradosOntology
 *
 * CHANGE HISTORY
 * 0.1	14/01/09	Base line file
 * 0.5	09/02/09	Externalised strings.
 * 		12/02/09	Fixed compiler warnings.
 * 0.8	05/03/09	Added new method for accepting a filter when grabbing all the 
 * 					instances in the ontology.
 */

/*
 * Copyright (c) 2007-2007, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados.simple;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.TreeMap;

import org.w3c.dom.Document;

import uk.ac.abdn.aries.DOMUtil;
import uk.ac.abdn.aries.DOMUtilException;
import uk.ac.abdn.carados.CaradosMessages;
import uk.ac.abdn.carados.OntologyClassHandler;
import uk.ac.abdn.carados.OntologyException;
import uk.ac.abdn.carados.OntologyHandler;
import uk.ac.abdn.carados.OntologyInstanceFilter;
import uk.ac.abdn.carados.OntologyInstanceHandler;
import uk.ac.abdn.carados.OntologyListener;

/**
 * <p>
 * A basic example of the OntologyHandler, which is not backed by an ontology.
 * This class is intended for applications whose knowledge representation
 * requirements are relatively simple (e.g. do not require a highly expressive
 * representation language like OWL, and do not require reasoning).
 * </p>
 * <h4>Release Notes</h4>
 * <p>
 * Version 0.8 - Ontology will now report to listener any missing properties that are 
 * being set. Add a new method: <code>getAllInstances(OntologyInstanceFilter)</code>
 * </p>
 * <p>Version 0.5 - Externalised strings</p>
 * @author A. Gatt & D. Westwater, University of Aberdeen
 */
public class SimpleCaradosOntology implements OntologyHandler {
	/** The list of listeners for this handler. */
	private ArrayList<OntologyListener> _listeners = new ArrayList<OntologyListener>();

	/** The map containing all the classes in this ontology. */
	private TreeMap<String, OntologyClassHandler> _classes = new TreeMap<String, OntologyClassHandler>();

	/** This is the root class for the ontology. */
	private SimpleCaradosClass _rootClass;

	/** The name of the root class. */
	private static final String ROOT_CLASS_NAME = 
		CaradosMessages.getString("Carados.ROOT_CLASS"); //$NON-NLS-1$

	/** The name of the ontology. */
	private String _name = ""; //$NON-NLS-1$

	/**
	 * Basic constructor for the Owl2CaradosOntology.
	 */
	public SimpleCaradosOntology() {
		this._rootClass = new SimpleCaradosClass(ROOT_CLASS_NAME, this);
		this._classes.put(ROOT_CLASS_NAME, this._rootClass);
	}

	@Override
	final public void attachListener(OntologyListener listener) {
		if (this._listeners != null) {
			this._listeners.add(listener);
		}
	}

	@Override
	final public boolean containsClass(String className) {
		return (this._classes != null && this._classes.containsKey(className));
	}

	@Override
	final public boolean containsClass(OntologyClassHandler classHandler) {
		return (this._classes != null && this._classes
				.containsValue(classHandler));
	}

	@Override
	final public boolean containsInstance(
			OntologyInstanceHandler instanceHandler) {
		boolean found = false;

		Collection<OntologyClassHandler> allClasses = this._classes.values();

		if (allClasses != null) {
			Iterator<OntologyClassHandler> eachClass = allClasses.iterator();
			while (!found && eachClass.hasNext()) {
				found = ((SimpleCaradosClass) eachClass.next())
				.containsInstance(instanceHandler);
			}
		}
		return found;
	}

	@Override
	final public boolean containsProperty(String propertyName) {
		boolean found = false;

		if (this._classes != null) {
			Iterator<OntologyClassHandler> eachClass = this._classes.values()
			.iterator();

			while (!found && eachClass.hasNext()) {
				found = eachClass.next().hasProperty(propertyName);
			}
		}
		return found;
	}

	@Override
	final public OntologyClassHandler createClass(String className) {
		OntologyClassHandler newClass = null;
		if (this._classes != null && !className.equals(ROOT_CLASS_NAME)) {
			newClass = this._classes.get(className);

			if (newClass == null) {
				newClass = new SimpleCaradosClass(className, this);
			}

			if (this._listeners != null) {
				for (OntologyListener eachListener : this._listeners) {
					eachListener.classCreated(newClass);
				}
			}
			this._classes.put(className, newClass);
			newClass.addDirectSuperClass(this._rootClass);
		}
		return className.equals(ROOT_CLASS_NAME) ? this._rootClass : newClass;
	}
	
	@Override
	final public Collection<OntologyInstanceHandler>
	getAllInstances() {
		return getAllInstances(null);
	}

	@Override
	final public Collection<OntologyInstanceHandler> getAllInstances(
			OntologyInstanceFilter filter) {
		HashSet<OntologyInstanceHandler> allInstances = new HashSet<OntologyInstanceHandler>();

		if (this._classes != null) {
			Collection<OntologyClassHandler> allClasses = this._classes
			.values();
			
			for (OntologyClassHandler eachClass : allClasses) {
				for (OntologyInstanceHandler instance : eachClass.getInstances(false))
				{
					if (filter == null || filter.canInclude(instance)) {
						allInstances.add(instance);
					}
				}
			}
		}

		return allInstances;
	}

	@Override
	final public Collection<OntologyClassHandler> getAllOntologyClasses() {
		return this._classes != null ? this._classes.values()
				: new HashSet<OntologyClassHandler>();
	}

	@Override
	final public OntologyClassHandler getOntologyClass(String className) {
		return this._classes != null ? this._classes.get(className) : null;
	}

	@Override
	final public Collection<String> getPropertyNames() {
		HashSet<String> allProperties = new HashSet<String>();
		if (this._classes != null)
		{
			for (OntologyClassHandler eachClass : this._classes.values())
			{
				allProperties.addAll(eachClass.getPropertyNames());
			}
		}
		return allProperties;
	}

	@Override
	final public OntologyClassHandler getRootClass() {
		return this._rootClass;
	}

	@Override
	final public void removeAllListeners() {
		this._listeners.clear();
	}

	@Override
	final public void removeListener(OntologyListener listener) {
		this._listeners.remove(listener);
	}

	@Override
	final public void saveOntology(String fileName) throws OntologyException {
		Document doc = SimpleCaradosXMLEncoder.encodeOntology(this);
		if (doc != null) {
			try {
				DOMUtil.saveDocument(doc, fileName);
			} catch (DOMUtilException e) {
				throw new OntologyException(CaradosMessages.getString("Carados.NO_ONTOLOGY_SAVING") //$NON-NLS-1$
						+ e.getMessage());
			}
		}
	}

	@Override
	final public Object translateDate(Date date) {
		return translateDate(date.getTime());
	}

	@Override
	final public Object translateDate(long timestamp) {
		return new Long(timestamp);
	}

	@Override
	final public Date translateToDate(Object object) {
		Date date = null;
		if (object instanceof Long) {
			date = new Date(((Long)object).longValue());
		} else if (object instanceof Date) {
			date = (Date) ((Date) object).clone();
		}
		return date;
	}

	@Override
	final public long translateToTimestamp(Object object) {
		long timestamp = 0;
		if (object instanceof Long) {
			timestamp = ((Long) object).longValue();
		} else if (object instanceof Date) {
			timestamp = ((Date) object).getTime();
		}
		return timestamp;
	}

	/**
	 * Creates a new instance of a particular class.
	 * 
	 * @param parentClass
	 *            the class to which the new instance belongs.
	 * @param name
	 *            the name of the instance.
	 * @return a reference to the new instance.
	 */
	final OntologyInstanceHandler createInstance(SimpleCaradosClass parentClass,
			String name) {
		OntologyInstanceHandler newInstance = 
			new SimpleCaradosInstance(name.toString(), parentClass, this);
		
		if (this._listeners != null) {
			for (OntologyListener eachListener : this._listeners) {
				eachListener.instanceCreated(newInstance);
			}
		}
		return newInstance;
	}

	/**
	 * Reports to the listeners that a super class has been added.
	 * 
	 * @param baseClass
	 *            the base class.
	 * @param superClass
	 *            the super class.
	 * @param overwritten
	 *            <code>true</code> if the new super class has replaced existing
	 *            super classes, <code>false</code> otherwise.
	 */
	final void addedSuperClass(OntologyClassHandler baseClass,
			OntologyClassHandler superClass, boolean overwritten) {

		if (this._listeners != null) {
			for (OntologyListener eachListener : this._listeners) {
				eachListener.superClassAdded(baseClass, superClass, overwritten);
			}
		}
	}

	/**
	 * Reports to the listeners that a sub class has been added.
	 * 
	 * @param baseClass
	 * @param subClass
	 */
	final void addedSubClass(OntologyClassHandler baseClass,
			OntologyClassHandler subClass) {
		if (this._listeners != null) {
			for (OntologyListener eachListener : this._listeners) {
				eachListener.subClassAdded(baseClass, subClass);
			}
		}
	}

	/**
	 * Reports to the listeners that a default property value has been changed.
	 * 
	 * @param ontologyClass
	 *            the class the property belongs to.
	 * @param propertyName
	 *            the property changed.
	 * @param propertyValue
	 *            the new value.
	 */
	final void changedDefaultValue(OntologyClassHandler ontologyClass,
			String propertyName, Object propertyValue) {
		if (this._listeners != null) {
			for (OntologyListener eachListener : this._listeners) {
				eachListener.defaultValueChanged(ontologyClass, propertyName,
						propertyValue);
			}
		}
	}

	/**
	 * Deletes the class from the list of classes associated with this ontology.
	 * 
	 * @param ontologyClass
	 *            the class to be removed.
	 */
	final void deleteClass(SimpleCaradosClass ontologyClass) {
		if (this._classes != null) {
			this._classes.remove(ontologyClass);
			if (this._listeners != null) {
				for (OntologyListener eachListener : this._listeners) {
					eachListener.classDeleted(ontologyClass.getName());
				}
			}
		}
	}

	/**
	 * Reports to the listeners that a property value has been changed.
	 * 
	 * @param instance
	 *            the class the property belongs to.
	 * @param propertyName
	 *            the property changed.
	 * @param propertyValue
	 *            the new value.
	 * @param overwritten
	 *            <code>true</code> if the new value has replaced all existing
	 *            values of the property, otherwise <code>false</code>.
	 */
	final void changedPropertyValue(OntologyInstanceHandler instance,
			String propertyName, Object propertyValue, boolean overwritten) {
		if (this._listeners != null) {
			for (OntologyListener eachListener : this._listeners) {
				eachListener.propertyValueChanged(instance, propertyName,
						propertyValue, overwritten);
			}
		}
	}

	/**
	 * Reports that the instance has been removed.
	 * 
	 * @param instance
	 *            the instance that has been deleted.
	 */
	final void deleteInstance(SimpleCaradosInstance instance) {
		if (this._listeners != null) {
			for (OntologyListener eachListener : this._listeners) {
				eachListener.instanceDeleted(instance.getName());
			}
		}
	}

	@Override
	final public void loadOntology(URI uri) throws OntologyException {
		this._name = uri.getFragment();
		try
		{
			Document doc = DOMUtil.loadDocument(uri);
			if (doc != null) {
				SimpleCaradosXMLDecoder.decodeOntology(doc, this);
			}
		} catch (DOMUtilException e)
		{
			throw new OntologyException(CaradosMessages.getString("Carados.NO_ONTOLOGY_LOADING") + e.getMessage()); //$NON-NLS-1$
		}
	}

	@Override
	final public String toString()
	{
		return this._name;
	}
}
