/*
 * File: SimpleCaradosXMLDecoder
 *
 * CHANGE HISTORY
 * 0.1	14/01/09	Base line file
 * 2.4	05/03/10	Added date formatter for reading in date types.
					Uses existing instances now rather than creating an instances for 
					every occurrence. This was required as it will now correctly parse 
					the instances from files.
 */

/*
 * Copyright (c) 2007-2007, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados.simple;

import java.lang.reflect.Constructor;
import java.text.SimpleDateFormat;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import uk.ac.abdn.carados.OntologyClassHandler;
import uk.ac.abdn.carados.OntologyEntityHandler;
import uk.ac.abdn.carados.OntologyHandler;
import uk.ac.abdn.carados.OntologyInstanceHandler;

/**
 * <p>
 * This class contains methods for creating a Owl2CaradosOntology from a DOM as
 * laid out in SimpleOntologyXML format. The layout for which, is:<br>
 * <code>
 * <blockquote>
 * &lt;ontology name="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;&lt;class name="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;property name="..." type="..." value="..." /&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;property name="..." type="..." value="..." /&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;/class&gt;<br>
 * &nbsp;&nbsp;&nbsp;&lt;class name="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;/class&gt;<br>
 * &nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;instance name="..." class="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;property name="..." type="..." value="..." /&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;property name="..." type="..." value="..." /&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;/instance&gt;<br>
 * &nbsp;&nbsp;&nbsp;&lt;instance name="..." class="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;/instance&gt;<br>
 * &nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;hierarchy&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;parent name="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;child name="..."/&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;child name="..."/&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;/parent&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;parent name="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;/parent&gt;<br>
 * &nbsp;&nbsp;&nbsp;&lt;/hierarchy&gt;<br>
 * &lt;/ontology&gt;<br>
 * </blockquote>
 * </code> The properties recorded under the <em>class</em> tag represent the
 * default values for those properties for the given class. The hierarchy of
 * classes is recorded separately from the class definitions to avoid any
 * looping with the class structure.
 * </p>
 * 
 * @author D.Westwater & A.Gatt, University of Aberdeen
 */
abstract public class SimpleCaradosXMLDecoder {
	private static final SimpleDateFormat DATE_FORMATTER = new SimpleDateFormat(
			"EEE MMM dd HH:mm:ss z yyyy");

	/**
	 * Decodes the DOM document into the ontology.
	 * 
	 * @param doc
	 *            the DOM Document
	 * @param ontology
	 *            the ontology where the information is to be written to.
	 */
	static final public void decodeOntology(Document doc,
			OntologyHandler ontology) {
		try {
			decodeClasses(doc, ontology);
			decodeInstances(doc, ontology);
			decodeHierarchy(doc, ontology);
		} catch (XPathExpressionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Decodes the XPath in the DOM document to create all the classes in the
	 * ontology.
	 * 
	 * @param doc
	 *            the DOM document.
	 * @param ontology
	 *            the ontology being created.
	 * @throws XPathExpressionException
	 *             if the class node path cannot be evaluated.
	 */
	static final private void decodeClasses(Document doc,
			OntologyHandler ontology) throws XPathExpressionException {
		XPath xpath = XPathFactory.newInstance().newXPath();

		NodeList nodelist;
		nodelist = (NodeList) xpath.evaluate(SimpleMessages
				.getString("SimpleXML.CLASS_NODE_PATH"), doc, //$NON-NLS-1$
				XPathConstants.NODESET);

		NamedNodeMap nnm = null;
		int length = nodelist.getLength();
		Node classNode = null;
		OntologyClassHandler ontologyClass = null;

		for (int i = 0; i < length; i++) {
			classNode = nodelist.item(i);
			nnm = classNode.getAttributes();
			ontologyClass = ontology
					.createClass(nnm
							.getNamedItem(
									SimpleMessages
											.getString("SimpleXML.NAME_ATTRIBUTE")).getNodeValue()); //$NON-NLS-1$

			if (ontologyClass != null) {
				decodeProperties(classNode.getChildNodes(), ontologyClass);
			}
		}
	}

	/**
	 * Decodes the XPath in the DOM document to create all the instances in the
	 * ontology. It checks first to see if the instance with the given name
	 * already exists. If it does then that instance is used.
	 * 
	 * @param doc
	 *            the DOM document.
	 * @param ontology
	 *            the ontology being created.
	 * @throws XPathExpressionException
	 *             if the instance node path cannot be evaluated.
	 */
	static final private void decodeInstances(Document doc,
			OntologyHandler ontology) throws XPathExpressionException {
		XPath xpath = XPathFactory.newInstance().newXPath();

		NodeList nodelist;
		nodelist = (NodeList) xpath.evaluate(SimpleMessages
				.getString("SimpleXML.INSTANCE_NODE_PATH"), doc, //$NON-NLS-1$
				XPathConstants.NODESET);

		NamedNodeMap nnm = null;
		int length = nodelist.getLength();
		String className = null;
		OntologyClassHandler ontologyClass = null;
		OntologyInstanceHandler instance = null;
		Node instanceNode = null;
		String instanceName = null;

		for (int i = 0; i < length; i++) {
			instanceNode = nodelist.item(i);
			nnm = instanceNode.getAttributes();
			className = nnm.getNamedItem(
					SimpleMessages.getString("SimpleXML.CLASS_ATTRIBUTE")). //$NON-NLS-1$
					getNodeValue();

			if (className != null) {
				ontologyClass = ontology.getOntologyClass(className);
				if (ontologyClass != null) {
					instanceName = nnm.getNamedItem(
							SimpleMessages
									.getString("SimpleXML.NAME_ATTRIBUTE")). //$NON-NLS-1$
							getNodeValue();
					instance = ((SimpleCaradosClass) ontologyClass)
							.getInstance(instanceName);

					if (instance == null) {
						instance = ontologyClass.createInstance(instanceName);
					}
					if (instance != null) {
						decodeProperties(instanceNode.getChildNodes(), instance);
					}
				}
			}
		}
	}

	/**
	 * Decodes the property nodes for classes and instances.
	 * 
	 * @param nodelist
	 *            the list of property nodes.
	 * @param entity
	 *            the class to which these properties belong.
	 */
	private static final void decodeProperties(NodeList nodelist,
			OntologyEntityHandler entity) {
		try {
			NamedNodeMap nnm = null;
			int length = nodelist.getLength();
			Node propertyNode = null;
			Object value = null;
			Class<?> StringClass = Class.forName("java.lang.String"); //$NON-NLS-1$
			Constructor<?> constructor = null;
			String type = null;
			OntologyHandler ontology = entity.getOntology();
			OntologyClassHandler propertyClass = null;
			OntologyInstanceHandler propertyInstance = null;
			Class<?> typeClass = null;
 
			for (int i = 0; i < length; i++) {
				propertyNode = nodelist.item(i);
				value = null;

				if (propertyNode.getNodeName().equals(
						SimpleMessages.getString("SimpleXML.PROPERTY_NODE"))) //$NON-NLS-1$
				{
					nnm = propertyNode.getAttributes();
					type = nnm.getNamedItem(
							SimpleMessages
									.getString("SimpleXML.TYPE_ATTRIBUTE"))
							.getNodeValue();

					if (type.equals(SimpleMessages
							.getString("SimpleXML.INSTANCE_NODE"))) {

						propertyClass = ontology
								.getOntologyClass(nnm
										.getNamedItem(
												SimpleMessages
														.getString("SimpleXML.CLASS_ATTRIBUTE"))
										.getNodeValue());

						propertyInstance = ((SimpleCaradosClass) propertyClass)
								.getInstance(nnm
										.getNamedItem(
												SimpleMessages
														.getString("SimpleXML.VALUE_ATTRIBUTE"))
										.getNodeValue());

						if (null == propertyInstance) {
							propertyInstance = propertyClass
									.createInstance(nnm
											.getNamedItem(
													SimpleMessages
															.getString("SimpleXML.VALUE_ATTRIBUTE"))
											.getNodeValue());
						}
						value = propertyInstance;
					} else if (type.equals("java.util.Date")) {
						value = DATE_FORMATTER
								.parse(nnm
										.getNamedItem(
												SimpleMessages
														.getString("SimpleXML.VALUE_ATTRIBUTE"))
										.getNodeValue());
					} else {
						try {
							typeClass = Class.forName(type);
							if (typeClass.isEnum()) {
								Object[] constants = typeClass.getEnumConstants();
								String enumValue = nnm
								.getNamedItem(
										SimpleMessages
												.getString("SimpleXML.VALUE_ATTRIBUTE")) //$NON-NLS-1$
								.getNodeValue();
								
								value = null;
								int j = 0; 
								while (null == value && j < constants.length) {
									if (constants[j].toString().equalsIgnoreCase(enumValue)) {
										value = constants[j];
									}
									j++;
								}
							} else {
								constructor = Class.forName(type).getConstructor(
										StringClass);
	
								if (constructor != null) {
									value = constructor
											.newInstance(nnm
													.getNamedItem(
															SimpleMessages
																	.getString("SimpleXML.VALUE_ATTRIBUTE")) //$NON-NLS-1$
													.getNodeValue());
								}
							}
						} catch (ClassNotFoundException e) {
							value = new String(
									type
											+ '.'
											+ nnm
													.getNamedItem(
															SimpleMessages
																	.getString("SimpleXML.VALUE_ATTRIBUTE")) //$NON-NLS-1$
													.getNodeValue());
						}
					}
					if (null != value && entity instanceof OntologyClassHandler) {
						((OntologyClassHandler) entity)
								.setDefaultValue(
										nnm
												.getNamedItem(
														SimpleMessages
																.getString("SimpleXML.NAME_ATTRIBUTE")) //$NON-NLS-1$
												.getNodeValue(), value);
					} else if (null != value
							&& entity instanceof OntologyInstanceHandler) {
						((OntologyInstanceHandler) entity)
								.addOwnValue(
										nnm
												.getNamedItem(
														SimpleMessages
																.getString("SimpleXML.NAME_ATTRIBUTE")) //$NON-NLS-1$
												.getNodeValue(), value);
					}
				}
			}
		} catch (Exception e) {
			// Do nothing
		}
	}

	/**
	 * Decodes the class hierarchy information.
	 * 
	 * @param doc
	 *            the document being decoded.
	 * @param ontology
	 *            the ontology being created.
	 */
	private static final void decodeHierarchy(Document doc,
			OntologyHandler ontology) {
		OntologyClassHandler root = ontology.getRootClass();

		if (root != null) {
			for (OntologyClassHandler eachClass : ontology
					.getAllOntologyClasses()) {
				root.removeSubClass(eachClass);
				eachClass.removeDirectSuperClass(root);
			}
		}

		try {
			XPath xpath = XPathFactory.newInstance().newXPath();

			NodeList nodelist;
			nodelist = (NodeList) xpath.evaluate(SimpleMessages
					.getString("SimpleXML.PARENT_NODE_PATH"), doc, //$NON-NLS-1$
					XPathConstants.NODESET);

			NamedNodeMap nnm = null;
			int length = nodelist.getLength();
			String parentName = null;
			OntologyClassHandler parentClass = null;
			Node parentNode = null;

			for (int i = 0; i < length; i++) {
				parentNode = nodelist.item(i);
				nnm = parentNode.getAttributes();
				parentName = nnm.getNamedItem(
						SimpleMessages.getString("SimpleXML.NAME_ATTRIBUTE")) //$NON-NLS-1$
						.getNodeValue();

				if (parentName != null) {
					parentClass = ontology.getOntologyClass(parentName);

					if (parentClass != null) {
						decodeSubClasses(parentNode.getChildNodes(),
								parentClass, ontology);
					}
				}
			}
		} catch (Exception e) {
			// Do nothing
		}
	}

	/**
	 * Adds the subclasses to the parent class in the hierarchy.
	 * 
	 * @param nodelist
	 *            the list of child nodes of the parent node
	 * @param parentClass
	 *            the ontology class of the parent.
	 * @param ontology
	 *            the ontology reference.
	 */
	private final static void decodeSubClasses(NodeList nodelist,
			OntologyClassHandler parentClass, OntologyHandler ontology) {
		NamedNodeMap nnm = null;
		int length = nodelist.getLength();
		String childName = null;
		OntologyClassHandler childClass = null;
		Node childNode = null;

		for (int i = 0; i < length; i++) {
			childNode = nodelist.item(i);

			if (childNode.getNodeName().equals(
					SimpleMessages.getString("SimpleXML.CHILD_NODE"))) //$NON-NLS-1$
			{
				nnm = childNode.getAttributes();
				childName = nnm.getNamedItem(
						SimpleMessages.getString("SimpleXML.NAME_ATTRIBUTE")) //$NON-NLS-1$
						.getNodeValue();

				if (childName != null) {
					childClass = ontology.getOntologyClass(childName);

					if (childClass != null) {
						parentClass.addSubClass(childClass);
					}
				}
			}
		}
	}
}
