/*
 * File: SimpleCaradosXMLEncoder
 *
 * CHANGE HISTORY
 * 0.1	16/01/09	Base line file
 * 0.5	09/02/09	Externalised strings.
 * 		12/02/09	Fixed compiler warnings
 * 2.4	05/03/10	Added a class attribute when storing instances as property values.
 */

/*
 * Copyright (c) 2007-2007, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados.simple;

import java.util.ArrayList;
import java.util.Collection;
import java.util.TreeMap;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import uk.ac.abdn.carados.CaradosMessages;
import uk.ac.abdn.carados.OntologyClassHandler;
import uk.ac.abdn.carados.OntologyException;
import uk.ac.abdn.carados.OntologyHandler;
import uk.ac.abdn.carados.OntologyInstanceHandler;

/**
 * <p>
 * This class will convert an ontology into a simple XML style format contained
 * within a DOM. The layout for which, is:<br>
 * <code>
 * <blockquote>
 * &lt;ontology name="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;&lt;class name="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;property name="..." type="..." value="..." /&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;property name="..." type="..." value="..." /&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;/class&gt;<br>
 * &nbsp;&nbsp;&nbsp;&lt;class name="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;/class&gt;<br>
 * &nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;instance name="..." class="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;property name="..." type="..." value="..." /&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;property name="..." type="..." value="..." /&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;/instance&gt;<br>
 * &nbsp;&nbsp;&nbsp;&lt;instance name="..." class="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;/instance&gt;<br>
 * &nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&lt;hierarchy&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;parent name="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;child name="..."/&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;child name="..."/&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;/parent&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;parent name="..."&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;/parent&gt;<br>
 * &nbsp;&nbsp;&nbsp;&lt;/hierarchy&gt;<br>
 * &lt;/ontology&gt;<br>
 * </blockquote>
 * </code> The properties recorded under the <em>class</em> tag represent the
 * default values for those properties for the given class. The hierarchy of
 * classes is recorded separately from the class definitions to avoid any
 * looping with the class structure.
 * </p>
 * <h4>Release Notes</h4>
 * <p>
 * Version 0.5 - Externalised strings.
 * </p>
 * 
 * @author D.Westwater & A.Gatt, University of Aberdeen
 */
abstract public class SimpleCaradosXMLEncoder {
	/**
	 * Encodes the ontology into the XML format.
	 * 
	 * @param ontology
	 *            the reference to the ontology to be encoded.
	 * @return the DOM document.
	 * @throws OntologyException
	 *             if there is a problem with the encoding.
	 */
	static final public Document encodeOntology(OntologyHandler ontology)
			throws OntologyException {
		Document doc = null;
		try {
			DocumentBuilder builder = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder();
			doc = builder.newDocument();
			Node docRoot = doc.createElement(SimpleMessages
					.getString("SimpleXML.ROOT_NODE")); //$NON-NLS-1$
			Node hierarchyNode = doc.createElement(SimpleMessages
					.getString("SimpleXML.HIERARCHY_NODE")); //$NON-NLS-1$

			Node subClassNode = null;

			for (OntologyClassHandler eachClass : ontology
					.getAllOntologyClasses()) {
				docRoot.appendChild(encodeClass(doc, eachClass));
				subClassNode = encodeHierarchy(doc, eachClass);
				if (subClassNode != null) {
					hierarchyNode.appendChild(subClassNode);
				}
			}
			for (OntologyInstanceHandler eachInstance : ontology
					.getAllInstances()) {
				docRoot.appendChild(encodeInstance(doc, eachInstance));
			}
			docRoot.appendChild(hierarchyNode);
			doc.appendChild(docRoot);
		} catch (ParserConfigurationException e) {
			throw new OntologyException(CaradosMessages
					.getString("Carados.NO_XML_ENCODER") + e.getMessage()); //$NON-NLS-1$
		}
		return doc;
	}

	/**
	 * Creates a Node to represent an ontology class.
	 * 
	 * @param doc
	 *            the base document.
	 * @param ontologyClass
	 *            the class to be encoded.
	 * @return the created Node.
	 */
	final private static Node encodeClass(Document doc,
			OntologyClassHandler ontologyClass) {
		Node classNode = doc.createElement(SimpleMessages
				.getString("SimpleXML.CLASS_NODE")); //$NON-NLS-1$
		NamedNodeMap attributes = classNode.getAttributes();
		attributes
				.setNamedItem(constructAttributeNode(
						doc,
						SimpleMessages.getString("SimpleXML.NAME_ATTRIBUTE"), ontologyClass.getName())); //$NON-NLS-1$

		TreeMap<String, Object> allValues = ontologyClass
				.getAllDefaultValues(false);
		Collection<Node> propertyNodes = null;

		for (String eachProperty : allValues.keySet()) {
			propertyNodes = constructPropertyNodes(doc, eachProperty, allValues
					.get(eachProperty));
			for (Node eachPropertyNode : propertyNodes) {
				classNode.appendChild(eachPropertyNode);
			}
		}
		return classNode;
	}

	/**
	 * Creates a Node to represent an instance.
	 * 
	 * @param doc
	 *            the base document.
	 * @param instance
	 *            the instance to be encoded.
	 * @return the created Node.
	 */
	final private static Node encodeInstance(Document doc,
			OntologyInstanceHandler instance) {
		Node instanceNode = doc.createElement(SimpleMessages
				.getString("SimpleXML.INSTANCE_NODE")); //$NON-NLS-1$
		NamedNodeMap attributes = instanceNode.getAttributes();

		attributes.setNamedItem(constructAttributeNode(doc, SimpleMessages
				.getString("SimpleXML.NAME_ATTRIBUTE"), //$NON-NLS-1$
				instance.getName()));

		attributes.setNamedItem(constructAttributeNode(doc, SimpleMessages
				.getString("SimpleXML.CLASS_ATTRIBUTE"), //$NON-NLS-1$
				instance.getOntologyClass().getName()));

		Collection<String> allProperties = instance.getPropertyNames();
		Collection<Node> propertyNodes = null;

		for (String eachProperty : allProperties) {
			propertyNodes = constructPropertyNodes(doc, eachProperty, instance
					.getOwnValue(eachProperty));
			for (Node eachPropertyNode : propertyNodes) {
				instanceNode.appendChild(eachPropertyNode);
			}
		}
		return instanceNode;
	}

	/**
	 * Creates an Attribute of the given name with the given value.
	 * 
	 * @param doc
	 *            the base document.
	 * @param name
	 *            the name for this attribute.
	 * @param value
	 *            the value of this attribute.
	 * @return the created Node.
	 */
	final private static Attr constructAttributeNode(Document doc, String name,
			String value) {
		Attr attribute = doc.createAttribute(name);
		attribute.setNodeValue(value);
		return attribute;
	}

	/**
	 * Creates the property nodes for a property on a class or instance. Note
	 * that a collection of nodes is returned as the property value may consist
	 * of more than one element. A node is created for each element. If the
	 * value is an array or a collection then this method is called recursively
	 * on each element.
	 * 
	 * @param doc
	 *            the base document.
	 * @param name
	 *            the name of the property.
	 * @param value
	 *            the value of the property.
	 * @return the collection of created nodes.
	 */
	final private static Collection<Node> constructPropertyNodes(Document doc,
			String name, Object value) {
		ArrayList<Node> nodes = new ArrayList<Node>();

		if (value.getClass().isArray()) {
			for (Object eachObject : (Object[]) value) {
				nodes.addAll(constructPropertyNodes(doc, name, eachObject));
			}
		} else if (value instanceof Collection) {
			for (Object eachObject : (Collection<?>) value) {
				nodes.addAll(constructPropertyNodes(doc, name, eachObject));
			}
		} else {
			Node propertyNode = doc.createElement(SimpleMessages
					.getString("SimpleXML.PROPERTY_NODE")); //$NON-NLS-1$

			NamedNodeMap attributes = propertyNode.getAttributes();
			attributes.setNamedItem(constructAttributeNode(doc, SimpleMessages
					.getString("SimpleXML.NAME_ATTRIBUTE"), name)); //$NON-NLS-1$

			attributes.setNamedItem(constructAttributeNode(doc, SimpleMessages
					.getString("SimpleXML.TYPE_ATTRIBUTE"), getType(value))); //$NON-NLS-1$

			attributes.setNamedItem(constructAttributeNode(doc, SimpleMessages
					.getString("SimpleXML.VALUE_ATTRIBUTE"), value.toString())); //$NON-NLS-1$

			if (value instanceof OntologyInstanceHandler) {
				attributes
						.setNamedItem(constructAttributeNode(
								doc,
								SimpleMessages
										.getString("SimpleXML.CLASS_ATTRIBUTE"), ((OntologyInstanceHandler) value).getOntologyClass().getName())); //$NON-NLS-1$
			}

			nodes.add(propertyNode);
		}
		return nodes;
	}

	/**
	 * Retrieves the type of an object.
	 * 
	 * @param value
	 *            the object whose type is desired.
	 * @return the type as a string.
	 */
	final private static String getType(Object value) {
		String type = null;

		if (value instanceof OntologyClassHandler) {
			type = SimpleMessages.getString("SimpleXML.CLASS_NODE"); //$NON-NLS-1$
		} else if (value instanceof OntologyInstanceHandler) {
			type = SimpleMessages.getString("SimpleXML.INSTANCE_NODE"); //$NON-NLS-1$
		} else {
			type = value.getClass().getName();
		}
		return type;
	}

	/**
	 * Creates a parent node for entry into the class hierarchy.
	 * 
	 * @param doc
	 *            the base document.
	 * @param parent
	 *            the ontology class acting as the parent.
	 * @return the created parent node or <code>null</code> if this class has no
	 *         subclasses.
	 */
	final private static Node encodeHierarchy(Document doc,
			OntologyClassHandler parent) {
		Node parentNode = null;
		Node childNode = null;
		Collection<OntologyClassHandler> subClasses = parent
				.getSubClasses(false);
		NamedNodeMap parentAttributes = null;
		NamedNodeMap childAttributes = null;

		if (subClasses != null && subClasses.size() > 0) {
			parentNode = doc.createElement(SimpleMessages
					.getString("SimpleXML.PARENT_NODE")); //$NON-NLS-1$
			parentAttributes = parentNode.getAttributes();
			parentAttributes
					.setNamedItem(constructAttributeNode(
							doc,
							SimpleMessages
									.getString("SimpleXML.NAME_ATTRIBUTE"), parent.getName())); //$NON-NLS-1$

			for (OntologyClassHandler eachSubClass : subClasses) {
				childNode = doc.createElement(SimpleMessages
						.getString("SimpleXML.CHILD_NODE")); //$NON-NLS-1$
				childAttributes = childNode.getAttributes();
				childAttributes.setNamedItem(constructAttributeNode(doc,
						SimpleMessages.getString("SimpleXML.NAME_ATTRIBUTE"), //$NON-NLS-1$
						eachSubClass.getName()));

				parentNode.appendChild(childNode);
			}
		}
		return parentNode;
	}
}
