/*
 * File: OntologySortTree
 *
 * CHANGE HISTORY
 * 0.11	23/06/09	Base line file
 * 1.0.1	13/07/09	Handles sorting on a tree with no nodes in it.
 */

/*
 * Copyright (c) 2007-2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados.util;

import java.util.ArrayList;
import java.util.Collection;

import uk.ac.abdn.carados.OntologyInstanceHandler;

/**
 * <p>This class represents a tree structure for sorting instances based on the values
 * of a particular property. It is more complex than the standard Java Tree implementations 
 * but the principles are the same.</p>
 * <p>Each instantiation of the Tree works on a given property of the class. This is the 
 * basis for comparing instances as they are added to the Tree. The instances are stored in 
 * the <code>OntologySortTreeNode</code>. The Tree is constructed in the natural ascending 
 * order of the property values. The left child branch forms the elements that come before the 
 * current node. The right child branch forms the elements that come after the current node 
 * but before the current node's parent. A simple example using integer values is given 
 * below:</p>
 * <pre>
 *                     76
 *                    /  \
 *                  56    123
 *                 /  \
 *               34    57
 * </pre>
 * <p>In most cases the property will have a set type in which case the elements are 
 * ordered according to their natural order as determined by their <code>compareTo</code> 
 * method. For non-Comparable objects the value returned by the <code>toString</code> method
 * is used for comparison. Should the compared objects have exactly the same value then 
 * the Tree orders the instances by their name, as returned by <code>getName</code>.</p>
 * <p>The unique feature of the <code>OntologySortTree</code> is its ability to sort on a  
 * property that may contain different types. In such cases, the following arbitrary order is 
 * used:
 * <ul>
 * <li><code><b>NULL</b></code> - null values are always placed first in the tree.</li>
 * <li><b>Boolean</b> - boolean values appear after null values.</li>
 * <li><b>Number</b> - numbers are next and are converted to doubles and then compared.</li>
 * <li><b>String</b> - strings come after numbers.</li>
 * <li><b>Date</b> - dates are next and are stored in chronological order.</li>
 * <li><b>Object</b> - all other objects are placed last. They are sorted on the value returned
 * from the <code>toString</code> method.</li>
 * </ul>
 * </p>
 * <p>For example, in the BabyTalk project we have a Class that notes a Trend in channel 
 * readings. The class has a property, <em>start_value</em>, which represents the first reading
 * of the trend. As the Trend class covers both numeric and symbolic channels the 
 * <em>start_value</em> property takes any property type. Now suppose we have the following 
 * instances of trend we wish to sort:
 * <ul>
 * <li>instance_1 : "CMV"</li>
 * <li>instance_2 : 3.56</li>
 * <li>instance_3 : 5.7</li>
 * <li>instance_4 : -78</li>
 * <li>instance_5 : "BiPaP"</li>
 * <li>instance_6 : 123</li>
 * <li>instance_7 : -0.2</li>
 * <li>instance_8 : "Setting C"</li>
 * </ul>
 * </p>
 * <p>The instances will be ordered as:
 * <ul>
 * <li>instance_4 : -78</li>
 * <li>instance_7 : -0.2</li>
 * <li>instance_2 : 3.56</li>
 * <li>instance_3 : 5.7</li>
 * <li>instance_6 : 123</li>
 * <li>instance_5 : "BiPaP"</li>
 * <li>instance_1 : "CMV"</li>
 * <li>instance_8 : "Setting C"</li>
 * </ul>
 * </p>
 * <p>The tree is stored in natural ascending order but the values can be returned in the 
 * reverse, or descending, order. In descending order the list of instances above would be 
 * returned as:
 * <ul>
 * <li>instance_8 : "Setting C"</li>
 * <li>instance_1 : "CMV"</li>
 * <li>instance_5 : "BiPaP"</li>
 * <li>instance_6 : 123</li>
 * <li>instance_3 : 5.7</li>
 * <li>instance_2 : 3.56</li>
 * <li>instance_7 : -0.2</li>
 * <li>instance_4 : -78</li>
 * </ul>
 * </p>
 * @author D. Westwater, University of Aberdeen
 * @since 0.11
 */
public class OntologySortTree {

	/** Collates the instances in the natural ascending order. */
	public static final boolean ASCENDING_ORDER = true;

	/** Collates the instances in the natural descending order. */
	public static final boolean DESCENDING_ORDER = false;

	/** The head or top node of the tree. */
	private OntologySortTreeNode head = null;

	/** The property to be sorted on. */
	private String sortProperty = null;
	
	/**
	 * Create a new sort tree based on the given property.
	 * @param propertyName the property to be sorted on.
	 */
	public OntologySortTree(String propertyName) {
		this.sortProperty = propertyName;
	}

	/**
	 * Puts the given instance into the tree. The tree is always sorted and 
	 * new instances are added in their correct place in the order.
	 * @param instance the instance to be added.
	 */
	public void put(OntologyInstanceHandler instance) {
		if (this.head == null) {
			this.head = new OntologySortTreeNode(
					instance.getOwnValue(this.sortProperty), instance);
		}
		else {
			this.head.put(instance.getOwnValue(this.sortProperty), instance);
		}
	}
	
	/**
	 * Puts all the elements in the supplied collection into the tree. They will all be 
	 * placed in the correct position to maintain the ordered nature of the tree.
	 * @param instances the collection of instances to be added.
	 */
	public void putAll(Collection<OntologyInstanceHandler> instances) {
		for (OntologyInstanceHandler eachInstance : instances) {
			put(eachInstance);
		}
	}
	
	/**
	 * Retrieve a sorted collection of the instances currently in this tree. The order can 
	 * be either ascending or descending.
	 * @param ascendingOrder the flag to determine in which direction the elements are to be 
	 * ordered.
	 * @return an ArrayList of instances preserving the sorted order of the tree.
	 */
	public ArrayList<OntologyInstanceHandler> sort(boolean ascendingOrder) {
		ArrayList<OntologyInstanceHandler> sortedList = new ArrayList<OntologyInstanceHandler>();
		
		if (this.head != null) {
			if (ascendingOrder) {
				sortedList = this.head.getLeftOrder(sortedList);
			}
			else {
				sortedList = this.head.getRightOrder(sortedList);
			}
		}
		return sortedList;
	}
	
	/**
	 * A convenience method for performing both the placing of elements into the tree and 
	 * the retrieving the sorted collection of those elements.
	 * @param instances the collection of instances to be sorted.
	 * @param ascendingOrder the direction in which the sorting is to be performed.
	 * @return an ArrayList of instances preserving the sorted order of the tree.
	 */
	public ArrayList<OntologyInstanceHandler> sort(Collection<OntologyInstanceHandler> instances, 
			boolean ascendingOrder) {
		
		putAll(instances);
		
		ArrayList<OntologyInstanceHandler> sortedList = new ArrayList<OntologyInstanceHandler>();
		if (this.head != null) {
			if (ascendingOrder) {
				sortedList = this.head.getLeftOrder(sortedList);
			}
			else {
				sortedList = this.head.getRightOrder(sortedList);
			}
		}
		return sortedList;
	}
}
