/*
 * File: OntologySortTreeNode
 *
 * CHANGE HISTORY
 * 0.11	23/06/09	Base line file
 * 0.13	02/07/09	Fixed bug with the comparison of keys being put into the tree.
 */

/*
 * Copyright (c) 2007-2009, the University of Aberdeen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted FOR RESEARCH PURPOSES ONLY, provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * 		this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation and/or 
 *    other materials provided with the distribution.
 * 3. Neither the name of the University of Aberdeen nor the names of its contributors 
 * 	  may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 *    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 *    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *     
 *  Redistribution and use for purposes other than research requires special permission by the
 *  copyright holders and contributors. Please contact Ehud Reiter (e.reiter@abdn.ac.uk) for
 *  more information.
 */
package uk.ac.abdn.carados.util;

import java.util.ArrayList;

import uk.ac.abdn.carados.OntologyInstanceHandler;

/**
 * This is the base class for representing a node in the OntologySortTree. It also performs
 * the sorting, which takes place as elements are added to the tree. Each node contains:
 * <ul>
 * <li>a key - the value of the property being sorted on.</li>
 * <li>a value - the actual ontology instance.</li>
 * <li>a left child - a reference to another tree node of an element that comes before this 
 * one.</li>
 * <li>a right child - a reference to another tree node of an element that comes after this 
 * one.</li>
 * </ul>
 * @since 0.11
 * @author D. Westwater, University of Aberdeen.
 */
class OntologySortTreeNode {
	
	/**
	 * The value of the property for the instance at this node. This is a local 
	 * reference to avoid having to make constant calls to 
	 * <code>instance.getOwnValue(property)</code>.
	 */
	Object key;
	
	/** The instance. */
	OntologyInstanceHandler value;
	
	/** The left child branch of the tree. */
	OntologySortTreeNode leftChild = null;
	
	/** THe right child branch of the tree. */
	OntologySortTreeNode rightChild = null;

	/**
	 * Creates a new node with the given key value and instance. The child branches are
	 * initially set to <code>null</code>.
	 * @param objectKey the value of the property for this instance.
	 * @param instance the instance.
	 */
	OntologySortTreeNode(Object objectKey, OntologyInstanceHandler instance) {
		this.key = objectKey;
		this.value = instance;
	}

	/**
	 * <p>Places the new instance into the tree hierarchy. If the instance is deemed to come 
	 * before the current Node then it is added on the left branch. If the instance is deemed
	 * to come after the current Node then it is added to the right branch.</p>
	 * @param instanceKey the value of the sorting property for this instance.
	 * @param instance the instance to be added.
	 */
	void put(Object instanceKey, OntologyInstanceHandler instance) {
		if (this.key == null) {
			if (instanceKey != null) {
				addToRightChild(instanceKey, instance);
			}
			else {
				compareNames(instanceKey, instance);
			}
		}
		else if (instanceKey == null) {
			addToLeftChild(instanceKey, instance);
		}
		else if (this.key instanceof Number && instanceKey instanceof Number) {
			compareDifferentNumbers(instanceKey, instance);
		}
		else if (this.key.getClass().equals(instanceKey.getClass())) {
			compareSameClass(instanceKey, instance);
		}
		else {
			compareDifferentClasses(instanceKey, instance);
		}
	}
	
	/**
	 * Convenience method that compares the natural order of two strings and taking care
	 * of <code>null</code> values. <code>null</code>s are deemed to come before any 
	 * formalised string.
	 * @param left the first string to be compared.
	 * @param right the second string to be compared against.
	 * @return <code>int</code> whose value is -1 if the left string comes before the 
	 * right string, 0 if both strings are exactly the same and +1 if the right string 
	 * comes before the left string. 
	 */
	private int compareStrings(String left, String right) {
		int compare = 0;
		
		if (left == null) {
			compare = -1;
		}
		else if (right == null) {
			compare = 1;
		}
		else {
			compare = left.compareTo(right);
		}
		return compare;
	}

	/**
	 * Compares the name of the instance to be added with the name of the current node instance.
	 * This is used when the normal property value being sorted on is exactly the same for the 
	 * current instance and the new instance.
	 * @param instanceKey the value of the sorting property for this instance.
	 * @param instance the instance to be added.
	 */
	private void compareNames(Object instanceKey,
			OntologyInstanceHandler instance) {
		
		if (compareStrings(instance.getName(), this.value.getName()) <= 0) {
			addToRightChild(instanceKey, instance);
		}
		else {
			addToLeftChild(instanceKey, instance);
		}
	}

	/**
	 * Converts the property values of the current node and the new instance into 
	 * <code>double</code> values and compares them. It assumes the values being converted
	 * are both <code>Number</code>s.
	 * @param instanceKey the value of the sorting property for this instance.
	 * @param instance the instance to be added.
	 */
	private void compareDifferentNumbers(Object instanceKey,
			OntologyInstanceHandler instance) {
		
		double keyValue = ((Number)this.key).doubleValue();
		double instanceValue = ((Number)instanceKey).doubleValue();
		
		if (instanceValue < keyValue) {
			addToLeftChild(instanceKey, instance);
		}
		else if (instanceValue > keyValue) {
			addToRightChild(instanceKey, instance);
		}
		else {
			compareNames(instanceKey, instance);
		}
	}

	/**
	 * Performs the comparison on the property values when they are both of the same type. 
	 * The natural approach is to use the base type's <code>compareTo</code> method if it 
	 * exists. Failing this, the values are converted to <code>String</code>s using the 
	 * <code>toString()</code> method and these <code>String</code>s are then compared.
	 * @param instanceKey the value of the sorting property for this instance.
	 * @param instance the instance to be added.
	 */
	@SuppressWarnings("unchecked")
	private void compareSameClass(Object instanceKey,
			OntologyInstanceHandler instance) {
		
		if (instanceKey instanceof Comparable) {
			if (((Comparable<Object>) instanceKey).compareTo(this.key) < 0) {
				addToLeftChild(instanceKey, instance);
			}
			else {
				addToRightChild(instanceKey, instance);
			}
		}
		else {
			int comparison = 
				compareStrings(this.key.toString(), instanceKey.toString());
			
			if (comparison < 0) {
				addToRightChild(instanceKey, instance);
			}
			else if (comparison > 0) {
				addToLeftChild(instanceKey, instance);
			}
			else {
				compareNames(instanceKey, instance);
			}
		}
	}

	/**
	 * Controls the comparison of values of different base types. The order these are sorted
	 * is explained in the <code>OntologySortTree</code> class.
	 * @param instanceKey the value of the sorting property for this instance.
	 * @param instance the instance to be added.
	 * @see OntologySortTree
	 */
	private void compareDifferentClasses(Object instanceKey,
			OntologyInstanceHandler instance) {
		
		if (this.key instanceof Boolean) {
			addToRightChild(instanceKey, instance);
		}
		else if (instanceKey instanceof Boolean) {
			addToLeftChild(instanceKey, instance);
		}
		else if (this.key instanceof Number) {
			addToRightChild(instanceKey, instance);
		}
		else if (instanceKey instanceof Number) {
			addToLeftChild(instanceKey, instance);
		}
		else if (this.key instanceof String) {
			addToRightChild(instanceKey, instance);
		}
		else if (instanceKey instanceof String) {
			addToLeftChild(instanceKey, instance);
		}
		else if (this.key instanceof java.util.Date) {
			addToRightChild(instanceKey, instance);
		}
		else {
			addToLeftChild(instanceKey, instance);
		}
	}

	/**
	 * The new instance is deemed to come before the current instance and will be added to the 
	 * left branch. If the leftChild of this node is null a new element is created for the new
	 * instance. Otherwise, <code>put(...)</code> is recursively called on the left child.
	 * @param instanceKey the value of the sorting property for this instance.
	 * @param instance the instance to be added.
	 */
	private void addToLeftChild(Object instanceKey,
			OntologyInstanceHandler instance) {
		
		if (this.leftChild == null) {
			this.leftChild = new OntologySortTreeNode(instanceKey, instance);
		}
		else {
			this.leftChild.put(instanceKey, instance);
		}
	}

	/**
	 * The new instance is deemed to come after the current instance and will be added to the 
	 * right branch. If the right Child of this node is null a new element is created for the 
	 * new instance. Otherwise, <code>put(...)</code> is recursively called on the right child.
	 * @param instanceKey the value of the sorting property for this instance.
	 * @param instance the instance to be added.
	 */
	private void addToRightChild(Object instanceKey,
			OntologyInstanceHandler instance) {
		
		if (this.rightChild == null) {
			this.rightChild = new OntologySortTreeNode(instanceKey, instance);
		}
		else {
			this.rightChild.put(instanceKey, instance);
		}
	}

	/**
	 * Retrieves the ordered <code>ArrayList</code> of instances in ascending order. The left
	 * branch is always added before the current element and then the right branch elements 
	 * are added after the current element but before going back to the parent node.
	 * @param sortedList the <code>ArrayList</code> the elements are added to.
	 * @return the sorted <code>ArrayList</code> for this node and all its children.
	 */
	ArrayList<OntologyInstanceHandler> getLeftOrder(
			ArrayList<OntologyInstanceHandler> sortedList) {
		
		if (this.leftChild != null) {
			this.leftChild.getLeftOrder(sortedList);
		}
		sortedList.add(this.value);
		if (this.rightChild != null) {
			this.rightChild.getLeftOrder(sortedList);
		}
		return sortedList;
	}

	/**
	 * Retrieves the ordered <code>ArrayList</code> of instances in descending order. The right
	 * branch is always added before the current element and then the left branch elements 
	 * are added after the current element but before going back to the parent node.
	 * @param sortedList the <code>ArrayList</code> the elements are added to.
	 * @return the sorted <code>ArrayList</code> for this node and all its children.
	 */
	ArrayList<OntologyInstanceHandler> getRightOrder(
			ArrayList<OntologyInstanceHandler> sortedList) {
		
		if (this.rightChild != null) {
			this.rightChild.getRightOrder(sortedList);
		}
		sortedList.add(this.value);
		if (this.leftChild != null) {
			this.leftChild.getRightOrder(sortedList);
		}
		return sortedList;
	}
}
